//-----------------------------------------------------------------------------
// Project            	GeometryKit
// Class		G3DLine
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Thu Sep  9 11:05:08 CEST 1999
//
// Copyright (c) Philippe C.D. Robert
//
// The GeometryKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DLine.m,v 1.9 2001/01/05 20:50:48 robert Exp $
//-----------------------------------------------------------------------------

#import "G3DLine.h"

#import "G3DVector3f.h"
#import "G3DTuple3f.h"
#import "G3DBox.h"
#import "G3DPlane.h"
#import "G3DSphere.h"
#import "G3DFunctions.h"
#import "G3DDefs.h"
#import "G3DVectorFunc.h"
#import "G3DMatrixFunc.h"

NSString *G3DLineException = @"G3DLineException";

@implementation G3DLine

//-----------------------------------------------------------------------------
// class methods
//-----------------------------------------------------------------------------

+ (G3DLine *)line
{
  return [[[G3DLine alloc] init] autorelease];
}

+ (G3DLine *)lineWithOrigin:(G3DTuple3f *)s endPoint:(G3DTuple3f *)e
{
  return [[[G3DLine alloc] initWithOrigin:s endPoint:e] autorelease];
}

+ (G3DLine *)lineWithOrigin:(G3DTuple3f *)s direction:(G3DVector3f *)vec factor:(float)f
{
  return [[[G3DLine alloc] initWithOrigin:s direction:vec factor:f] autorelease];
}

+ (G3DLine *)lineWithLine:(G3DLine *)aLine
{
  return [[[G3DLine alloc] initWithLine:aLine] autorelease];
}

//-----------------------------------------------------------------------------
// init and free
//-----------------------------------------------------------------------------

- (id)init
{
  static float tmp1[3] = {0.0,0.0,0.0};
  static float tmp2[3] = {1.0,1.0,1.0};

  return [self initWithOrigin:[G3DTuple3f tuple3fWithElements:tmp1] direction:[G3DVector3f vector3fWithValues:tmp2] factor:1.0f];
}

- (id)initWithOrigin:(G3DTuple3f *)s endPoint:(G3DTuple3f *)e
{
  float *tmp = [e elements];
  float f;

  tmp[0] -= [s x];
  tmp[1] -= [s y];
  tmp[2] -= [s z];

  f = G3DDistance3fv(tmp,[s elements]);

  return [self initWithOrigin:s direction:[G3DVector3f vector3fWithValues:tmp] factor:f];
}

- (id)initWithOrigin:(G3DTuple3f *)s direction:(G3DVector3f *)vec factor:(float)f
{
  if ((self = [super init])) {
    origin = [s copy];
    direction = [vec copy];
    [direction normalise];
    t = f;
  }
  return self;
}

- (id)initWithLine:(G3DLine *)aLine
{
  return [self initWithOrigin:[aLine origin] direction:[aLine direction] factor:[aLine factor]];
}

- (void)dealloc
{
  [direction release];
  [origin release];

  [super dealloc];
}

//-----------------------------------------------------------------------------
// math
//-----------------------------------------------------------------------------

- (BOOL)isParallelLine:(G3DLine *)aLine
{
  float *d1 = [direction elements];
  float *d2 = [[aLine direction] elements];
  float cr_pr[3];

  G3DMultiplyVector3fv(cr_pr,d1,d2);

  return (G3DScalarProduct3fv(cr_pr,cr_pr) == 0.0f) ? YES : NO;
}

- (BOOL)intersectsObject:(id)anObject
{
  Class objClass = [anObject class];

  if (objClass == [G3DLine class]) {
    return [self intersectsLine:anObject];
  }
  else if (objClass == [G3DSphere class]) {
    return [self intersectsSphere:anObject];
  }
  else if (objClass == [G3DBox class]) {
    return [self intersectsBox:anObject];
  }
  else if (objClass == [G3DPlane class]) {
    return [self intersectsPlane:anObject intersection:nil];
  }
  else {
    [NSException raise:G3DLineException format:@"Unknown volume class!"];
  }
  return NO;
}

- (BOOL)intersectsLine:(G3DLine *)ln
{
  float *o1 = [origin elements];
  float *d1 = [direction elements];
  float *o2 = [[ln origin] elements];
  float *d2 = [[ln direction] elements];
  float denom;
  float _t_v1[3];
  float _t_v2[3];
  float t2 = [ln factor];
  float _r1, _r2;

  /*
   * o1 + t * d1 = o2 + t2 * d2 
   *
   * <=>
   *
   *      |o2 -o1 d2 d1xd2|
   * t == -----------------
   *          ||d1xd2||^2
   *
   * AND
   *
   *       |o2 -o1 d1 d1xd2|
   * t2 == -----------------
   *           ||d1xd2||^2
   *
   */

  G3DMultiplyVector3fv(_t_v1,d1,d2);
  denom = G3DScalarProduct3fv(_t_v1,_t_v1);

  // Parallel if equals 0
  if (denom == 0.0f) {
    return NO;
  }

  // Faster...
  denom = 1.0f/denom;

  G3DSubVectors3fv(_t_v1,o2,o1);
  G3DMultiplyVector3fv(_t_v2,d1,d2);

  _r1 = G3DDeterminante3fv(_t_v1,d2,_t_v2) * denom;
  _r2 = G3DDeterminante3fv(_t_v1,d1,_t_v2) * denom;

  return (_r1 == t && _r2 == t2) ? YES : NO;
}

- (BOOL)intersectsSphere:(G3DSphere *)sphere
{
  return [sphere intersectsLine:self];
}

- (BOOL)intersectsBox:(G3DBox *)box
{
    return [box intersectsLine:self];
}

- (BOOL)intersectsPlane:(G3DPlane *)plane intersection:(G3DTuple3f *)pt
{
  return [plane intersectsLine:self intersection:pt];
}

//-----------------------------------------------------------------------------
// accessor methods
//-----------------------------------------------------------------------------

- (void)setFactor:(float)val
{
  NSAssert(val!=0.0f,@"The line factor must not be 0!");

  t = val;
}

- (float)factor
{
  return t;
}

- (G3DTuple3f *)origin
{
  return origin;
}

- (void)setOrigin:(G3DTuple3f *)o
{
  [origin autorelease];
  origin = [o copy];
}

- (G3DVector3f *)direction
{
  return direction;
}

- (void)setDirection:(G3DVector3f *)d
{
  [direction autorelease];
  direction = [d copy];

  [direction normalise];
}

- (BOOL)isEqualToLine:(G3DLine *)aLine
{
  //~ This is not everything!!!
  if ([origin isEqualToTuple:[aLine origin]] &&
      [direction isEqualToTuple:[aLine direction]]) {
    return YES;
  }
  else {
    return NO;
  }
}

//-----------------------------------------------------------------------------
// NSCoding
//-----------------------------------------------------------------------------

- (void)encodeWithCoder:(NSCoder *)aCoder
{
  [aCoder encodeObject:origin];
  [aCoder encodeObject:direction];
  [aCoder encodeValueOfObjCType:@encode(float) at:&t];
}

- (id)initWithCoder:(NSCoder *)aCoder
{
  self = [super init];
  origin = [[aCoder decodeObject] retain];
  direction = [[aCoder decodeObject] retain];
  [aCoder decodeValueOfObjCType:@encode(float) at:&t];
  return self;
}

//-----------------------------------------------------------------------------
// NSCopying
//-----------------------------------------------------------------------------

- (id)copyWithZone:(NSZone *)zone
{
  return [[G3DLine allocWithZone:zone] initWithLine:self];
}

@end






