//-----------------------------------------------------------------------------
// Project            	RenderKit
// Class		G3DMatrix
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Thu Sep  9 11:05:08 CEST 1999
//
// Copyright (c) Philippe C.D. Robert
//
// The GeometryKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DMatrix.m,v 1.7 2001/01/05 20:50:48 robert Exp $
//-----------------------------------------------------------------------------

#import "G3DMatrix.h"
#import "G3DVector.h"
#import "G3DFunctions.h"
#import "G3DDefs.h"
#import "G3DVectorFunc.h"
#import "G3DMatrixFunc.h"

NSString *G3DMatrixException = @"G3DMatrixException";
NSString *InvalidMatrixSize = @"Invalid size of matrix!";

@implementation G3DMatrix

//-----------------------------------------------------------------------------
// class methods
//-----------------------------------------------------------------------------

+ (G3DMatrix *)matrixWithRows:(int)rows columns:(int)cols
{
  return [[[G3DMatrix alloc] initWithRows:rows columns:cols] autorelease];
}

+ (G3DMatrix *)matrixWithRows:(int)rows columns:(int)cols elements:(double *)vals
{
  return [[[G3DMatrix alloc] initWithRows:rows columns:cols elements:vals] autorelease];
}

+ (G3DMatrix *)matrixWithMatrix:(G3DMatrix *)aMatrix
{
  return [[[G3DMatrix alloc] initWithRows:[aMatrix rows] columns:[aMatrix columns] elements:[aMatrix elements]] autorelease];
}

//-----------------------------------------------------------------------------
// init and free
//-----------------------------------------------------------------------------

- (id)initWithRows:(int)rows columns:(int)cols
{
  return [self initWithRows:rows columns:cols elements:NULL];
}

- (id)initWithRows:(int)rows columns:(int)cols elements:(double *)vals
{
  NSAssert(rows>0 && cols>0,InvalidMatrixSize);

  if ((self = [super init])) {
    size_t n = rows * cols * sizeof(double);
    
    matrix = malloc(n);
    
    _rows = rows;
    _columns = cols;

    if (vals != NULL) {
      matrix = memcpy(matrix,vals,n);
    }
    else {
      memset(matrix,0,n);
    }
  }
  return self;
}

- (id)initWithMatrix:(G3DMatrix *)aMatrix
{
  return [self initWithRows:[aMatrix rows] columns:[aMatrix columns] elements:[aMatrix elements]];
}

- (void)dealloc
{
  free(matrix);

  // this is faster than [super dealloc]...
  NSDeallocateObject (self);
  //  [super dealloc];
}

//-----------------------------------------------------------------------------
// math
//-----------------------------------------------------------------------------

- (void)addMatrix:(G3DMatrix *)a
{
  int i;
  int cnt = _rows * _columns;
  double *m = [a elements];

  if ([a columns] != _columns || [a rows] != _rows) {
    [NSException raise:G3DMatrixException format:InvalidMatrixSize];
    return;
  }

  for (i=0;i<cnt;i++) {
    matrix[i] += *m++;
  }
}

- (void)addMatrix:(G3DMatrix *)a plus:(G3DMatrix *)b
{
  int i;
  int cnt = _rows * _columns;
  double *m1 = [a elements];
  double *m2 = [b elements];

  if(([a columns] != [b columns] && [a columns] != _columns) || ([a rows] != [b rows] && [a rows] != _rows)) {
    [NSException raise:G3DMatrixException format:InvalidMatrixSize];
    return;
  }

  for (i=0;i<cnt;i++) {
    matrix[i] = *m1++ + *m2++;
  }
}

- (void)subMatrix:(G3DMatrix *)a
{
  int i;
  int cnt = _rows * _columns;
  double *m = [a elements];

  if ([a columns] != _columns || [a rows] != _rows) {
    [NSException raise:G3DMatrixException format:InvalidMatrixSize];
    return;
  }

  for (i=0;i<cnt;i++) {
    matrix[i] -= *m++;
  }
}

- (void)subMatrix:(G3DMatrix *)a minus:(G3DMatrix *)b
{
  int i;
  int cnt = _rows * _columns;
  double *m1 = [a elements];
  double *m2 = [b elements];

  if(([a columns] != [b columns] && [a columns] != _columns) || ([a rows] != [b rows] && [a rows] != _rows)) {
    [NSException raise:G3DMatrixException format:InvalidMatrixSize];
    return;
  }

  for (i=0;i<cnt;i++) {
    matrix[i] = *m1++ - *m2++;
  }
}

- (void)multiplyByMatrix:(G3DMatrix *)a
{
  int i,j,k;
  double *tmp = malloc(_rows*_columns*sizeof(double));
  double *m = [a elements];
  double val;

  if(_columns != [a rows]) {
    [NSException raise:G3DMatrixException format:InvalidMatrixSize];
    return;
  }

  for (i=0;i<_rows;i++) {
    for (j=0;j<_columns;j++) {
      val = 0;
      for (k=0;k<_columns;k++) {
	val += matrix[k*_rows+i] * m[j*_rows+k];
      }
      tmp[j*_rows+i] = val;
    }
  }

  [self setElements:tmp];

  free(tmp);
}

- (void)multiplyByMatrix:(G3DMatrix *)a with:(G3DMatrix *)b
{
  int i,j,k;
  double *tmp = malloc(_rows*_columns*sizeof(double));
  double *m1 = [a elements];
  double *m2 = [b elements];
  double val;

  if([a columns] != [a rows] || _columns != [a columns]) {
    [NSException raise:G3DMatrixException format:InvalidMatrixSize];
    return;
  }

  for (i=0;i<_rows;i++) {
    for (j=0;j<_columns;j++) {
      val = 0;
      for (k=0;k<_columns;k++) {
	val += m1[k*_rows+i] * m2[j*_rows+k];
      }
      tmp[j*_rows+i] = val;
    }
  }

  [self setElements:tmp];

  free(tmp);
}

- (void)multiplyByMatrix:(G3DMatrix *)a withTransposed:(G3DMatrix *)b
{
  G3DMatrix *tmp = [b transposedMatrix];

  [self multiplyByMatrix:a with:tmp];
}

- (void)multiplyByTransposedMatrix:(G3DMatrix *)a with:(G3DMatrix *)b
{
  G3DMatrix *tmp = [a transposedMatrix];

  [self multiplyByMatrix:tmp with:b];
}

- (void)multiplyByTransposedMatrix:(G3DMatrix *)a withTransposed:(G3DMatrix *)b
{
  G3DMatrix *tmp1 = [a transposedMatrix];
  G3DMatrix *tmp2 = [b transposedMatrix];

  [self multiplyByMatrix:tmp1 with:tmp2];
}

- (void)invert
{
}

- (G3DMatrix *)invertedMatrix
{
  G3DMatrix *m = [G3DMatrix matrixWithMatrix:self];

  [m invert];

  return m;
}

- (void)adjoint
{
}

- (G3DMatrix *)adjointedMatrix
{
  G3DMatrix *m = [G3DMatrix matrixWithMatrix:self];

  [m adjoint];

  return m;
}

- (void)transpose
{
}

- (G3DMatrix *)transposedMatrix
{
  G3DMatrix *m = [G3DMatrix matrixWithMatrix:self];

  [m transpose];

  return m;
}

- (void)negate
{
}

- (G3DMatrix *)negatedMatrix
{
  G3DMatrix *m = [G3DMatrix matrixWithMatrix:self];

  [m negate];

  return m;
}

- (void)multiplyByScalar:(double)scalar
{
  int i;
  int c = _rows * _columns;

  for (i=0;i<c;i++) {
    matrix[i] *= scalar;
  }
}

- (void)divideByScalar:(double)scalar
{
  int i;
  int c = _rows * _columns;

  if (scalar == 0.0) {
    [NSException raise:G3DMatrixException format:@"Division by zero!"];
    return;
  }

  for (i=0;i<c;i++) {
    matrix[i] /= scalar;
  }
}

- (int)svdWithU:(G3DMatrix **)u W:(G3DMatrix **)w V:(G3DMatrix **)v
{
    return -1;
}

- (double)determinant
{
    return -1.0;
}

- (double)trace
{
    return -1.0;
}

- (int)rank
{
    return -1;
}

//-----------------------------------------------------------------------------
// accessor methods
//-----------------------------------------------------------------------------

- (int)rows
{
  return _rows;
}

- (int)columns
{
  return _columns;
}

- (BOOL)isQuadratic
{
  return (_columns == _rows);
}

- (G3DVector *)rowVectorAtIndex:(int)n
{
  G3DVector *vec = [G3DVector vectorWithSize:_columns];
  double *tmp = malloc(_columns * sizeof(double));
  int i;

  for (i=0;i<_columns;i++) {
    *(tmp+i)=*matrix+n+(i*_rows);
  }

  [vec setElements:tmp];
  free(tmp);

  return vec;
}

- (G3DVector *)columnVectorAtIndex:(int)n
{
  G3DVector *vec = [G3DVector vectorWithSize:_rows];
  double *tmp = malloc(_rows * sizeof(double));
  int r = (n*_rows);
  int i;

  for (i=0;i<_rows;i++) {
    *(tmp+i)=*matrix+r+i;
  }

  [vec setElements:tmp];
  free(tmp);

  return vec;
}

- (void)setIdentity
{
}

- (void)setZero
{
  memset(matrix,0,_rows*_columns*sizeof(double));
}

- (void)setMatrix:(G3DMatrix *)aMatrix
{
  double *vals = [aMatrix elements];
  
  [self setElements:vals];
}

- (void)setElements:(double *)vals
{
  int i;
  int c = _rows * _columns;

  for (i=0;i<c;i++) {
    *(matrix+i) = *vals++;
  }
}

- (void)setElement:(double)val atRow:(int)row column:(int)col
{
  matrix[(col*_rows) + row] = val;
}

- (double)elementAtRow:(int)row column:(int)col
{
  return matrix[(col*_rows) + row];
}

- (double *)elements
{
  return matrix;
}

- (void)setElements:(double *)vals inRow:(int)row
{
  int i;

  for (i=0;i<_columns;i++) {
    matrix[row + (i*_rows)] = vals[i];
  }
}

- (void)setVector:(G3DVector *)vec inRow:(int)row
{
  int i;
  double *vals = [vec elements];

  if ([vec size] != _columns) {
    [NSException raise:G3DMatrixException format:@"Invalid size of vector!"];
    return;
  }

  for (i=0;i<_columns;i++) {
    matrix[row + (i*_rows)] = vals[i];
  }
}

- (void)setElements:(double *)vals inColumn:(int)col
{
  int i;

  for (i=0;i<_rows;i++) {
    matrix[i + (col*_rows)] = vals[i];
  }
}

- (void)setVector:(G3DVector *)vec inColumn:(int)col
{
  int i;
  double *vals = [vec elements];

  if ([vec size] != _rows) {
    [NSException raise:G3DMatrixException format:@"Invalid size of vector!"];
    return;
  }

  for (i=0;i<_rows;i++) {
    matrix[i + (col*_rows)] = vals[i];
  }
}

//-----------------------------------------------------------------------------
// misc
//-----------------------------------------------------------------------------

- (BOOL)isEqualToMatrix:(G3DMatrix *)aMatrix
{
  int i;
  int cnt = _rows*_columns;
  double *m = [aMatrix elements];

  if (_columns != [aMatrix columns] || _rows != [aMatrix rows]) {
    return NO;
  }

  for (i=0;i<cnt;i++) {
    if (matrix[i] != m[i]) {
      return NO;
    }
  }

  return YES;
}

- (NSString *)description
{
  return [NSString stringWithFormat:@"%dx%d G3DMatrix",_rows,_columns];
}

//-----------------------------------------------------------------------------
// NSCoding
//-----------------------------------------------------------------------------

- (void)encodeWithCoder:(NSCoder *)aCoder
{
  [aCoder encodeValueOfObjCType:@encode(int) at:&_rows];
  [aCoder encodeValueOfObjCType:@encode(int) at:&_columns];

  [aCoder encodeBytes:&matrix length:_rows*_columns*sizeof(double)];
}

- (id)initWithCoder:(NSCoder *)aCoder
{
  unsigned n;

  self = [super init];

  [aCoder decodeValueOfObjCType:@encode(int) at:&_rows];
  [aCoder decodeValueOfObjCType:@encode(int) at:&_columns];

  n=_rows*_columns*sizeof(double);
  matrix = (double*)[aCoder decodeBytesWithReturnedLength:&n];

  return self;
}

//-----------------------------------------------------------------------------
// NSCopying
//-----------------------------------------------------------------------------

- (id)copyWithZone:(NSZone *)zone
{
  return [[G3DMatrix allocWithZone:zone] initWithMatrix:self];
}

@end






