//-----------------------------------------------------------------------------
// Project            	GeometryKit
// Class		G3DQuaternion
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Thu Sep  9 11:05:08 CEST 1999
//
// Copyright (c) Philippe C.D. Robert
//
// The GeometryKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DQuaternion.m,v 1.9 2001/01/05 20:50:48 robert Exp $
//-----------------------------------------------------------------------------

#import "G3DQuaternion.h"

#import "G3DMatrix3f.h"
#import "G3DMatrix4f.h"
#import "G3DVector3f.h"
#import "G3DTuple4f.h"

#import "G3DDefs.h"
#import "G3DQuatFunc.h"
#import "G3DVectorFunc.h"
#import "G3DMatrixFunc.h"

NSString *G3DQuaternionException = @"G3DQuaternionException";

@implementation G3DQuaternion

//-----------------------------------------------------------------------------
// class methods
//-----------------------------------------------------------------------------

static Class G3DTuple4fClass;
static Class G3DMatrix4fClass;
static Class G3DMatrix3fClass;

+ (void)initialize
{
  if (self == [G3DQuaternion class]) {
    G3DTuple4fClass  = [G3DTuple4f class];
    G3DMatrix4fClass  = [G3DMatrix4f class];
    G3DMatrix3fClass  = [G3DMatrix3f class];
  }
}

+ (G3DQuaternion *)quaternion
{
  return [[[G3DQuaternion alloc] init] autorelease];
}

+ (G3DQuaternion *)quaternionWithElements:(float *)values
{
  return [[[G3DQuaternion alloc] initWithElements:values] autorelease];
}

+ (G3DQuaternion *)quaternionWithQuaternion:(G3DQuaternion *)aTuple
{
  return [[(G3DTuple4f*)[G3DQuaternion alloc] initWithTuple:aTuple] autorelease];
}

+ (G3DQuaternion *)quaternionWithMatrix4f:(G3DMatrix4f *)aMatrix
{
  return [[[G3DQuaternion alloc] initWithMatrix4f:aMatrix] autorelease];
}

+ (G3DQuaternion *)quaternionWithEulerRep:(float *)ev
{
  return [[[G3DQuaternion alloc] initWithEulerRep:ev] autorelease];
}

//-----------------------------------------------------------------------------
// init and free
//-----------------------------------------------------------------------------

- (id)init
{
  if ((self = [super initWithX:0.0 y:0.0 z:0.0 w:1.0])) {
  }
  return self;
}

- (id)initWithMatrix4f:(G3DMatrix4f *)aMatrix
{
  if ((self = [self init])) {
    G3DQuatFromMatrixf(tuple,[aMatrix elements]);
  }
  return self;
}

- (id)initWithEulerRep:(float *)ev
{
  if ((self = [self init])) {
    G3DQuatFromEulerRepf(tuple,ev);
  }
  return self;
}

//-----------------------------------------------------------------------------
// Math
//-----------------------------------------------------------------------------

- (float)norm
{
  return (tuple[0]*tuple[0] + tuple[1]*tuple[1] + tuple[2]*tuple[2] + tuple[3]*tuple[3]);
}

- (void)conjugate
{
  tuple[0] = -tuple[0];
  tuple[1] = -tuple[1];
  tuple[2] = -tuple[2];
}

- (G3DQuaternion *)conjugatedQuaternion
{
  G3DQuaternion *tmp = [G3DQuaternion quaternionWithQuaternion:self];

  [tmp conjugate];
  return tmp;
}

- (void)invert
{
  register float sql = tuple[0]*tuple[0] + tuple[1]*tuple[1] + tuple[2]*tuple[2] + tuple[3]*tuple[3];

  if (sql != 0.0) {
    sql = 1.0/sql;
  }
  else {
    sql = 1.0;
  }

  tuple[0] *= -sql;
  tuple[1] *= -sql;
  tuple[2] *= -sql;
  tuple[3] *=  sql;
}

- (G3DQuaternion *)invertedQuaternion
{
  G3DQuaternion *tmp = [G3DQuaternion quaternionWithQuaternion:self];

  [tmp invert];
  return tmp;
}

- (void)normalise
{
  register float sql = tuple[0]*tuple[0] + tuple[1]*tuple[1] + tuple[2]*tuple[2] + tuple[3]*tuple[3];
  
  if (sql > 0.0) {
    sql = (1.0/ sqrt(sql));

    tuple[0] *= sql;
    tuple[1] *= sql;
    tuple[2] *= sql;
    tuple[3] *= sql;
  }
}

- (G3DQuaternion *)normalisedQuaternion
{
  G3DQuaternion *tmp = [G3DQuaternion quaternionWithQuaternion:self];

  [tmp normalise];
  return tmp;
}

- (void)interpolate:(G3DQuaternion *)quat factor:(float)val
{
  float tmp[4];
  float *to = [quat elements];

  G3DInterpolateQuatf(tmp,tuple,to,val);
  G3DCopyVector4fv(tuple,tmp);
}

- (void)multiplyByQuaternion:(G3DQuaternion *)aQuat
{
  float *a = [aQuat elements];

  G3DMultiplyQuatf(tuple, tuple, a);
}

- (void)multiplyByQuaternion:(G3DQuaternion *)quatA and:(G3DQuaternion *)quatB
{
  float *a = [quatA elements];
  float *b = [quatB elements];

  G3DMultiplyQuatf(tuple, a, b);
}

- (void)multiplyByInvertedQuaternion:(G3DQuaternion *)aQuat
{
  [self multiplyByQuaternion:[aQuat invertedQuaternion]];
}

- (void)multiplyByQuaternion:(G3DQuaternion *)a andInverted:(G3DQuaternion *)b
{
  [self multiplyByQuaternion:a and:[b invertedQuaternion]];
}

- (void)multiplyByScalar:(float)value
{
  tuple[0] *= value;
  tuple[1] *= value;
  tuple[2] *= value;
  tuple[3] *= value;
}

- (void)addQuaternion:(G3DQuaternion *)aQuat
{
  float *t = [aQuat elements];

  tuple[0] += t[0];
  tuple[1] += t[1];
  tuple[2] += t[2];
  tuple[3] += t[3];
}

- (void)subQuaternion:(G3DQuaternion *)aQuat
{
  float *t = [aQuat elements];

  tuple[0] -= t[0];
  tuple[1] -= t[1];
  tuple[2] -= t[2];
  tuple[3] -= t[3];
}

- (void)rotateByAngle:(float)angle axis:(G3DVector3f *)aVec
{
  float tmp[4];
  float *v = [aVec elements];
  float *a = tuple;
  float *b = tmp;

  G3DCopyVector4f(tmp,v[0],v[1],v[2],angle);
  G3DMultiplyQuatf(tuple, a, b);

  [self normalise];
}

//-----------------------------------------------------------------------------
// Accessor methods
//-----------------------------------------------------------------------------

- (void)setQuaternionWithMatrix4f:(G3DMatrix4f *)m
{
  float *mat = [m elements];

  G3DQuatFromMatrixf(tuple,mat);
}

- (G3DMatrix4f *)rotationMatrix4f
{
  float m[16];

  G3DMatrixFromQuatf(m,tuple);

  return [G3DMatrix4fClass matrix4fWithElements:m];
}

- (G3DMatrix3f *)rotationMatrix3f
{
  float m[9];

  float x2x = tuple[0] * (tuple[0] + tuple[0]);
  float x2y = tuple[0] * (tuple[1] + tuple[1]);
  float x2z = tuple[0] * (tuple[2] + tuple[2]);

  float w2x = tuple[3] * (tuple[0] + tuple[0]);
  float w2y = tuple[3] * (tuple[1] + tuple[1]);
  float w2z = tuple[3] * (tuple[2] + tuple[2]);

  float y2y = tuple[1] * (tuple[1] + tuple[1]);
  float y2z = tuple[1] * (tuple[2] + tuple[2]);
  float z2z = tuple[2] * (tuple[2] + tuple[2]);

  G3DCopyVector3f(m, 1.0-(y2y+z2z), x2y-w2z, x2z+w2y);
  G3DCopyVector3f(m+3, x2y+w2z, 1.0-(x2x+z2z), y2z-w2x);
  G3DCopyVector3f(m+6, x2z-w2y, y2z+w2x, 1.0-(x2x+y2y));

  return [G3DMatrix3fClass matrix3fWithElements:m];
}

- (void)setQuaternion:(G3DQuaternion *)q
{
  [super setValuesWithTuple:q];
}

- (void)setVector:(G3DVector3f *)vec angle:(float)val
{
  float *tmp = [vec elements];

  G3DQuatFromAngleAxisf(tuple,val,tmp);
}

- (G3DTuple4f *)angleAxisRepresentation
{
  float angleAxis[4];

  G3DAngleAxisFromQuatf(angleAxis,angleAxis,tuple);

  return [G3DTuple4fClass tuple4fWithElements:angleAxis];
}

@end






