//-----------------------------------------------------------------------------
// Project            	RenderKit
// Class		G3DVector
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Thu Sep  9 11:05:08 CEST 1999
//
// Copyright (c) Philippe C.D. Robert
//
// The GeometryKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DVector.m,v 1.7 2001/01/05 20:50:49 robert Exp $
//-----------------------------------------------------------------------------

#import "G3DVector.h"
#import "G3DMatrix.h"
#import "G3DFunctions.h"
#import "G3DDefs.h"
#import "G3DVectorFunc.h"
#import "G3DMatrixFunc.h"

NSString *G3DVectorException = @"G3DVectorException";

@implementation G3DVector

//-----------------------------------------------------------------------------
// Class methods
//-----------------------------------------------------------------------------

+ (G3DVector *)vectorWithSize:(int)val;
{
  return [[[G3DVector alloc] initWithSize:val] autorelease];
}

+ (G3DVector *)vectorWithValues:(double *)vals size:(int)val 
{
  return [[[G3DVector alloc] initWithElements:vals size:val] autorelease];
}

+ (G3DVector *)vectorWithVector:(G3DVector *)vec
{
  return [[[G3DVector alloc] initWithVector:vec] autorelease];
}

//-----------------------------------------------------------------------------
// init and free
//-----------------------------------------------------------------------------

- (id)initWithSize:(int)val
{
  return [self initWithElements:NULL size:val];
}

- (id)initWithElements:(double *)vals size:(int)val
{
  NSAssert(val > 0,@"Invalid vector size!");

  if ((self = [super init])) {
    size_t n = _size * sizeof(double);

    vector = malloc(n);
    _size = val;

    if (vals != NULL) {
      vector = memcpy(vector,vals,n);
    }
    else {
      memset(vector,0,n);
    }
  }
  return self;
}

- (id)initWithVector:(G3DVector *)vec
{
  return [self initWithElements:[vec elements] size:[vec size]];
}

- (void)dealloc
{
  free(vector);

  [super dealloc];
}

//-----------------------------------------------------------------------------
// Math
//-----------------------------------------------------------------------------

- (void)vectorByAdding:(G3DVector *)vec
{
  double *tmp = [vec elements];
  int i;
  
  if (_size != [vec size]) {
    [NSException raise:G3DVectorException format:@"Invalid size of vector!"];
    return;
  }

  for(i=0;i<_size;i++) {
    vector[i] += *tmp++;
  }
}

- (void)vectorBySubtracting:(G3DVector *)vec
{
  double *tmp = [vec elements];
  int i;
  
  if (_size != [vec size]) {
    [NSException raise:G3DVectorException format:@"Invalid size of vector!"];
    return;
  }

  for(i=0;i<_size;i++) {
    vector[i] -= *tmp++;
  }
}

- (void)vectorByMultiplyingBy:(double)aScalar
{
  int i;
  
  for(i=0;i<_size;i++) {
    vector[i] *= aScalar;
  }
}

- (void)vectorByDividingBy:(double)aScalar
{
  int i;
  
  if (aScalar == 0.0) {
    [NSException raise:G3DVectorException format:@"Division by zero!"];
    return;
  }

  for(i=0;i<_size;i++) {
    vector[i] /= aScalar;
  }
}

- (void)interpolateBetween:(G3DVector *)first and:(G3DVector *)second factor:(double)factor
{
}

- (void)vectorByMultiplyingMatrix:(G3DMatrix *)aMatrix and:(G3DVector *)vec
{
  int i,j;
//  int cols = [aMatrix columns];
  double *m = [aMatrix elements];
  double *v = [vec elements];
  double val;
  int rows = [aMatrix rows];

  if ([aMatrix columns] != [vec size]) {
    [NSException raise:G3DVectorException format:@"Invalid sizes!"];
    return;
  }

  // Dynamically resize self
  realloc(vector,[aMatrix rows]*sizeof(double));

  for (i=0;i<[aMatrix rows];i++) {
    val = 0.0;
    for (j=0;j<[vec size];j++) {
      val += m[j*rows+i]*v[j];
    }
    vector[i] = val;
  }
}

- (void)vectorByMultiplyingVector:(G3DVector *)vec and:(G3DMatrix *)m
{
}

- (double)dotProduct:(G3DVector *)aVec
{
  double t = 0.0;
  double *v = [aVec elements];
  int i;

  if (_size != [aVec size]) {
    [NSException raise:G3DVectorException format:@"Invalid size of vector!"];
    return -1.0;
  }

  for (i=0;i<_size;i++) {
    t += vector[i]*v[i];
  }

  return t;
}

- (double)norm
{
  double t = 0.0;
  int i;

  for (i=0;i<_size;i++) {
    t += vector[i]*vector[i];
  }

  return sqrt(t);
}

- (double)squaredLength
{
  double t = 0.0;
  int i;

  for (i=0;i<_size;i++) {
    t += vector[i]*vector[i];
  }

  return t;
}

- (void)negate
{
  int i;

  for (i=0;i<_size;i++) {
    vector[i] *= -1.0;
  }
}

- (void)zero
{
  size_t n = _size * sizeof(double);

  memset(vector,0,n);
}

- (void)normalise
{
  double l = [self norm];
  int i;

  for (i=0;i<_size;i++) {
    vector[i] /= l;
  }

}

- (G3DVector *)normalisedVector
{
  G3DVector *tmp = [G3DVector vectorWithVector:self];

  [tmp normalise];

  return tmp;
}

- (BOOL)isEqualToVector:(G3DVector *)vec
{
  int i;
  double *v = [vec elements];

  if (_size != [vec size]) {
    return NO;
  }

  for (i=0;i<_size;i++) {
    if (vector[i] != v[i]) {
      return NO;
    }
  }
  return YES;
}

//-----------------------------------------------------------------------------
// Accessing
//-----------------------------------------------------------------------------

- (int)size
{
  return _size;
}

- (double *)elements
{
  return vector;
}

- (void)setElements:(double *)el
{
  int i;

  for (i=0;i<_size;i++) {
    vector[i] = el[i];
  }
}

- (void)setElement:(double)el atIndex:(int)index
{
  if (index < 0 || index >= _size) {
    [NSException raise:G3DVectorException format:@"Invalid index!"];
    return;
  }

  vector[index] = el;
}

- (double)elementAtIndex:(int)index
{
  if (index < 0 || index >= _size) {
    [NSException raise:G3DVectorException format:@"Invalid index!"];
    return -1.0;
  }

  return vector[index];
}

- (NSString *)description
{
  return [NSString stringWithFormat:@"G3DVector of size %d",_size];
}

//-----------------------------------------------------------------------------
// NSCoding
//-----------------------------------------------------------------------------

- (void)encodeWithCoder:(NSCoder *)aCoder
{
  [aCoder encodeValueOfObjCType:@encode(int) at:&_size];

  [aCoder encodeBytes:&vector length:_size*sizeof(double)];
}

- (id)initWithCoder:(NSCoder *)aCoder
{
  unsigned n;

  self = [super init];

  [aCoder decodeValueOfObjCType:@encode(int) at:&_size];

  n=_size*sizeof(double);
  vector = (double*)[aCoder decodeBytesWithReturnedLength:&n];

  return self;
}

//-----------------------------------------------------------------------------
// NSCopying
//-----------------------------------------------------------------------------

- (id)copyWithZone:(NSZone *)zone
{
  return [[G3DVector allocWithZone:zone] initWithVector:self];
}

@end






