//-----------------------------------------------------------------------------
// Project            	RenderKit
// Class		G3DArchiver
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Wed Nov 24 18:49:25 CET 1999
//
// Copyright (c) Philippe C.D. Robert
//
// The RenderKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DArchiver.m,v 1.11 2001/07/07 15:35:47 probert Exp $
//-----------------------------------------------------------------------------

#import "G3DArchiver.h"
#import "G3DUnarchiver.h"
#import "G3DFunctions.h"

NSString *G3DArchiverRootObjectDuplicationException = @"G3DArchiverRootObjectDuplicationException";
NSString *G3DArchiverObjectNotWrittenException = @"G3DArchiverObjectNotWrittenException";

NSString *ARCH_VERSION = @"1";

@interface G3DArchiver (Private)

- (void)_firstPassWithRoot:(id)rObj named:(NSString *)name;
- (id)_secondPassWithRoot:(id)rObj named:(NSString *)name;

@end

@implementation G3DArchiver (Private)

- (void)_firstPassWithRoot:(id)rObj named:(NSString *)name
{
  writingRoot = YES;
  findingConditionals = YES;

#ifdef DEBUG
  NSLog(@"G3DArchiver: _firstPassWithRoot:%@ named:%@",[rObj description],name);
#endif DEBUG

  NSResetHashTable(conditionals);
  NSResetMapTable(objects);

  [self encodeObject:rObj named:name];
}

- (id)_secondPassWithRoot:(id)rObj named:(NSString *)name
{
  id label = nil;

#ifdef DEBUG
  NSLog(@"G3DArchiver: _secondPassWithRoot:%@ named:%@",[rObj description],name);
#endif DEBUG

  findingConditionals = NO;

  NSResetMapTable(objects);

  label = [self encodeObject:rObj named:name];

  writingRoot = NO;

  return label;
}

@end

@implementation G3DArchiver

//-----------------------------------------------------------------------------
// Class methods
//-----------------------------------------------------------------------------

+ (BOOL)archiveObjectTree:(id)obj toFile:(NSString *)path
{
  G3DArchiver* archiver = [[[self alloc] init] autorelease];

  [archiver encodeRootObject:obj named:@"RootObject"];
  return [archiver writeToFile:path];
}

+ (id)archivedObjectTreeWithRoot:(id)obj
{
  G3DArchiver* archiver = [[[self alloc] init] autorelease];

  return [archiver encodeRootObject:obj named:@"RootObject"];
}

//-----------------------------------------------------------------------------
// Init and free
//-----------------------------------------------------------------------------

- (id)init
{
  if ((self = [super init])) {
    topLevelObjects = [[NSMutableArray alloc] init];
    pList = [[NSMutableDictionary alloc] init];
    [pList setObject:topLevelObjects forKey: @"TopLevelObjects"];

    lastObjectRep = pList;
    
    objects = NSCreateMapTable (NSNonRetainedObjectMapKeyCallBacks, NSObjectMapValueCallBacks, 119);
    conditionals = NSCreateHashTable(NSNonRetainedObjectHashCallBacks, 19);
    classes = NSCreateMapTable(NSObjectMapKeyCallBacks,NSObjectMapValueCallBacks, 19);

    [pList setObject:ARCH_VERSION forKey:@"Version"];
  }
  return self;
}

- (void)dealloc
{
  [pList release];
  [topLevelObjects release];

  NSFreeMapTable(objects);
  NSFreeHashTable(conditionals);
  NSFreeMapTable(classes);

  [super dealloc];
}

//-----------------------------------------------------------------------------
// Archiving
//-----------------------------------------------------------------------------

- (NSDictionary *)propertyList
{
  return pList;
}

- (NSString*)newLabel
{
  return [NSString stringWithFormat: @"Object%5d", ++counter];
}

- (BOOL)writeToFile:(NSString*)path
{
  return [pList writeToFile:path atomically:YES];
}

//-----------------------------------------------------------------------------
// Encoding Objects
//-----------------------------------------------------------------------------

- (id)encodeRootObject:(id)rootObject named:(NSString*)name
{
  id originalPList = pList;
  int oldCounter = counter;

  /*
   *
   * Prepare for writing the graph objects for which `rootObject' is the root 
   * node. The algorithm consists of 2 passes. In the first pass it determines
   * the nodes encoded *only* with -encodeConditionalObject:. They represent 
   * nodes that are not related directly to the graph. In the second pass 
   * objects are encoded normally, except for the conditional objects which are
   * encoded as nil. This is the same algorithm as used in GMArchiver.
   *
   */

  if (writingRoot) {
    [NSException raise:G3DArchiverRootObjectDuplicationException format:@"Root is already written!"];
  }

  /*
   * First pass
   */

  lastObjectRep = pList = nil;
  [self _firstPassWithRoot:rootObject named:name];

  /*
   * Second pass.
   */

  counter = oldCounter;
  lastObjectRep = pList = originalPList;

  return [self _secondPassWithRoot:rootObject named:name];
}

- (id)encodeConditionalObject:(id)anObject named:(NSString*)name
{
#if 0
  if (!writingRoot) {
    [NSException raise:G3DArchiverObjectNotWrittenException format:@"Object has not yet been written!"];
  }
#endif

  if (findingConditionals) {
    id value;

    if (!anObject) {
      return nil;
    }

    // Lookup anObject into the `conditionals' set.
    if ((value = (id)NSHashGet(conditionals, anObject))) {
      return value;
    }

    // Maybe it has received -encodeObject: and now is in the `objects' set. 
    if ((value = (id)NSMapGet(objects, anObject))) {
      return value;
    }

    // anObject was not written previously.
    NSHashInsert(conditionals, anObject);
  }
  else {
    // If anObject is in the `conditionals' set, it is encoded as nil.
    if (!anObject || NSHashGet(conditionals, anObject)) {
      return [self encodeObject: nil named:name];
    }
    else {
      return [self encodeObject:anObject named:name];
    }
  }

  return nil;
}

- (id)encodeObject:(id)anObject named:(NSString*)name
{
#ifdef DEBUG
  NSLog(@"G3DArchiver: encodeObject:%@ named:%@",[anObject description],name);
#endif DEBUG

  if (!anObject) {
    if (!findingConditionals && name) {
      [lastObjectRep setObject:@"nil" forKey:name];
    }
    return @"nil";
  }
  else {
    id label;
    id upperObjectRep;

    anObject = [anObject replacementObjectForSceneArchiver:self];
    label = NSMapGet(objects, anObject);

    if (findingConditionals && !label) {      
       // Lookup the object in the `conditionals' set. If the object is there, then remove it because it is no longer a conditional one.
      if ((label = NSHashGet(conditionals, anObject))) {
	NSHashRemove(conditionals, anObject);
	NSMapInsert(objects, anObject, [self newLabel]);
	return label;
      }
    }

    if (!label) {
      Class archiveClass;

      // If the object gets encoded at top level, set the label to be `name'.
      if (!level) {
	if (!name) {
	  NSLog (@"Can't encode top level object with a nil name!");
	  return nil;
	}
	label = name;
      }
      else {
	label = [self newLabel];
      }

      NSMapInsert(objects, anObject, label);

      // Temporary save the last object into upperObjectRep so we  can restore the stack of objects being encoded after anObject is encoded.

      upperObjectRep = lastObjectRep;
      archiveClass = [anObject classForSceneArchiver];

      if (!findingConditionals) {
	NSMutableDictionary* objectPList = [NSMutableDictionary dictionary];

	// If anObject is the first object in the graph that receives the -encodeObject:named: message, save its label into the topLevelObjects array.

	if (!level) {
	  [topLevelObjects addObject: (name ? name : label)];
	}

	lastObjectRep = objectPList;

	if (level) {
	  if (name) {
	    [upperObjectRep setObject:label forKey:name];
	  }
	  // Description
	  [pList setObject: objectPList forKey:label];
	}
	else {
	  // The encoded object is on the top level so encode under 'name'
	  if (name) {
	    label = name;
	  }
	  [pList setObject:objectPList forKey:label];
	}

	[objectPList setObject: NSStringFromClass(archiveClass) forKey:@"isa"];
      }
      else {
	// Remove anObject from the `conditionals' set if it is there and insert it into the `objects' set.	  
	NSHashRemove(conditionals, anObject);
      }

      level++;
      [anObject encodeWithSceneArchiver:self];
      level--;

      lastObjectRep = upperObjectRep;
    }
    else if (!findingConditionals) {
      if (name) {
	[lastObjectRep setObject:label forKey:name];
      }
    }
    return label;
  }
}

- (id)encodeString:(NSString*)anObject named:(NSString*)name
{
  if (!findingConditionals) {
    if (!anObject) {
      if (name) {
	[lastObjectRep setObject:@"nil" forKey:name];
      }
      return @"nil";
    }
    else {
      if (name) {
	[lastObjectRep setObject:anObject forKey:name];
      }
      return anObject;
    }
  }
  return @"nil";
}

- (id)encodeArray:(NSArray*)array named:(NSString*)name
{
#ifdef DEBUG
  NSLog(@"G3DArchiver: encodeArray:%@ named:%@",[array description],name);
#endif DEBUG

  if (array) {
    int i;
    int count = [array count];
    NSMutableArray *descr = [NSMutableArray arrayWithCapacity: count];

    for (i = 0; i < count; i++) {
      id object = [array objectAtIndex:i];
      [descr addObject:[self encodeObject:object named:nil]];
    }
    if (name) {
      [lastObjectRep setObject:descr forKey:name];
    }
    return descr;
  }
  else {
    if (name) {
      [lastObjectRep setObject:@"nil" forKey:name];
    }
    return @"nil";
  }
}

- (id)encodeDictionary:(NSDictionary*)dictionary named:(NSString*)name
{
  if (dictionary) {
    NSEnumerator *enumerator;
    id key;
    int c = [dictionary count];
    NSMutableDictionary *descr=[NSMutableDictionary dictionaryWithCapacity:c];

    enumerator = [dictionary keyEnumerator];

    while ((key = [enumerator nextObject])) {
      id value = [dictionary objectForKey:key];
      id keyDesc = [self encodeObject:key named:nil];
      id valueDesc = [self encodeObject:value named:nil];

      [descr setObject:valueDesc forKey:keyDesc];
    }
    if (name) {
      [lastObjectRep setObject:descr forKey:name];
    }
    return descr;
  }
  else {
    if (name) {
      [lastObjectRep setObject:@"nil" forKey:name];
    }
    return @"nil";
  }
}

- (id)encodeData:(NSData*)anObject named:(NSString*)name
{
  if (!findingConditionals) {
    if (!anObject) {
      if (name) {
	[lastObjectRep setObject: @"nil" forKey: name];
      }
      return @"nil";
    }
    else {
      if (name) {
	[lastObjectRep setObject:anObject forKey:name];
      }
      return anObject;
    }
  }

  return @"nil";
}

- (id)encodeClass:(Class)class named:(NSString*)name
{
  if (class) {
    return [self encodeString:NSStringFromClass(class) named:name];
  }
  else {
    return [self encodeString:nil named:name];
  }
}

- (id)encodeSelector:(SEL)selector named:(NSString*)name
{
  if (selector) {
    return [self encodeString:NSStringFromSelector(selector) named:name];
  }
  else {
    return [self encodeString:nil named:name];
  }
}

//-----------------------------------------------------------------------------
// Encoding C Types
//-----------------------------------------------------------------------------

- (void)encodeBytes:(const void *)b length:(unsigned)l named:(NSString *)name
{
#warning encodeBytes:length:named: not yet done!
}

- (void)encodeArrayOfTypeDoubleAt:(const double *)address count:(unsigned)cnt named:(NSString*)name
{
  if (!findingConditionals && name) {
    int i;
    NSMutableArray *array = [NSMutableArray arrayWithCapacity:cnt]; 

    for (i=0;i<cnt;i++) {
      [array addObject:[NSString stringWithFormat:@"%f", address[i]]];
    }

    [lastObjectRep setObject:array forKey:name];
  }
}

- (void)encodeArrayOfTypeFloatAt:(const float *)address count:(unsigned)cnt named:(NSString*)name
{
  if (!findingConditionals && name) {
    int i;
    NSMutableArray *array = [NSMutableArray arrayWithCapacity:cnt]; 

    for (i=0;i<cnt;i++) {
      [array addObject:[NSString stringWithFormat:@"%f", address[i]]];
    }

    [lastObjectRep setObject:array forKey:name];
  }
}

- (void)encodeArrayOfTypeIntAt:(const int *)address count:(unsigned)cnt named:(NSString*)name
{
  if (!findingConditionals && name) {
    int i;
    NSMutableArray *array = [NSMutableArray arrayWithCapacity:cnt]; 

    for (i=0;i<cnt;i++) {
      [array addObject:[NSString stringWithFormat:@"%d", address[i]]];
    }

    [lastObjectRep setObject:array forKey:name];
  }
}

- (void)encodeChar:(char)value named:(NSString*)name
{
  if (!findingConditionals && name) {
    NSString *valueString = [NSString stringWithFormat: @"%c", value];

    [lastObjectRep setObject:valueString forKey:name];
  }
}

- (void)encodeUnsignedChar:(unsigned char)value named:(NSString*)name
{
  if (!findingConditionals && name) {
    NSString  *valueString = [NSString stringWithFormat:@"%uc",value];

    [lastObjectRep setObject:valueString forKey:name];
  }
}

- (void)encodeBOOL:(BOOL)value named:(NSString*)name
{
  if (!findingConditionals && name) {
    NSString *valueString = [NSString stringWithFormat:@"%s",value?"YES":"NO"];

    [lastObjectRep setObject:valueString forKey:name];
  }
}

- (void)encodeShort:(short)value named:(NSString*)name
{
  if (!findingConditionals && name) {
    NSString *valueString = [NSString stringWithFormat:@"%s",value];

    [lastObjectRep setObject:valueString forKey:name];
  }
}

- (void)encodeUnsignedShort:(unsigned short)value named:(NSString*)name
{
  if (!findingConditionals && name) {
    NSString *valueString = [NSString stringWithFormat:@"%us",value];

    [lastObjectRep setObject:valueString forKey:name];
  }
}

- (void)encodeInt:(int)value named:(NSString*)name
{
  if (!findingConditionals && name) {
    NSString *valueString = [NSString stringWithFormat:@"%i",value];

    [lastObjectRep setObject:valueString forKey:name];
  }
}

- (void)encodeUnsignedInt:(unsigned int)value named:(NSString*)name
{
  if (!findingConditionals && name) {
    NSString *valueString = [NSString stringWithFormat:@"%u",value];

    [lastObjectRep setObject:valueString forKey:name];
  }
}

- (void)encodeLong:(long)value named:(NSString*)name
{
  if (!findingConditionals && name) {
    NSString *valueString = [NSString stringWithFormat:@"%l",value];

    [lastObjectRep setObject:valueString forKey:name];
  }
}

- (void)encodeUnsignedLong:(unsigned long)value named:(NSString*)name
{
  if (!findingConditionals && name) {
    NSString *valueString = [NSString stringWithFormat:@"%lu",value];

    [lastObjectRep setObject:valueString forKey:name];
  }
}

- (void)encodeFloat:(float)value named:(NSString*)name
{
  if (!findingConditionals && name) {
    NSString *valueString = [NSString stringWithFormat:@"%f",value];

    [lastObjectRep setObject:valueString forKey:name];
  }
}

- (void)encodeDouble:(double)value named:(NSString*)name
{
  if (!findingConditionals && name) {
    NSString *valueString = [NSString stringWithFormat:@"%f",value];

    [lastObjectRep setObject:valueString forKey:name];
  }
}

//-----------------------------------------------------------------------------
// Encoding other DTs
//-----------------------------------------------------------------------------


- (void)encodePoint:(NSPoint)point named:(NSString*)name
{
  if (!findingConditionals && name) {
    [lastObjectRep setObject:NSStringFromPoint(point) forKey:name];
  }
}

- (void)encodeSize:(NSSize)size named:(NSString*)name
{
  if (!findingConditionals && name) {
    [lastObjectRep setObject:NSStringFromSize(size) forKey:name];
  }
}

- (void)encodeRect:(NSRect)rect named:(NSString*)name
{
  if (!findingConditionals && name) {
    [lastObjectRep setObject:NSStringFromRect(rect) forKey:name];
  }
}

- (void)encodeFrustum:(G3DFrustumValues)fr named:(NSString*)name
{
  if (!findingConditionals && name) {
    [lastObjectRep setObject:G3DStringFromFrustum(fr) forKey:name];
  }
}

- (void)encodeCoordinate:(G3DCoordinate)coord named:(NSString*)name
{
  if (!findingConditionals && name) {
    [lastObjectRep setObject:G3DStringFromCoordinate(coord) forKey:name];
  }
}

- (void)encodeG3DRange:(G3DRange)range named:(NSString*)name
{
  if (!findingConditionals && name) {
    [lastObjectRep setObject:G3DStringFromRange(range) forKey:name];
  }
}

- (void)encodeBoundingSphere:(G3DBoundingSphere)bsph named:(NSString*)name
{
  if (!findingConditionals && name) {
    [lastObjectRep setObject:G3DStringFromBoundingSphere(bsph) forKey:name];
  }
}

- (void)encodeBoundingBox:(G3DBoundingBox)bbox named:(NSString*)name
{
  if (!findingConditionals && name) {
    [lastObjectRep setObject:G3DStringFromBoundingBox(bbox) forKey:name];
  }
}

@end

@implementation NSObject (SceneArchiving)

+ (id)createObjectForSceneUnarchiver:(G3DUnarchiver *)unarchiver
{
  return [[[self allocWithZone:[unarchiver objectZone]] init] autorelease];
}

- (id)replacementObjectForSceneArchiver:(G3DArchiver *)archiver
{
  return [self replacementObjectForCoder:nil];
}

- (Class)classForSceneArchiver
{
  return [self classForCoder];
}

@end





