//-----------------------------------------------------------------------------
// Project            	RenderKit
// Class		G3DBVol
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Sun Jun 25 10:35:50 CEST 2000
//
// Copyright (c) Philippe C.D. Robert
//
// The RenderKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DBVol.m,v 1.18 2001/07/07 15:35:47 probert Exp $
//-----------------------------------------------------------------------------

#import "G3DBVol.h"
#import "G3DVtxArray.h"
#import "G3DTransform.h"
#import "G3DGraphNode.h"
#import "G3DArchiver.h"
#import "G3DUnarchiver.h"

#import <GeometryKit/GeometryKit.h>

#define SL 10
#define ST 10

NSString *G3DBVolTypeException = @"G3DBVolTypeException";
NSString *G3DBVolException = @"G3DBVolException";

@implementation G3DBVol

void drawBox(float *min, float *max)
{
  float v[8][3];
  int i;
  static GLint faces[6][4] = {{0, 1, 2, 3},
			      {3, 2, 6, 7},
			      {7, 6, 5, 4},
			      {4, 5, 1, 0},
			      {5, 6, 2, 1},
			      {7, 4, 0, 3}};

  v[0][0] = v[1][0] = v[2][0] = v[3][0] = min[0];
  v[4][0] = v[5][0] = v[6][0] = v[7][0] = max[0];
  v[0][1] = v[1][1] = v[4][1] = v[5][1] = min[1];
  v[2][1] = v[3][1] = v[6][1] = v[7][1] = max[1];
  v[0][2] = v[3][2] = v[4][2] = v[7][2] = min[2];
  v[1][2] = v[2][2] = v[5][2] = v[6][2] = max[2];

  for (i = 5; i >= 0; i--) {
    glBegin(GL_LINE_LOOP);
    glVertex3fv(&v[faces[i][0]][0]);
    glVertex3fv(&v[faces[i][1]][0]);
    glVertex3fv(&v[faces[i][2]][0]);
    glVertex3fv(&v[faces[i][3]][0]);
    glEnd();
  }
}

//-----------------------------------------------------------------------------
// Init and free
//-----------------------------------------------------------------------------

- (id)initWithType:(G3D_BVolType)volType
{
  if ((self = [super init])) {
    // Volume zero indicates that self is not yet ready!
    volume = 0.0f; 
    volumeType = volType;

    _vObjIsDirty = YES;

    switch (volumeType) {
      case G3D_Sphere:
        volumeObject = [[G3DSphere alloc] init];
        transformedVolumeObject = [[G3DSphere alloc] init];
        break;
      case G3D_AABB:
        volumeObject = [[G3DBox alloc] init];
        transformedVolumeObject = [[G3DBox alloc] init];
        break;
      case G3D_OBB:
        [NSException raise:G3DBVolTypeException format:@"Oriented bounding boxes are not yet supported!"];
        break;
      default:
        [NSException raise:G3DBVolTypeException format:@"Unsupported volume type!"];
        break;
    }

    _trvoNeedsUpdate = YES;
    _tmpMat = [[G3DMatrix4f alloc] init];
  }
  return self;
}

- (id)initWithType:(G3D_BVolType)volType geometry:(id<G3DGeometryBackend>)geom
{
  if ((self = [self initWithType:volType])) {
    [self updateWithGeometry:geom];
  }
  return self;  
}

- (void)dealloc
{
  [volumeObject release];
  [transformedVolumeObject release];
  [_tmpMat release];
  
  [super dealloc];
}

//-----------------------------------------------------------------------------
// Updating
//-----------------------------------------------------------------------------

- (void)mergeWithBVolume:(G3DBVol *)bVol
{
  G3D_BVolType _o_type = [bVol type];
  id _o_vol = [bVol transformedVolumeObject];
  
  /*
   * Algorithm:
   *
   *	1st, transform both volumes into world coordinates
   *	2nd, merge volumes - if different, choose the existing type to survive!
   *	3rd, set the new volume, incl.the transformed!
   */

  if (volumeType == G3D_Sphere && _o_type == G3D_Sphere) {
    [(G3DSphere *)transformedVolumeObject wrapSphere:(G3DSphere *)_o_vol];
  }
  else if (volumeType == G3D_AABB && _o_type == G3D_AABB) {
    [(G3DBox *)transformedVolumeObject wrapBox:(G3DBox *)_o_vol];
  }
  else if (volumeType == G3D_Sphere && _o_type == G3D_AABB) {
    [(G3DSphere *)transformedVolumeObject wrapBox:(G3DBox *)_o_vol];
  }
  else if (volumeType == G3D_AABB && _o_type == G3D_Sphere) {
    [(G3DBox *)transformedVolumeObject wrapSphere:(G3DSphere *)_o_vol];
  }
  else {
    [NSException raise:G3DBVolTypeException format:@"Oriented bounding boxes are not yet supported!"];
    return;
  }

  [volumeObject autorelease];
  volumeObject = [(NSObject *)transformedVolumeObject copy];

  _tmpMat = [[node modelTransformation] invertedMatrix];
  [volumeObject transform:_tmpMat];

  DLOG(@"Merged %@ %x (type %d) with %@ %x (type %d) [name %@, class %@].",[self class],self,volumeType,[bVol class],bVol,[bVol type],[[self node] name],[[self node] class]);
  
  volume = [volumeObject volume];  
}

- (void)updateWithGeometry:(id<G3DGeometryBackend>)geom
{
  float min[4];
  float max[4];

  DLOG(@"<%@ %x>: updating bounding volume using %@ %x (%@) [name %@, class %@]",[self class],self,[geom class],geom,(volumeType == G3D_Sphere) ? @"Sphere" : @"AABB",[[self node] name],[[self node] class]);

  // This takes place in the local object coordinate space!
  [geom getMin:&min[0] max:&max[0]];

  if (volumeType == G3D_Sphere) {
    // Slow -> direct use of C API...
    [[(G3DSphere *)volumeObject center] setElements:min];
    [(G3DSphere *)volumeObject setRadius:0.001f];    
    [(G3DSphere *)volumeObject wrapPoint:[G3DTuple3f tuple3fWithElements:max]];
  }
  else if (volumeType == G3D_AABB) {
    [[(G3DBox *)volumeObject min] setElements:min];
    [[(G3DBox *)volumeObject max] setElements:max];
  }
  else if (volumeType == G3D_OBB) {
    [NSException raise:G3DBVolTypeException format:@"Oriented bounding boxes are not yet supported!"];
    return;
  }

  // For caching purposes...
  if (node) {
    G3DMatrix4f *tr = [node modelTransformation];
    [transformedVolumeObject autorelease];
    transformedVolumeObject = [(NSObject *)volumeObject copy];

    // Updates the transformed object, which is needed for view-frustum culling
    [transformedVolumeObject transform:tr];

    _trvoNeedsUpdate = NO;
  }
  else {
    _trvoNeedsUpdate = YES;
  }

  // Cache the colume
  volume = [volumeObject volume];  
}

//-----------------------------------------------------------------------------
// Accessor methods
//-----------------------------------------------------------------------------

- (G3D_BVolType)type
{
  return volumeType;
}

- (void)setNode:(G3DGraphNode *)nd;
{
  node = nd;
}

- (G3DGraphNode *)node
{
  return node;
}

- (id<G3DIntersecting>)volumeObject
{
  return volumeObject; 
}

- (id<G3DIntersecting>)transformedVolumeObject
{
  if (_trvoNeedsUpdate) {
    if (node) {
      G3DMatrix4f *tr = [node modelTransformation];

      [transformedVolumeObject autorelease];
      transformedVolumeObject = [(NSObject *)volumeObject copy];
      
      [transformedVolumeObject transform:tr];
      
      _trvoNeedsUpdate = NO;
    }
    else {
      NSLog(@"*** No node given!!");
      return nil;
    }
  }
  return transformedVolumeObject; 
}

//-----------------------------------------------------------------------------
// Enclosing
//-----------------------------------------------------------------------------

- (float)volume
{
  return volume;
}

- (BOOL)intersectsWithVolume:(G3DBVol *)bVol
{
  id<G3DIntersecting> _o_vol = [bVol transformedVolumeObject];

  return [transformedVolumeObject intersectsObject:_o_vol];
}

//-----------------------------------------------------------------------------
// NSCoding
//-----------------------------------------------------------------------------

- (void)encodeWithCoder:(NSCoder *)aCoder
{
  [super encodeWithCoder:aCoder];

  [aCoder encodeObject:volumeObject];
  [aCoder encodeObject:transformedVolumeObject];
  [aCoder encodeValueOfObjCType:@encode(float) at:&volume];
  [aCoder encodeValueOfObjCType:@encode(G3D_BVolType) at:&volumeType];
  [aCoder encodeObject:node];
}

- (id)initWithCoder:(NSCoder *)aCoder
{
  self = [super initWithCoder:aCoder];

  volumeObject = [[aCoder decodeObject] retain];
  transformedVolumeObject = [[aCoder decodeObject] retain];
  [aCoder decodeValueOfObjCType:@encode(float) at:&volume];
  [aCoder decodeValueOfObjCType:@encode(G3D_BVolType) at:&volumeType];
  node = [aCoder decodeObject]; // Correct?

  _vObjIsDirty = YES;
  _trvoNeedsUpdate = YES;

  return self;
}

//-----------------------------------------------------------------------------
// NSCopying
//-----------------------------------------------------------------------------

- (id)copyWithZone:(NSZone *)zone
{
  G3DBVol *copy = [super copyWithZone:zone];

  /*
   * Deep copy!
   */

  copy->volumeObject = [(NSObject *)volumeObject copy];
  copy->transformedVolumeObject = [(NSObject *)transformedVolumeObject copy];
  copy->volume = [self volume];
  copy->volumeType = volumeType;

  copy->_vObjIsDirty = YES;
  copy->_trvoNeedsUpdate = _trvoNeedsUpdate;
 
  [copy setNode:node];

  return copy;
}

//-----------------------------------------------------------------------------
// G3DSceneCoding
//-----------------------------------------------------------------------------

- (void)encodeWithSceneArchiver:(G3DArchiver *)archiver
{
  [super encodeWithSceneArchiver:archiver];

  [archiver encodeObject:volumeObject named:@"volumeObject"];
  [archiver encodeObject:transformedVolumeObject named:@"transformedVolumeObject"];

  [archiver encodeFloat:volume named:@"volume"];
  [archiver encodeInt:volumeType named:@"volumeType"];

  [archiver encodeObject:node named:@"node"];
}

- (id)initWithSceneUnarchiver:(G3DUnarchiver *)unarchiver
{
  self = [super initWithSceneUnarchiver:unarchiver];
  
  volumeObject = [[unarchiver decodeObjectNamed:@"volumeObject"] retain];
  transformedVolumeObject = [[unarchiver decodeObjectNamed:@"transformedVolumeObject"] retain];

  [unarchiver decodeFloat:&volume named:@"volume"];
  [unarchiver decodeInt:&volumeType named:@"volumeType"];

  node = [unarchiver decodeObjectNamed:@"node"]; // Correct?

  _vObjIsDirty = YES;
  _trvoNeedsUpdate = YES;

  return self;
}

@end

@implementation G3DBVol (RenderingExtensions)
     
- (void)updateVisObj
{
  /*
  if ((q = gluNewQuadric()) == NULL) {
    [NSException raise:G3DBVolException format:@"Quadric generation failed!"];
    return;
  }
  
  gluQuadricDrawStyle(q,GLU_LINE);
  gluQuadricNormals(q,GLU_NONE);
  gluQuadricTexture(q,GL_FALSE);
  */

  _vObjIsDirty = NO;
}

- (void)renderBoundingVolume
{
  G3DTuple3f *min;
  G3DTuple3f *max;

  if (_vObjIsDirty) {
    [self updateVisObj];
  }

  switch (volumeType) {
  case G3D_Sphere:
    //gluSphere(q,[(G3DSphere *)volumeObject radius],SL,ST);
    break;
  case G3D_AABB:
    min = [G3DTuple3f tuple3f];
    max = [G3DTuple3f tuple3f];

    [(G3DBox *)volumeObject getMin:&min max:&max];
    drawBox([min elements],[max elements]);
    break;
  case G3D_OBB:
    break;
  default:
    break;
  }
}

@end





