//-----------------------------------------------------------------------------
// Project            	RenderKit
// Class		G3DCamera
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Thu Sep  9 11:05:08 CEST 1999
//
// Copyright (c) Philippe C.D. Robert
//
// The RenderKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DCamera.m,v 1.17 2001/07/22 11:04:21 probert Exp $
//-----------------------------------------------------------------------------

#import "G3DCamera.h"
#import <RenderKit/RenderKit.h>

#include <time.h>
#define FRAME 25

static float gettime(void)
{
  static clock_t told=0;
  clock_t tnew;
  clock_t ris;

  tnew=clock();
  ris=tnew-told;
  told=tnew;

  return(ris/(float)CLOCKS_PER_SEC);
}

#define KNOWS_SEL(x) (renderDelegate && [renderDelegate respondsToSelector:@selector(x)])

NSString *G3DCameraSetupException = @"G3DCameraSetupException";

@implementation G3DCamera

static G3DCamera *activeCamera = nil;
static BOOL _measureFramerate = NO;

+ (void)initialize
{
  static BOOL tooLate = NO;

  [super initialize];

  if (!tooLate) {
    tooLate = YES;
  }
}

//-----------------------------------------------------------------------------
// Context handling
//-----------------------------------------------------------------------------

+ (G3DCamera *)activeCamera
{
    return activeCamera;
}

+ (void)setActiveCamera:(G3DCamera *)aCamera
{
  if (aCamera != activeCamera) {
    DLOG(@"%@ %x in thread %x becomes the current context.",[aCamera class],aCamera,[NSThread currentThread]);
    activeCamera = aCamera;
  }
}

+ (void)measureFramerates:(BOOL)yn
{
  _measureFramerate = yn;
}

//-----------------------------------------------------------------------------
// Init and free
//-----------------------------------------------------------------------------

- (id)initWithSize:(NSSize)aSize
{
  if ((self = [super init])) {
    _size = aSize;

    frustum = [[G3DFrustum alloc] init];
    state = [[G3DState alloc] init];

    G3DCopyMatrix4fv(position.euler,G3DIdentityMatrix4f);
    G3DCopyMatrix4fv(position.ieuler,G3DIdentityMatrix4f);

    position.pos[0] = 0.0f;
    position.pos[1] = 0.0f;
    position.pos[2] = 0.0f;
    
    position.quat[0] = 0.0f;
    position.quat[1] = 0.0f;
    position.quat[2] = 0.0f;
    position.quat[3] = 1.0f;

    _rdist[3] = 0.0f;

    // Some defaults
    cameraFlags.doFlush = NO;
    cameraFlags.degenerate = (_size.width == 0 || _size.height == 0) ? YES : NO;
    cameraFlags.basisChanged = NO;
    cameraFlags.canRender = NO;
    cameraFlags.usePreTM = NO;
    cameraFlags.drawBackground = YES;
    cameraFlags.isFullscreen = NO;
    cameraFlags.doDepthTest = YES;

    backgroundColour = [[G3DColour alloc] initWithR:1.0f g:1.0f b:1.0f a:1.0f];

    _hasDelegate = NO;
    _useTwoSideModel = YES;

    _rMask = 0;
  }
  return self;
}

- (void)dealloc
{
  [preTransformation release];  
  [backgroundColour release];
  [world release];

  [frustum release];
  [state release];

  if (fog) {
    [fog release];
  }

  [super dealloc];
}

//-----------------------------------------------------------------------------
// Delegate
//-----------------------------------------------------------------------------

- (id)renderDelegate
{
  return renderDelegate;
}

- (void)setRenderDelegate:(id)aDelegate
{
  renderDelegate = aDelegate;
  _hasDelegate = YES;

  _notifyBefore = (KNOWS_SEL(cameraWillRenderFrame:)) ? YES : NO;
  _notifyAfter = (KNOWS_SEL(cameraDidRenderFrame:)) ? YES : NO;
}

//-----------------------------------------------------------------------------
// Light information
//-----------------------------------------------------------------------------

- (void)setEnableLighting:(BOOL)flag
{
  if (yn) {
    _rMask |= (1<<ENABLE_GL_LIGHTS);
  }
  else {
    _rMask &= ~(1<<ENABLE_GL_LIGHTS);
  }
}

- (BOOL)enableLighting
{
  return _rMask & (1<<ENABLE_GL_LIGHTS);
}

//-----------------------------------------------------------------------------
// Rendering information
//-----------------------------------------------------------------------------

- (void)setCullingEnabled:(BOOL)yn
{
  if (yn) {
    _rMask |= (1<<VIEW_FRUSTUM_CULLING);
  }
  else {
    _rMask &= ~(1<<VIEW_FRUSTUM_CULLING);
  }
}

- (BOOL)cullingEnabled
{
  return _rMask & (1<<VIEW_FRUSTUM_CULLING);
}

- (void)setUseTwoSideModel:(BOOL)flag
{
  _useTwoSideModel = flag;
}

- (BOOL)useTwoSideModel
{
  return _useTwoSideModel;
}

- (G3DFrustum *)frustum
{
  return frustum;
}

- (G3DState *)state
{
  return state;
}

//-----------------------------------------------------------------------------
// Background drawing
//-----------------------------------------------------------------------------

- (void)setFog:(G3DFog *)f
{
  [fog autorelease];
  fog = [f retain];
}

- (G3DFog *)fog
{
  return fog;
}

- (void)setBackgroundColour:(G3DColour *)bg
{
  [backgroundColour autorelease];
  backgroundColour = [bg retain];
}

- (G3DColour *)backgroundColour
{
  return backgroundColour;
}

- (BOOL)doesDrawBackgroundColour
{
  return cameraFlags.drawBackground;
}

- (void)setDrawBackgroundColour:(BOOL)yn
{
  cameraFlags.drawBackground = yn;
}

//-----------------------------------------------------------------------------
// Accessor methods
//-----------------------------------------------------------------------------

- (void)setWorld:(G3DWorld *)aWorld
{
  [world autorelease];
  world = [aWorld retain];
}

- (G3DWorld *)world
{
  return world;
}

- (void)setViewportWithSize:(NSSize)s
{
  float width = s.width;
  float height = s.height;
  G3DFrustumValues _fr_vals = [frustum frustumValues];

  _size = s;
  cameraFlags.degenerate = (width <= 0 || height <= 0) ? YES : NO;

  DLOG(@"Viewport with size w:%f h:%f and frustum %@",
       width,height,G3DStringFromFrustum(_fr_vals));

  [G3DCamera setActiveCamera:self];

  /*
   * Frustum
   *
   */

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glFrustum(_fr_vals.left,
	    _fr_vals.right,
	    _fr_vals.bottom,
	    _fr_vals.top,
	    _fr_vals.near,
	    _fr_vals.far);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  /*
   * Viewport
   *
   */

  glViewport(0,0,width,height);

  if (renderDelegate && 
      [renderDelegate respondsToSelector:@selector(cameraViewportDidChange:)]){
    [renderDelegate cameraViewportDidChange:self];
  }
}

- (NSSize)size
{
  return _size;
}

- (G3D_CameraPosition *)position
{
  return &position;
}

//-----------------------------------------------------------------------------
// Drawing
//-----------------------------------------------------------------------------

- (void)setupCamera
{
  /*
  if ([self window] == nil) {
    [NSException raise:G3DCameraSetupException 
		 format:@"The camera is not yet attached to a window!"];
  }
  */

  [self setViewportWithSize:_size];

  glClearColor([backgroundColour red],
	       [backgroundColour green],
	       [backgroundColour blue],
	       [backgroundColour alpha]);

  // Z Buffering
  if (cameraFlags.doDepthTest) {
    glEnable(GL_DEPTH_TEST);
  }

  [[G3DRenderEngine renderEngine] initEngine];

  cameraFlags.canRender = YES;

  // Notify the renderDelegate
  if (renderDelegate && 
      [renderDelegate respondsToSelector:@selector(cameraDidInitialise:)]) {
    [renderDelegate cameraDidInitialise:self];
  }
}

- (void)preprocessScene
{
  if (renderDelegate && 
      [renderDelegate respondsToSelector:@selector(cameraWillPreprocessScene:)]) {
      [renderDelegate cameraWillPreprocessScene:self];
    }
  
  [G3DCamera setActiveCamera:self];
  [world preprocess];
  
  if (renderDelegate && 
      [renderDelegate respondsToSelector:@selector(cameraDidPreprocessScene:)]){
    [renderDelegate cameraDidPreprocessScene:self];
  }
}

- (void)render
{
  static int count=0;

  if (!cameraFlags.canRender || cameraFlags.degenerate) {
    return;
  }
  
  if (_measureFramerate) {
    if((count % FRAME)==0) {
      float fr=gettime();
      NSLog(@"Frame rate: %f",FRAME/fr);
    }
  }

  if (_notifyBefore) {
    [renderDelegate cameraWillRenderFrame:self];
  }
  
  // Start rendering!
  [G3DCamera setActiveCamera:self];
  
  G3D_PUSH_MATRIX;

  glClearColor([backgroundColour red],
	       [backgroundColour green],
	       [backgroundColour blue],
	       [backgroundColour alpha]);
  
  if (cameraFlags.drawBackground) {
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  }
  else {
    glClear(GL_DEPTH_BUFFER_BIT);
  }

  // -> current render state?
  if (!_useTwoSideModel) {
    glEnable(GL_CULL_FACE);
    glCullFace(GL_BACK);
    glFrontFace(GL_CCW);
  }

  /*
   * Setting the camera's view matrix
   *
   */

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  glMultMatrixf(position.ieuler);
  glTranslatef(-position.pos[0],-position.pos[1],-position.pos[2]);

  /*
   * Use the pretransformation if there is one
   *
   */

  if (cameraFlags.usePreTM && preTransformation != nil) {
    [preTransformation transform];
  }

  /*
   * Finally render the scenegraph
   *
   */

  [world renderWithMask:_rMask frustum:frustum];

  if (_notifyAfter) {
    [renderDelegate cameraDidRenderFrame:self];
  }

  G3D_POP_MATRIX;

  count++;
}

//-----------------------------------------------------------------------------
// Transformation
//-----------------------------------------------------------------------------

- (void)setPreTransformation:(G3DTransform *)tf
{
  [preTransformation autorelease];
  preTransformation = [tf retain];
}

- (G3DTransform *)preTransformation
{
  return preTransformation;
}

- (void)setUsePreTransformation:(BOOL)yn
{
  cameraFlags.usePreTM = yn;
}

- (BOOL)usePreTransformation
{
  return cameraFlags.usePreTM;
}

//-----------------------------------------------------------------------------
// Camera Movements
//-----------------------------------------------------------------------------

- (void)moveX:(const float)x
{
  [self moveX:x y:0.0f z:0.0f];
}

- (void)moveY:(const float)y
{
  [self moveX:0.0f y:y z:0.0f];
}

- (void)moveZ:(const float)z
{
  [self moveX:0.0f y:0.0f z:z];
}

- (void)moveX:(const float)x y:(const float)y z:(const float)z
{
  float d[4];

  _rdist[0] = x;
  _rdist[1] = y;
  _rdist[2] = z;

  G3DMatrix4fXVector4f(d,position.euler,_rdist);

  position.pos[0] += d[0];
  position.pos[1] += d[1];
  position.pos[2] += d[2];

  if (_rMask & (1 << VIEW_FRUSTUM_CULLING)) {
    /* Culling */
    [frustum transformWithRotation:&position.euler[0] position:&position.pos[0]];
  }
  else {
    /* Not culling */
  }
}

- (void)rotateX:(const float)angle
{
  static float axis[4] = {1.0f,0.0f,0.0f,0.0f};
  float res[4];
  float q[4];
  float new[4];

  G3DMatrix4fXVector4f(res,position.euler,axis);
  G3DQuatFromAngleAxisf(q,angle,res);

  G3DMultiplyQuatf(new,q,position.quat);
  G3DCopyVector4fv(position.quat,new);

  G3DMatrixFromQuatf(position.euler,new);
  G3DTransposeMatrix4fv(position.ieuler,position.euler);

  if (_rMask & (1 << VIEW_FRUSTUM_CULLING)) {
    /* Culling */
    [frustum transformWithRotation:&position.euler[0] position:&position.pos[0]];
  }
  else {
    /* Not culling */
  }
}

- (void)rotateY:(const float)angle
{
  static float axis[4] = {0.0f,1.0f,0.0f,0.0f};
  float res[4];
  float q[4];
  float new[4];

  G3DMatrix4fXVector4f(res,position.euler,axis);
  G3DQuatFromAngleAxisf(q,angle,res);

  G3DMultiplyQuatf(new,q,position.quat);
  G3DCopyVector4fv(position.quat,new);

  G3DMatrixFromQuatf(position.euler,new);
  G3DTransposeMatrix4fv(position.ieuler,position.euler);

  if (_rMask & (1 << VIEW_FRUSTUM_CULLING)) {
    /* Culling */
    [frustum transformWithRotation:&position.euler[0] position:&position.pos[0]];
  }
  else {
    /* Not culling */
  }
}

- (void)rotateZ:(const float)angle
{
  static float axis[4] = {0.0f,0.0f,1.0f,0.0f};
  float res[4];
  float q[4];
  float new[4];

  G3DMatrix4fXVector4f(res,position.euler,axis);
  G3DQuatFromAngleAxisf(q,angle,res);

  G3DMultiplyQuatf(new,q,position.quat);
  G3DCopyVector4fv(position.quat,new);

  G3DMatrixFromQuatf(position.euler,new);
  G3DTransposeMatrix4fv(position.ieuler,position.euler);

  if (_rMask & (1 << VIEW_FRUSTUM_CULLING)) {
    /* Culling */
    [ frustum transformWithRotation:&position.euler[0] position:&position.pos[0]];
  }
  else {
    /* Not culling */
  }
}

- (void)rotateX:(const float)xA y:(const float)yA z:(const float)zA
{
}

- (void)rotateByAngleAxis:(float *)vec4
{
  G3DQuatFromAngleAxisf(position.quat, vec4[0], &vec4[1]);

  G3DMatrixFromQuatf(position.euler,position.quat);
  G3DTransposeMatrix4fv(position.ieuler,position.euler);
  
  if (_rMask & (1 << VIEW_FRUSTUM_CULLING)) {
    /* Culling */
    [frustum transformWithRotation:&position.euler[0] position:&position.pos[0]];
  }
  else {
    /* Not culling */
  }
}

- (void)rotateByQuaternion:(G3DQuaternion *)quat
{
  float *q = [quat elements];
  float new[4];

  G3DMultiplyQuatf(new,q,position.quat);
  G3DCopyVector4fv(position.quat,new);

  G3DMatrixFromQuatf(position.euler,new);
  G3DTransposeMatrix4fv(position.ieuler,position.euler);
  
  if (_rMask & (1 << VIEW_FRUSTUM_CULLING)) {
    /* Culling */
    [frustum transformWithRotation:&position.euler[0] position:&position.pos[0]];
  }
  else {
    /* Not culling */
  }
}

- (void)setCamera:(const float *)pos toward:(const float *)to up:(const float *)up
{
  float x[3], y[3], z[3];
  float _dt[3];
  float _xt[3], _yt[3];

  _dt[0] = pos[0] - to[0];
  _dt[1] = pos[1] - to[1];
  _dt[2] = pos[2] - to[2];
  G3DNormalise3fv(z, _dt);

  y[0] = up[0];
  y[1] = up[1];
  y[2] = up[2];
  
  /* x = y cross z */
  G3DMultiplyVector3fv(_xt,y,z);
  
  /* y = z cross x */
  G3DMultiplyVector3fv(_yt,z,_xt);

  /* normalise x and y, is this needed? */
  G3DNormalise3fv(x, _xt);
  G3DNormalise3fv(y, _yt);
  
  /* #define taken from Mesa */
#define M(row,col)  position.ieuler[col*4+row]
  M(0,0) = x[0];  M(0,1) = x[1];  M(0,2) = x[2];  M(0,3) = 0.0;
  M(1,0) = y[0];  M(1,1) = y[1];  M(1,2) = y[2];  M(1,3) = 0.0;
  M(2,0) = z[0];  M(2,1) = z[1];  M(2,2) = z[2];  M(2,3) = 0.0;
  M(3,0) = 0.0;   M(3,1) = 0.0;   M(3,2) = 0.0;   M(3,3) = 1.0;
#undef M

  G3DTransposeMatrix4fv(position.euler,position.ieuler);
  G3DQuatFromMatrixf(position.quat,position.ieuler);
  
  /* Tranlation into origin */

  position.pos[0] = pos[0];
  position.pos[1] = pos[1];
  position.pos[2] = pos[2];

  if (_rMask & (1 << VIEW_FRUSTUM_CULLING)) {
    /* Culling */
    [frustum transformWithRotation:&position.euler[0] position:&position.pos[0]];
  }
  else {
    /* Not culling */
  }
}

@end











