//-----------------------------------------------------------------------------
// Project            	RenderKit
// Class		G3DFrustum
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Thu Sep  9 11:05:08 CEST 1999
//
// Copyright (c) Philippe C.D. Robert
//
// The RenderKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DFrustum.m,v 1.25 2001/01/05 20:50:51 robert Exp $
//-----------------------------------------------------------------------------

#import "G3DFrustum.h"
#import "G3DBVol.h"

#import <GeometryKit/GeometryKit.h>

NSString *G3DFrustumException = @"G3DFrustumException";

static BOOL cullPlaneAABB(float* plane,float* m1, float* m2)
{
  float v_min[3];
  float v_max[3];

  /*
   * First, find the diagonal of the AABB that fits most the normal 
   * of the plane. Then insert the 2 vertices of that diagonal into the
   * plane equation, if the signs of the results differ or one is 0.0 then
   * the plane intersects the box. Using loop unrolling is perhaps faster...
   *
   */

  // Parameter a
  if (plane[0] >= 0.0f) {
    v_min[0] = m1[0];
    v_max[0] = m2[0];
  }
  else {
    v_min[0] = m2[0];
    v_max[0] = m1[0];
  }

  // Parameter b
  if (plane[1] >= 0.0f) {
    v_min[1] = m1[1];
    v_max[1] = m2[1];
  }
  else {
    v_min[1] = m2[1];
    v_max[1] = m1[1];
  }

  // Parameter c
  if (plane[2] >= 0.0f) {
    v_min[2] = m1[2];
    v_max[2] = m2[2];
  }
  else {
    v_min[2] = m2[2];
    v_max[2] = m1[2];
  }

  // Overlap!
  if (G3DDistanceFromPlanef(v_max,plane) >= 0.0f) {
    return NO;  
  }

  return YES;
}

@implementation G3DFrustum

//-----------------------------------------------------------------------------
// init and free
//-----------------------------------------------------------------------------

- (id)init
{
  return [self initWithLeft:-1.0 
	       right:1.0 
	       top:1.0 
	       bottom:-1.0 
	       near:1.0 
	       far:100.0];
}

- (id)initWithLeft:(float)l right:(float)r top:(float)t bottom:(float)b near:(float)n far:(float)f
{
  if ((self = [super init])) {
    frustum.near = n;
    frustum.far = f;
    frustum.top = t;
    frustum.bottom = b;
    frustum.left = l;
    frustum.right = r;

    frustum.hfov = 0.0;
    frustum.vfov = 0.0;

    [self update];
  }
  return self;
}

//-----------------------------------------------------------------------------
// Frustum handling
//-----------------------------------------------------------------------------

- (void)update
{
  float tv1[3];
  float tv2[3];
  float tv3[3];
  float tv4[3];
  float ntv1[3];
  float ntv2[3];
  float ntv3[3];
  float ntv4[3];

  float r1;
  float r2;
  float _r;
  float _l;

  if (ABS(frustum.far - frustum.near) < 0.1 ) {
    [NSException raise:G3DFrustumException format:@"Invalid frustum!"];
    return;
  }
  
  if (frustum.hfov != 0.0 && frustum.vfov != 0.0) {
    if (ABS(frustum.hfov) < 0.1 || ABS(frustum.vfov) < 0.1) {
      [NSException raise:G3DFrustumException format:@"Invalid frustum!"];
      return ;
    }

    frustum.right = frustum.near * (float)tan(frustum.hfov * DEG2RAD/2.0);
    frustum.top   = frustum.near * (float)tan(frustum.vfov * DEG2RAD/2.0);
    frustum.left  = -frustum.right;
    frustum.bottom   = -frustum.top;
  }

  /*
   * The 4 vectors (topleft, top/right, bottom/left, bottom/right)
   */

  G3DCopyVector3f(tv1,frustum.left,frustum.top,-frustum.near);
  G3DNormalise3fv(ntv1,tv1);

  G3DCopyVector3f(tv2,frustum.right,frustum.top,-frustum.near);
  G3DNormalise3fv(ntv2,tv2);

  G3DCopyVector3f(tv3,frustum.left,frustum.bottom,-frustum.near);
  G3DNormalise3fv(ntv3,tv3);

  G3DCopyVector3f(tv4,frustum.right,frustum.bottom,-frustum.near);
  G3DNormalise3fv(ntv4,tv4);

  /*
   * The frustum planes (facing inwards!)
   */

  G3DVectorProduct3fv(left,ntv3,ntv1);
  left[3] = 0.0f;
  G3DNormalise3fv(_tr_left,left);

  G3DVectorProduct3fv(right,ntv2,ntv4);
  right[3] = 0.0f;
  G3DNormalise3fv(_tr_right,right);

  G3DVectorProduct3fv(top,ntv1,ntv2);
  top[3] = 0.0f;
  G3DNormalise3fv(_tr_top,top);

  G3DVectorProduct3fv(bottom,ntv4,ntv3);
  bottom[3] = 0.0f;
  G3DNormalise3fv(_tr_bottom,bottom);

  /*
   * Calculating the max radius for a spherical like frustum. This is needed
   * for fast frustum culling.
   */

  _r = frustum.far * (frustum.right/frustum.near);
  _l = frustum.far * (frustum.left/frustum.near);

  r1 = sqrt(SQR(frustum.far) + SQR(_r));
  r2 = sqrt(SQR(frustum.far) + SQR(_l));

  _maxRad = MAX(r1,r2);
}

- (void)transformWithRotation:(float *)mat position:(float *)pos
{
  float tmp[4];

  /*
   * rotation should not alter the normalised state of a vector, but it seems
   * that normalisation actually results in slightly different values?!
   *
   */

  G3DMatrix4fXVector3f(tmp,mat,left);
  G3DNormalise3fv(_tr_left,tmp);

  G3DMatrix4fXVector3f(tmp,mat,right);
  G3DNormalise3fv(_tr_right,tmp);

  G3DMatrix4fXVector3f(tmp,mat,top);
  G3DNormalise3fv(_tr_top,tmp);

  G3DMatrix4fXVector3f(tmp,mat,bottom);
  G3DNormalise3fv(_tr_bottom,tmp);

  /*
   * Now set the proper 'd' values
   */

  _tr_left[3]   = -(_tr_left[0]*pos[0]+
		    _tr_left[1]*pos[1]+
		    _tr_left[2]*pos[2]);

  _tr_right[3]  = -(_tr_right[0]*pos[0]+
		    _tr_right[1]*pos[1]+
		    _tr_right[2]*pos[2]);

  _tr_top[3]    = -(_tr_top[0]*pos[0] + 
		    _tr_top[1]*pos[1] + 
		    _tr_top[2]*pos[2]);

  _tr_bottom[3] = -(_tr_bottom[0]*pos[0]+
		    _tr_bottom[1]*pos[1]+
		    _tr_bottom[2]*pos[2]);

  // Cache teh current camera position - needed for culling
  G3DCopyVector3fv(_currentPos,pos);
}

//-----------------------------------------------------------------------------
// Accessor methods
//-----------------------------------------------------------------------------

- (void)setHFOV:(float)hAngle vFOV:(float)vAngle
{
  NSAssert(ABS(hAngle) > 0.1,@"HFOV cannot be equal to 0.0");
  NSAssert(ABS(vAngle) > 0.1,@"VFOV cannot be equal to 0.0");

  frustum.hfov = hAngle;
  frustum.vfov = vAngle;

  [self update];
}

- (void)setNear:(float)n far:(float)f
{
  frustum.near = n;
  frustum.far = f;

  [self update];
}

- (void)setFrustumValues:(G3DFrustumValues)aFrustum
{
  frustum.near = aFrustum.near;
  frustum.far = aFrustum.far;
  frustum.top = aFrustum.top;
  frustum.bottom = aFrustum.bottom;
  frustum.left = aFrustum.left;
  frustum.right = aFrustum.right;

  frustum.hfov = 0.0;
  frustum.vfov = 0.0;
 
  [self update];
}

- (G3DFrustumValues)frustumValues
{
  return frustum;
}

- (void)setProjectionType:(G3DProjectionType)tp
{
  projectionType = tp;
}

- (G3DProjectionType)projectionType
{
  return projectionType;
}

- (G3DMatrix4f *)projectionMatrix
{
  G3DMatrix4f *matrix;
  float rl;
  float tb;
  float fn;

  matrix = [[G3DMatrix4d alloc] init]; // This gives us an identity matrix

  rl = frustum.right - frustum.left;
  tb = frustum.top - frustum.bottom;
  fn = frustum.far - frustum.near;

  /*
   * The projection matrix as defined by OpenGL...
   */

  [matrix setElement:(2.0 * frustum.near/rl) atRow:0 column:0];
  [matrix setElement:(2.0 * frustum.near/tb) atRow:1 column:1];
  [matrix setElement:((frustum.right + frustum.left)/rl) atRow:0 column:2];
  [matrix setElement:((frustum.top + frustum.bottom)/rl) atRow:1 column:2];
  [matrix setElement:(-(frustum.far + frustum.near)/fn) atRow:2 column:2];
  [matrix setElement:-1 atRow:3 column:2];
  [matrix setElement:(-2*(frustum.far * frustum.near)/fn) atRow:2 column:3];

  return [matrix autorelease];
}

//-----------------------------------------------------------------------------
// Testing
//-----------------------------------------------------------------------------

- (G3DLocation)containsVolume:(G3DBVol *)vol
{
  G3DLocation ret = G3DInside;
  G3D_BVolType type = [vol type];
  id _vol = [vol transformedVolumeObject];

  if (!_vol) {
    DLOG(@"*** %@ %x has no transformed bounding volume!?",[vol class],vol);
    return G3DOutside;
  }

  switch (type) {
    case G3D_Sphere:
      ret = [self containsSphere:_vol];
      break;
    case G3D_AABB:
      ret = [self containsAABB:_vol];
      break;
    case G3D_OBB:
      ret = [self containsOBB:_vol];
      break;
    default:
      break;
  }

  return ret;
}

- (G3DLocation)containsSphere:(G3DSphere *)sph
{
  float *c = sph->center;
  float radius = [sph radius];
  float dist_sph_pl[4];

  if (projectionType == G3DOrthographic) {
    // For the future!
    DLOG(@"%@ %x is orthograpic - using box/sphere test!",[self class],self);
    return G3DInside;
  }

  /*
   * Idea: Test if the center's distance to each plane is greater
   * than the sphere's radius. If yes, the sphere lies outside
   * the frustum.
   *
   */

  DLOG(@"Testing sphere at %f %f %f (%f)",c[0],c[1],c[2],radius);

  if (G3DDistance3fv(c,_currentPos) > _maxRad) {
    DLOG(@"%@ %x is outside of sphere with radius %f",[sph class],sph,_maxRad);
    return G3DOutside;
  }

  // Now the 4 other planes...
  dist_sph_pl[0] = G3DDistanceFromPlanef(c,_tr_left);
  DLOG(@"d1 to %@ is %f",G3DStringFromVectorf(_tr_left,4),dist_sph_pl[0]);
  if (-dist_sph_pl[0] > radius) {
    DLOG(@"%@ %x lies outside of %@ %x (left)",[sph class],sph,[self class],self);
    return G3DOutside;
  }
 
  dist_sph_pl[1] = G3DDistanceFromPlanef(c,_tr_right);
  DLOG(@"d2 to %@ is %f",G3DStringFromVectorf(_tr_right,4),dist_sph_pl[1]);
  if (-dist_sph_pl[1] > radius) {
    DLOG(@"%@ %x lies outside of %@ %x (right)",[sph class],sph,[self class],self);
    return G3DOutside;
  }

  dist_sph_pl[2] = G3DDistanceFromPlanef(c,_tr_top);
  DLOG(@"d3 to %@ is %f",G3DStringFromVectorf(_tr_top,4),dist_sph_pl[2]);
  if (-dist_sph_pl[2] > radius) {
    DLOG(@"%@ %x lies outside of %@ %x (top)",[sph class],sph,[self class],self);
    return G3DOutside;
  }

  dist_sph_pl[3] = G3DDistanceFromPlanef(c,_tr_bottom);
  DLOG(@"d4 to %@ is %f",G3DStringFromVectorf(_tr_bottom,4),dist_sph_pl[3]);
  if (-dist_sph_pl[3] > radius) {
    DLOG(@"%@ %x lies outside of %@ %x (bottom)",[sph class],sph,[self class],self);
    return G3DOutside;
  }

  /*
   * We do not test against complete or inbetween, since in both cases we would
   * have to render the node anyway!
   */

  return G3DInside;
}

- (G3DLocation)containsAABB:(G3DBox *)box
{
  G3DLocation r;
  float *min;
  float *max;
  float minz;
  float maxz;

  if (projectionType == G3DOrthographic) {
    // For the future!
    DLOG(@"%@ %x is orthograpic - using box/box test!",[self class],self);
  }

  return G3DInside;

  min = [[box min] elements];
  max = [[box max] elements];

  minz = MIN(min[2],max[2]);
  maxz = MAX(min[2],max[2]);

  if ((r = cullPlaneAABB(_tr_left,min,max))) {
    return G3DOutside;
  }

  if ((r = cullPlaneAABB(_tr_right,min,max))) {
    return G3DOutside;
  }

  if ((r = cullPlaneAABB(_tr_top,min,max))) {
    return G3DOutside;
  }

  if ((r = cullPlaneAABB(_tr_bottom,min,max))) {
    return G3DOutside;
  }

  return G3DInside;
}

- (G3DLocation)containsOBB:(G3DBox *)box
{
  return G3DInside;
}

- (BOOL)containsVector:(G3DVector4d *)aVec
{
  /*
  float *pt = [aVec elements];
  float tmp[4];

  tmp[0] = pt[0];
  tmp[1] = pt[1];
  tmp[2] = pt[2];
  tmp[3] = 1.0;

  G3DTransformVector4dv(tmp,tmp,[matrix elements]);

  // Test between [0,1]

  return (((tmp[0] >= -tmp[3]) | ((tmp[0] <= tmp[3]) << 1) | (( tmp[1] <=  tmp[3] ) << 2) | ((tmp[1] >= -tmp[3] ) << 3) | ((tmp[2] >= -tmp[3]) << 4) | ((tmp[2] <=  tmp[3]) << 5)) == 0x3F);
  */
}

@end






