//-----------------------------------------------------------------------------
// Project            	RenderKit
// Class		G3DGraphNode
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Thu Sep  9 11:05:08 CEST 1999
//
// Copyright (c) Philippe C.D. Robert
//
// The RenderKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DGraphNode.m,v 1.29 2001/01/09 15:30:21 robert Exp $
//-----------------------------------------------------------------------------

#import "G3DGraphNode.h"

#import "G3DLeaf.h"
#import "G3DFrustum.h"
#import "G3DTransform.h"
#import "G3DWorld.h"
#import "G3DGroup.h"
#import "G3DBVol.h"
#import "G3DArchiver.h"
#import "G3DUnarchiver.h"

NSString *G3DGraphNodeException = @"G3DGraphNodeException";
NSString *G3DGraphNodeDesignFlawException = @"G3DGraphNodeDesignFlawException";
NSString *G3DGraphNodeOptimisationException = @"G3DGraphNodeOptimisationException";
NSString *G3DGraphNodeCompilationException = @"G3DGraphNodeCompilationException";
NSString *G3DGraphNodeCollisionDetectionException = @"G3DGraphNodeCollisionDetectionException";

@implementation G3DGraphNode

//-----------------------------------------------------------------------------
// Init and Free
//-----------------------------------------------------------------------------

- (id)init
{
  if ((self = [super init])) {
    name = [[NSString alloc] initWithString:@"Generic Graphnode"];

    // Private ivars
    _isCompiled = NO;

    isVisible = YES;

    _shouldNotifyBeforeRendering = NO;
    _shouldNotifyAfterRendering = NO;
    _shouldNotifyAfterUpdating = NO;
    
    _shouldVFCull = YES;
    _bVolIsDirty = YES;
    _recentlyUpdated = NO;
    _renderAsBVol = NO;
    _hasBVolume = YES;

    transformation = [[G3DTransform alloc] init];
    [transformation setNode:self];

    _fwdT = [[G3DMatrix4f alloc] init];
    _transfIsDirty = NO;

    level = -1;
    cullDepth = 3;

    renderPriority.priority = G3D_PRIORITY_GENERIC;
    renderPriority.value = 0;

    // To be sure...
    parent = nil;
  }
  return self;
}

- (void)dealloc
{
  [name release];
  [transformation release];
  [_fwdT release];

  if (userData) {
    [userData release];
  }

  [super dealloc];
}

//-----------------------------------------------------------------------------
// Render delegate
//-----------------------------------------------------------------------------

- (void)setDelegate:(id)aDelegate
{
  delegate = aDelegate;

  if (delegate) {
    _shouldNotifyBeforeRendering = ([delegate respondsToSelector:@selector(graphNodeWillRender:)]) ? YES : NO;

    _shouldNotifyAfterRendering = ([delegate respondsToSelector:@selector(graphNodeDidRender:)]) ? YES : NO;

    _shouldNotifyAfterUpdating = ([delegate respondsToSelector:@selector(graphNodeDidUpdate:)]) ? YES : NO;
  }
  else {
    _shouldNotifyBeforeRendering = NO;
    _shouldNotifyAfterRendering = NO;
    _shouldNotifyAfterUpdating = NO;
  }
}

- (id)delegate
{
  return delegate;
}

//-----------------------------------------------------------------------------
// Rendering and traversing the scenegraph
//-----------------------------------------------------------------------------

- (void)renderWithMask:(unsigned)mask frustum:(G3DFrustum *)fr
{
  [NSException raise:G3DGraphNodeDesignFlawException format:@"You have to override this method in order to use it!"];
}

- (void)setShouldVFCull:(BOOL)yn
{
  if (yn && !_hasBVolume) {
    DLOG(@"%@ %x: no bounding volume for culling available - ignoring!",[self class],self);
    return;
  }

  _shouldVFCull = yn;
}

- (BOOL)shouldVFCull
{
  return _shouldVFCull;
}

- (void)setCullDepth:(unsigned)dep
{
  cullDepth = dep;
}

- (unsigned)cullDepth
{
  return cullDepth;
}

- (void)setIsVisible:(BOOL)flag
{
  isVisible = flag;
}

- (BOOL)isVisible
{
  return isVisible;
}

- (void)toggleVisibility
{
  isVisible = !isVisible;
}

//-----------------------------------------------------------------------------
// The parents...
//-----------------------------------------------------------------------------

- (G3DGraphNode *)parent
{
  return parent;
}

- (void)setParent:(G3DGraphNode *)aParent
{
  parent = aParent;
}

//-----------------------------------------------------------------------------
// User data
//-----------------------------------------------------------------------------

- (id)userData
{
  return userData;
}

- (void)setUserData:(id)data
{
  [userData autorelease];
  userData = [data retain];
}

//-----------------------------------------------------------------------------
// Transformation
//-----------------------------------------------------------------------------

- (void)update
{
  G3DMatrix4f *tr;

  /*
   * If we are the top most node, use the identity singleton! Updating occurs
   * only when we are in a graph context, this eliminates CPU load...
   */

  if (!parent) {
    tr = [G3DMatrix4f matrix4f];
  }
  else {
    tr = [parent modelTransformation];
  }

  [self updateWithTransformation:tr];
}

- (void)updateIfNeeded
{
  if (_transfIsDirty) {
    [self update];
  }
}

-(void)updateWithTransformation:(G3DMatrix4f *)t
{
  float _transf[16];

  /*
   * Calculates modelTransformation = t * transformation, can be optimised,
   * if _transf is the identity!
   *
   */

  [transformation getTransformationValues:_transf];
  G3DMultiplyMatrix4fv(_fwdT->matrix,t->matrix,_transf);

  _transfIsDirty = NO;

  if (_shouldNotifyAfterUpdating) {
    [delegate graphNodeDidUpdate:self];
  }

  DLOG(@"Updated %@ %x with transformation %x -> new: %@ %x",[self class],self,t,[_fwdT class],_fwdT);
}

- (G3DMatrix4f *)modelTransformation
{
  return _fwdT;
}

- (G3DTransform *)transformation
{
  return transformation;
}

- (void)setTransformation:(G3DTransform *)transf;
{
  NSAssert(transf != nil,@"You cannot assign nil to this node as a valid transformation!");
  
  if (_isCompiled) {
    [NSException raise:G3DGraphNodeCompilationException format:@"Cannot modify a compiled node."];
  }

  [transformation autorelease];
  transformation = [transf copy];

  [transformation setNode:self];

  // Update the transformations
  _transfIsDirty = YES;
}

//-----------------------------------------------------------------------------
// Optimisation
//-----------------------------------------------------------------------------

- (void)compile
{
  [NSException raise:G3DGraphNodeDesignFlawException format:@"You have to override this method in order to use it!"];  
}

- (BOOL)isCompiled
{
  return _isCompiled;
}

- (void)updatePriority
{
  DLOG(@"<%@ %x>: default implementation does nothing!",[self class],self);
}

- (G3D_SortPriority*)renderPriority
{
  return &renderPriority;
}

//-----------------------------------------------------------------------------
// Node information
//-----------------------------------------------------------------------------

- (void)setName:(NSString *)aName
{
  [name autorelease];
  name = [aName copy];
}

- (NSString *)name
{
  return name;
}

- (NSString *)description
{
  return [NSString stringWithFormat:@"<%@ %x, thread %x>: name %@",[self class], self, [NSThread currentThread],[self name]];
}

- (int)level
{
  /*
   * parent == nil => we are the root, that is level 0!
   */

  if (level == -1) {
    if (parent) {
      level += [parent level];
    }
    else {
      level = 0;
    }
  }
  return level;
}

//-----------------------------------------------------------------------------
// NSCoding
//-----------------------------------------------------------------------------

- (void)encodeWithCoder:(NSCoder *)aCoder
{
  [aCoder encodeObject:transformation];
  [aCoder encodeObject:_fwdT];
  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&_transfIsDirty];

  [aCoder encodeObject:parent];
  [aCoder encodeObject:name];
  [aCoder encodeObject:delegate];
  [aCoder encodeObject:userData];
 
  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&isVisible];
  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&_shouldNotifyBeforeRendering];
  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&_shouldNotifyAfterRendering];
  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&_shouldNotifyAfterUpdating];
  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&_hasDelegate];

  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&_shouldVFCull];
  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&_isCompiled];
 
  [aCoder encodeObject:bVolume];
  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&_bVolIsDirty];
  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&_renderAsBVol];
  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&_hasBVolume];

  [aCoder encodeValueOfObjCType:@encode(unsigned) at:&cullDepth];

  [aCoder encodeValueOfObjCType:@encode(int) at:&renderPriority.priority];
  [aCoder encodeValueOfObjCType:@encode(int) at:&renderPriority.value];
}

- (id)initWithCoder:(NSCoder *)aCoder
{
  self = [super init];

  transformation = [[aCoder decodeObject] retain];
  _fwdT = [[aCoder decodeObject] retain];
  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&_transfIsDirty];

  parent = [[aCoder decodeObject] retain];
  name = [[aCoder decodeObject] retain];
  delegate = [[aCoder decodeObject] retain];
  userData = [[aCoder decodeObject] retain];

  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&isVisible];
  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&_shouldNotifyBeforeRendering];
  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&_shouldNotifyAfterRendering];
  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&_shouldNotifyAfterUpdating];
  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&_hasDelegate];

  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&_shouldVFCull];
  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&_isCompiled];

  bVolume = [[aCoder decodeObject] retain];
  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&_bVolIsDirty];
  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&_renderAsBVol];
  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&_hasBVolume];

  [aCoder decodeValueOfObjCType:@encode(unsigned) at:&cullDepth];

  [aCoder decodeValueOfObjCType:@encode(int) at:&renderPriority.priority];
  [aCoder decodeValueOfObjCType:@encode(int) at:&renderPriority.value];

  return self;
}

//-----------------------------------------------------------------------------
// NSCopying
//-----------------------------------------------------------------------------

- (id)copyWithZone:(NSZone *)zone
{
  G3DGraphNode *copy = [[[self class] allocWithZone:zone] init];

  /*
   * We remove the node from the hierarchy, that is we don't copy
   * the parent nor the model transformation. This means, the node
   * cannot be compiled nor optimised!
   *
   */

  [copy setTransformation:transformation];
  [copy setIsVisible:isVisible];

  copy->parent = nil;
  copy->level = -1;

  copy->renderPriority = renderPriority;

  [copy setDelegate:delegate];
  [copy setName:name];
  [copy setUserData:userData];
  [copy setShouldVFCull:_shouldVFCull];
  [copy setCullDepth:cullDepth];

  copy->_shouldNotifyBeforeRendering = _shouldNotifyBeforeRendering;
  copy->_shouldNotifyAfterRendering = _shouldNotifyAfterRendering;
  copy->_shouldNotifyAfterUpdating = _shouldNotifyAfterUpdating;
  copy->_hasDelegate = _hasDelegate;

  copy->_isCompiled = NO;

  if (bVolume) {
    copy->bVolume = [bVolume copy];
  }
  copy->_bVolIsDirty = _bVolIsDirty;
  copy->_renderAsBVol = _renderAsBVol;
  copy->_hasBVolume = _hasBVolume;
  
  return copy;
}

//-----------------------------------------------------------------------------
// G3DSceneCoding
//-----------------------------------------------------------------------------

- (void)encodeWithSceneArchiver:(G3DArchiver *)archiver
{
  [archiver encodeObject:transformation named:@"transformation"];
  [archiver encodeObject:_fwdT named:@"_fwdT"];
  [archiver encodeBOOL:_transfIsDirty named:@"_transfIsDirty"];

  [archiver encodeObject:parent named:@"parent"];
  [archiver encodeObject:name named:@"name"];
  [archiver encodeObject:delegate named:@"delegate"];
  [archiver encodeObject:userData named:@"userData"];
 
  [archiver encodeBOOL:isVisible named:@"isVisible"];
  [archiver encodeBOOL:_shouldNotifyBeforeRendering named:@"_shouldNotifyBeforeRendering"];
  [archiver encodeBOOL:_shouldNotifyAfterRendering named:@"_shouldNotifyAfterRendering"];
  [archiver encodeBOOL:_shouldNotifyAfterUpdating named:@"_shouldNotifyAfterUpdating"];
  [archiver encodeBOOL:_hasDelegate named:@"_hasDelegate"];

  [archiver encodeBOOL:_shouldVFCull named:@"_shouldVFCull"];
  [archiver encodeBOOL:_isCompiled named:@"_isCompiled"];
 
  [archiver encodeObject:bVolume named:@"bVolume"];
  [archiver encodeBOOL:_bVolIsDirty named:@"_bVolIsDirty"];
  [archiver encodeBOOL:_renderAsBVol named:@"_renderAsBVol"];
  [archiver encodeBOOL:_hasBVolume named:@"_hasBVolume"];

  [archiver encodeUnsignedInt:cullDepth named:@"cullDepth"];

  [archiver encodeInt:renderPriority.priority named:@"renderPriority.priority"];
  [archiver encodeInt:renderPriority.value named:@"renderPriority.value"];
}

- (id)initWithSceneUnarchiver:(G3DUnarchiver *)unarchiver
{
  self = [super init];

  transformation = [[unarchiver decodeObjectNamed:@"transformation"] retain];
  _fwdT = [[unarchiver decodeObjectNamed:@"_fwdT"] retain];
  [unarchiver decodeBOOL:&_transfIsDirty named:@"_transfIsDirty"];

  parent = [[unarchiver decodeObjectNamed:@"parent"] retain];
  name = [[unarchiver decodeObjectNamed:@"name"] retain];
  delegate = [[unarchiver decodeObjectNamed:@"delegate"] retain];
  userData = [[unarchiver decodeObjectNamed:@"userData"] retain];

  [unarchiver decodeBOOL:&isVisible named:@"isVisible"];
  [unarchiver decodeBOOL:&_shouldNotifyBeforeRendering named:@"_shouldNotifyBeforeRendering"];
  [unarchiver decodeBOOL:&_shouldNotifyAfterRendering named:@"_shouldNotifyAfterRendering"];
  [unarchiver decodeBOOL:&_shouldNotifyAfterUpdating named:@"_shouldNotifyAfterUpdating"];
  [unarchiver decodeBOOL:&_hasDelegate named:@"_hasDelegate"];

  [unarchiver decodeBOOL:&_shouldVFCull named:@"_shouldVFCull"];
  [unarchiver decodeBOOL:&_isCompiled named:@"_isCompiled"];

  bVolume = [[unarchiver decodeObjectNamed:@"bVolume"] retain];
  [unarchiver decodeBOOL:&_bVolIsDirty named:@"_bVolIsDirty"];
  [unarchiver decodeBOOL:&_renderAsBVol named:@"_renderAsBVol"];
  [unarchiver decodeBOOL:&_hasBVolume named:@"_hasBVolume"];

  [unarchiver decodeUnsignedInt:&cullDepth named:@"cullDepth"];

  [unarchiver decodeInt:&renderPriority.priority named:@"renderPriority.priority"];
  [unarchiver decodeInt:&renderPriority.value named:@"renderPriority.value"];

  return self;
}

@end

@implementation G3DGraphNode (BoundingVolumeHandling)

- (G3DBVol *)boundingVolume
{
  if (!_hasBVolume) {
    DLOG(@"%@ %x has no bounding volume - ignoring...",[self class],self);
    return nil;
  }
  
  if (_bVolIsDirty) {
    [self updateBVolume];
  }
  return bVolume;
}

- (void)setNeedsUpdate:(BOOL)yn
{
  if (!_hasBVolume) {
    DLOG(@"%@ %x has no bounding volume - ignoring...",[self class],self);
    return;
  }
  
  _bVolIsDirty = yn;

  if (yn && parent) {
    [parent setNeedsUpdate:YES];
  }
}

- (BOOL)needsUpdate
{
  return _bVolIsDirty;
}

- (BOOL)recentlyUpdated
{
  return _recentlyUpdated;
}

- (void)setRecentlyUpdated:(BOOL)yn
{
  _recentlyUpdated = yn;
}

- (void)setRenderAsBVolume:(BOOL)flag
{
  if (!_hasBVolume) {
    DLOG(@"%@ %x has no bounding volume - ignoring...",[self class],self);
    return;
  }
  
  _renderAsBVol = flag;
}

- (BOOL)renderAsBVolume
{
  return _renderAsBVol;
}

- (void)setHasBVolume:(BOOL)yn
{
  BOOL _old = _hasBVolume;
  
  _hasBVolume = yn;

  if (_hasBVolume && _hasBVolume != _old) {
    _bVolIsDirty = YES;
  }
}

- (BOOL)hasBVolume
{
  return _hasBVolume;
}

- (void)updateBVolume
{
  DLOG(@"<%@ %x, thread %x>: updateBVolume invoked on a node with no bounding volume!",[self class],self,[NSThread currentThread]);
}

- (void)updateBVolumeIfNeeded
{
  if (_bVolIsDirty && _hasBVolume) {
    [self updateBVolume];
    _recentlyUpdated = YES;
  }
}

- (BOOL)collidesWithNode:(G3DGraphNode *)anotherNode
{
  G3DBVol *_volB;

  if (!_hasBVolume || !(_volB = [anotherNode boundingVolume])) {
    [NSException raise:G3DGraphNodeCollisionDetectionException format:@"%@ %x cannot test %@ %x!",self,[self class],anotherNode,[anotherNode class]];
    return NO;
  }

  return [bVolume intersectsWithVolume:_volB];
}

@end





