//-----------------------------------------------------------------------------
// Project            	RenderKit
// Class		G3DGroup
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Wed Nov 24 11:10:03 CET 1999
//
// Copyright (c) Philippe C.D. Robert
//
// The RenderKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DGroup.m,v 1.34 2001/07/15 12:39:24 probert Exp $
//-----------------------------------------------------------------------------

#import "G3DGroup.h"

#import "G3DTransform.h"
#import "G3DLeaf.h"
#import "G3DSwitchGroup.h"
#import "G3DFrustum.h"
#import "G3DFunctions.h"
#import "G3DWorld.h"
#import "G3DBVol.h"
#import "G3DArchiver.h"
#import "G3DUnarchiver.h"
#import "G3DLight.h"
#import "G3DNodeList.h"

NSString *G3DGroupException = @"G3DGroupException";

NSString *G3DLightRemovedNotification = @"G3DLightRemovedNotification";
NSString *G3DLightAddedNotification = @"G3DLightAddedNotification";

@interface G3DGraphNode (PrivateTraversals)

- (void)_joinArray:(NSMutableArray **)_arr;

@end

@implementation G3DGraphNode (PrivateTraversals)

- (void)_joinArray:(NSMutableArray **)_arr
{
  if ([self isKindOfClass:[G3DLeaf class]]) {
    [*_arr addObject:self];
  }
  else {
    int i;
    G3DGroup *g = self;
    int _l = [g->children count];

    for (i=0;i<_l;i++) {
      [[g->children objectAtIndex:i] _joinArray:_arr];
    }
  }
}

@end

@implementation G3DGroup

static Class G3DLightClass = NULL;
static Class G3DLeafClass = NULL;
static Class G3DGroupClass = NULL;
static Class G3DSwitchGroupClass = NULL;
static Class G3DWorldNodeClass = NULL;

int reorderChildren(G3DGraphNode *n1, G3DGraphNode *n2, void *context)
{
  G3D_SortPriority *_p1 = [n1 renderPriority];
  G3D_SortPriority *_p2 = [n2 renderPriority];

  DLOG(@"*** Reordering %@ %x (p: %d) and %@ %x (p: %d)",[n1 class],n1,_p1->value,[n2 class],n2,_p2->value);

  if (_p1->value < _p2->value) {
    return NSOrderedAscending;
  }
  else if (_p1->value > _p2->value) {
    return NSOrderedDescending;
  }

  return NSOrderedSame;
}

//-----------------------------------------------------------------------------
// Class methods
//-----------------------------------------------------------------------------

+ (void)initialize
{
  if (self == [G3DGroup class]) {
    G3DLightClass = [G3DLight class];
    G3DLeafClass = [G3DLeaf class];
    G3DGroupClass = [G3DGroup class];
    G3DSwitchGroupClass = [G3DSwitchGroup class];
    G3DWorldNodeClass = [G3DWorld class];
  }
}

+ (id)group
{
  return [[[[self class] alloc] initWithCapacity:0 allowReordering:YES] autorelease];
}

+ (id)groupWithCapacity:(unsigned int)cap allowReordering:(BOOL)yn
{
  return [[[[self class] alloc] initWithCapacity:cap allowReordering:yn] autorelease];
}

//-----------------------------------------------------------------------------
// Init and free
//-----------------------------------------------------------------------------

- (id)init
{
  return [self initWithCapacity:0 allowReordering:YES];
}

- (id)initWithCapacity:(unsigned int)cap allowReordering:(BOOL)yn
{
  if ((self = [super init])) {
    children = [[NSMutableArray alloc] initWithCapacity:cap];

    _children = 0;
    _allowReordering = yn;
    _isOptimised = NO;
    _needsOptimisation = YES;
    _isOptimised = NO;
  }
  return self;
}

- (void)dealloc
{
  [children release];

  if (_flattenedNodes) {
    [_flattenedNodes release];
  }

  [super dealloc];
}

//-----------------------------------------------------------------------------
// Rendering and traversing the scenegraph
//-----------------------------------------------------------------------------

- (void)renderWithMask:(unsigned)mask frustum:(G3DFrustum *)fr
{
  register int i;

  if (!isVisible) {
    return;
  }

  if ((mask & (1 << VIEW_FRUSTUM_CULLING))) {
    if (_shouldVFCull && [fr containsVolume:bVolume] == G3DOutside) {
      DLOG(@"View-Frustum culling: Node and its children won't be rendered");
      return;
    }
  }

  if (_shouldNotifyBeforeRendering) {
    [delegate graphNodeWillRender:self];
  }

  if (_flattenedNodes) {
    [_flattenedNodes renderWithMask:mask frustum:fr];
    return;
  }

  G3D_PUSH_MATRIX;

  if (!_isCompiled) {
    [transformation transform];
  }

  // Render! (if _isCompiled and nodeList -> optimise!)
  for (i=0;i<_children;i++) {
    [[children objectAtIndex:i] renderWithMask:mask frustum:fr];
  }
  
  if (_shouldNotifyAfterRendering) {
    [delegate graphNodeDidRender:self];
  }

  G3D_POP_MATRIX;
}

- (void)updateIfNeeded
{
  [super updateIfNeeded];

  // Query all children...
  if (_isCompiled == NO && _children) {
    register int i;

    for (i=0;i<_children;i++) {
      [[children objectAtIndex:i] updateIfNeeded];
    }
  }
}

-(void)updateWithTransformation:(G3DMatrix4f *)t
{
  [super updateWithTransformation:t];

  // Do it on all children...
  if (_isCompiled == NO && _children) {
    register int i;

    for (i=0;i<_children;i++) {
      [[children objectAtIndex:i] updateWithTransformation:_fwdT];
    }
  }
}

//-----------------------------------------------------------------------------
// Bounding Volume
//-----------------------------------------------------------------------------

- (void)updateBVolume
{
  if (_isCompiled) {
    [NSException raise:G3DGraphNodeCompilationException format:@"Cannot modify a compiled node."];
    return;
  }

  /*
   * Traverse down until _bVolIsDirty = NO, then update all bVolumes up until 
   * here.
   *
   */

  if (_children != 0) {
    int i;
    int touched[_children];
    int _cntr = 0;
    
    DLOG(@"%@ %x is updating bounding volume...",[self class],self);

    for (i=0;i<_children;i++) {
      G3DGraphNode *obj = [children objectAtIndex:i];

      [obj updateBVolumeIfNeeded];
      if ([obj recentlyUpdated]) {
        touched[_cntr++] = i;
        [obj setRecentlyUpdated:NO];
      }
    }

    // Now, update the current bounding box with above updated bVolumes
    if (!_hasBVolume) {
      DLOG(@"%@ %x has no bounding volume - ignoring...",[self class],self);
    }
    else if (_bVolIsDirty) {      
      if (!bVolume) {
	bVolume = [[G3DBVol alloc] initWithType:G3D_Sphere];
	[bVolume setNode:self];
      }
      for (i=0;i<_cntr;i++) {
	G3DGraphNode *_gn = [children objectAtIndex:touched[i]];
        G3DBVol *bv = [_gn boundingVolume];
	DLOG(@"%d of %d (in %@ %x named %@)",i+1,_cntr,[self class],self,name);
	if (bv) {
	  [bVolume mergeWithBVolume:bv];
	}
	else {
	  DLOG(@"<%@ %x> ignoring invalid bv of child %@",self,[self class],[_gn description]);
	}
      }
      _bVolIsDirty = NO;
    }
  }
}

//-----------------------------------------------------------------------------
// G3DGroupingNode methods
//-----------------------------------------------------------------------------

- (NSArray *)children
{
  return children;
}

- (NSArray *)leaves
{
  int i;
  NSMutableArray *_ma;

  if (!_children) {
    return nil;
  }

  _ma = [NSMutableArray arrayWithCapacity:_children];

  for (i=0;i<_children;i++) {
    [[children objectAtIndex:i] _joinArray:&_ma];
  }

  return ([_ma count]) ? _ma : nil;
}

- (void)addChild:(G3DGraphNode *)aChild
{
  if (_isCompiled) {
    [NSException raise:G3DGroupException format:@"Cannot modify a compiled node."];
    return;
  }

  if ([aChild parent] != nil) {
    [NSException raise:G3DGroupException format:@"Cannot add prebound child."];
    return;
  }

  [children addObject:aChild];
  
  _children++;
  
  // Set self as the parent
  [aChild setParent:self];
  aChild->level = [self level] + 1;

  if ([aChild cullDepth] < aChild->level) {
    [aChild setShouldVFCull:NO];
  }

  if ([aChild isKindOfClass:G3DLightClass]) {
    if ([(G3DLight *)aChild isGlobal]) {
      [[NSNotificationCenter defaultCenter] postNotificationName:G3DLightAddedNotification object:aChild];
    }
  }

  // This is handled lazily!
  [self setNeedsUpdate:YES];
  _isOptimised = NO;
}

- (void)insertChild:(G3DGraphNode *)aChild atIndex:(int)ind
{
  if (_isCompiled) {
    [NSException raise:G3DGroupException format:@"Cannot modify a compiled node."];
    return;
  }

  if ([aChild parent]) {
    [NSException raise:G3DGroupException format:@"Cannot add prebound child."];
    return;
  }

  [children insertObject:aChild atIndex:ind];

  _children++;
  
  // Set self as the parent
  [aChild setParent:self];
  aChild->level = [self level] + 1;

  if ([aChild cullDepth] < aChild->level) {
    [aChild setShouldVFCull:NO];
  }

  if ([aChild isKindOfClass:G3DLightClass]) {
    if ([(G3DLight *)aChild isGlobal]) {
      [[NSNotificationCenter defaultCenter] postNotificationName:G3DLightAddedNotification object:aChild];
    }
  }

  // This is handled lazily!
  [self setNeedsUpdate:YES];
  _isOptimised = NO;
}

- (int)childrenCount
{
  return _children;
}

- (G3DGraphNode *)childAtIndex:(int)index
{
  return [children objectAtIndex:index];
}

- (NSEnumerator *)childEnumerator
{
  return [children objectEnumerator];
}

- (void)removeChildAtIndex:(int)ind
{
  G3DGraphNode *child;

  if (_isCompiled) {
    [NSException raise:G3DGroupException format:@"Cannot modify a compiled node."];
    return;
  }

  if ((child = [children objectAtIndex:ind])) {
    [child retain];
    [children removeObject:child];

    _children--;

    if ([child isKindOfClass:G3DLightClass]) {
      if ([(G3DLight *)child isGlobal]) {
	[[NSNotificationCenter defaultCenter] postNotificationName:G3DLightRemovedNotification object:child];
      }
    }
    
    // Set self as the parent
    [child setParent:nil];
    child->level = -1;
    [child release];

    // This is handled lazily!
    [self setNeedsUpdate:YES];
    _isOptimised = NO;
  }
}

- (void)removeChild:(G3DGraphNode *)aChild
{
  if (_isCompiled) {
    [NSException raise:G3DGroupException format:@"Cannot modify a compiled node."];
    return;
  }

  if ([children containsObject:aChild]) {
    [aChild retain];
    [children removeObject:aChild];

    _children--;

    if ([aChild isKindOfClass:G3DLightClass]) {
      if ([(G3DLight *)aChild isGlobal]) {
	[[NSNotificationCenter defaultCenter] postNotificationName:G3DLightRemovedNotification object:aChild];
      }
    }
    
    // Set self as the parent
    [aChild setParent:nil];
    aChild->level = -1;
    [aChild release];

    // This is handled lazily!
    [self setNeedsUpdate:YES];
    _isOptimised = NO;
  }
}

- (void)removeAllChildren
{
  NSEnumerator *e;
  G3DGraphNode *child;

  if (_isCompiled) {
    [NSException raise:G3DGroupException format:@"Cannot modify a compiled node."];
    return;
  }

  e = [self childEnumerator];
  while ((child = [e nextObject])) {
    [child setParent:nil];
    child->level = -1;

    if ([child isKindOfClass:G3DLightClass]) {
      if ([(G3DLight *)child isGlobal]) {
	[[NSNotificationCenter defaultCenter] postNotificationName:G3DLightRemovedNotification object:child];
      }
    }
  }

  _children = 0;

  [children removeAllObjects];

  // This is handled lazily!
  [self setNeedsUpdate:YES];
  _isOptimised = NO;
}

- (void)replaceChildAtIndex:(int)ind withChild:(G3DGraphNode *)aChild
{
  G3DGraphNode *old;

  if (_isCompiled) {
    [NSException raise:G3DGroupException format:@"Cannot modify a compiled node."];
    return;
  }

  old = [children objectAtIndex:ind];
  [old setParent:nil];
  old->level = -1;

  if ([old isKindOfClass:G3DLightClass]) {
    if ([(G3DLight *)old isGlobal]) {
      [[NSNotificationCenter defaultCenter] postNotificationName:G3DLightRemovedNotification object:old];
    }
  }
  
  [children replaceObjectAtIndex:ind withObject:aChild];
  [aChild setParent:self];
  aChild->level = [self level] + 1;

  if ([aChild isKindOfClass:G3DLightClass]) {
    if ([(G3DLight *)aChild isGlobal]) {
      [[NSNotificationCenter defaultCenter] postNotificationName:G3DLightAddedNotification object:aChild];
    }
  }

  // This is handled lazily!
  [self setNeedsUpdate:YES];
  _isOptimised = NO;
}

//-----------------------------------------------------------------------------
// Sorting
//-----------------------------------------------------------------------------

- (void)sortUsingFunction:(int(*)(id,id,void*))comparator context:(void *)ctx
{
  if (_isCompiled) {
    [NSException raise:G3DGroupException format:@"Cannot modify a compiled node."];
    return;
  }

  [children sortUsingFunction:comparator context:ctx];
}

- (void)sortUsingSelector:(SEL)comparator
{
  if (_isCompiled) {
    [NSException raise:G3DGroupException format:@"Cannot modify a compiled node."];
    return;
  }

  [children sortUsingSelector:comparator];
}

- (BOOL)allowReordering
{
  return _allowReordering;
}

//-----------------------------------------------------------------------------
// NSCoding
//-----------------------------------------------------------------------------

- (void)encodeWithCoder:(NSCoder *)aCoder
{
  [super encodeWithCoder:aCoder];
  
  [aCoder encodeObject:children];
  [aCoder encodeObject:_flattenedNodes];

  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&_isOptimised];
  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&_needsOptimisation];
  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&_allowReordering];
  [aCoder encodeValueOfObjCType:@encode(int) at:&_children];
}

- (id)initWithCoder:(NSCoder *)aCoder
{
  self = [super initWithCoder:aCoder];

  children = [[aCoder decodeObject] retain];
  _flattenedNodes = [[aCoder decodeObject] retain];

  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&_isOptimised];
  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&_needsOptimisation];
  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&_allowReordering];
  [aCoder decodeValueOfObjCType:@encode(int) at:&_children];

  return self;
}

//-----------------------------------------------------------------------------
// NSCopying
//-----------------------------------------------------------------------------

- (id)copyWithZone:(NSZone *)zone
{
  G3DGroup *copy = [super copyWithZone:zone];

  copy->children = [children copy];
  copy->_flattenedNodes = [_flattenedNodes copy];

  copy->_isOptimised = _isOptimised;
  copy->_needsOptimisation = _needsOptimisation;
  copy->_allowReordering = _allowReordering;
  copy->_children = _children;

  return copy;
}

//-----------------------------------------------------------------------------
// G3DSceneCoding
//-----------------------------------------------------------------------------

- (void)encodeWithSceneArchiver:(G3DArchiver *)archiver
{
  [super encodeWithSceneArchiver:archiver];

  [archiver encodeArray:children named:@"children"];
  [archiver encodeArray:_flattenedNodes named:@"_flattenedNodes"];

  [archiver encodeBOOL:_isOptimised named:@"_isOptimised"];
  [archiver encodeBOOL:_needsOptimisation named:@"_needsOptimisation"];
  [archiver encodeBOOL:_allowReordering named:@"_allowReordering"];
  [archiver encodeInt:_children named:@"_children"];
}

- (id)initWithSceneUnarchiver:(G3DUnarchiver *)unarchiver
{
  self = [super initWithSceneUnarchiver:unarchiver];

  children = [[unarchiver decodeArrayNamed:@"children"] retain];
  _flattenedNodes = [[unarchiver decodeArrayNamed:@"_flattenedNodes"] retain];

  [unarchiver decodeBOOL:&_isOptimised named:@"_isOptimised"];
  [unarchiver decodeBOOL:&_needsOptimisation named:@"_needsOptimisation"];
  [unarchiver decodeBOOL:&_allowReordering named:@"_allowReordering"];
  [unarchiver decodeInt:&_children named:@"_children"];

  return self;
}

@end

@implementation G3DGroup (GraphOptimisation)

//-----------------------------------------------------------------------------
// Optimisation
//-----------------------------------------------------------------------------

- (void)compile
{
  if (!_isCompiled) {
    BOOL _canBeFlattened = YES;
    int i;

    DLOG([NSString stringWithFormat:@"Compiling %@ %x (thread %x) named %@",[self class], self, [NSThread currentThread],name]);

    /*
     * 1st step: flatten the subgraph's transformations - this is already done!
     * 2nd step: generate a node list, if possible
     * 3rd step: (optional) generate a (display) list if needed
     */

    for (i=0;i<_children;i++) {
      G3DGraphNode *_child = [children objectAtIndex:i];

      [_child compile];

      if ([_child isKindOfClass:G3DGroupClass]){
	if (![(G3DGroup*)_child canBeFlattened]) {
	  _canBeFlattened = NO;
	}
      }
    }

    if (_canBeFlattened) {
      if (!parent) {
	_flattenedNodes = [[G3DNodeList alloc] initWithNodes:[self leaves]];

	DLOG(@"Node list built: %@ (node %@)",
	     [_flattenedNodes description],
	     name);
      }
    }

    _isCompiled = YES;
  }
}

- (void)optimise
{
  if (_isCompiled) {
    [NSException raise:G3DGraphNodeCompilationException 
		 format:@"Cannot optimise a compiled node."];
  }

  if (_isOptimised == NO) {
    int i;

    DLOG([NSString stringWithFormat:@"Optimising %@ %x (thread %x) named %@",
		   [self class], 
		   self, 
		   [NSThread currentThread],
		   name]);

    /*
     * 1st step: updating bounding volumes of all children
     * 2nd step: update node's bounding volume
     * 3rd step: sort the children if possible (texure object, ...)
     *
     */
    
    for (i=0;i<_children;i++) {
      G3DGraphNode *_ch = [children objectAtIndex:i];
      
      if ([_ch isKindOfClass:G3DLeafClass]) {
	[_ch updateBVolumeIfNeeded];
	[_ch updatePriority];
      }
      else {
	[(G3DGroup *)_ch optimise];
      }
    }
    
    [self updateBVolumeIfNeeded];
    [self updatePriority];

    if (_allowReordering) {
      [children sortUsingFunction:reorderChildren context:NULL];
    }

    _isOptimised = YES;
  }
#ifdef DEBUG
  else {
    NSLog(@"<%@ %x>: already optimised! Ignoring...",[self class],self);
  }
#endif DEBUG
}

- (void)optimiseIfNeeded
{
  if (_needsOptimisation) {
    [self optimise];
  }
}

- (BOOL)isOptimised
{
  return _isOptimised;
}

- (BOOL)canBeFlattened
{
  return YES;
}

- (void)updatePriority
{
  register int i;

  for (i=0;i<_children;i++) {
    G3D_SortPriority *p = [[children objectAtIndex:i] renderPriority];

    if (p->value > renderPriority.value) {
      renderPriority.value = p->value;
      renderPriority.priority = p->priority;
    }
  }
}

@end
