//-----------------------------------------------------------------------------
// Project            	RenderKit
// Class		G3DIVtxArray
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Thu Sep  9 11:05:08 CEST 1999
//
// Copyright (c) Philippe C.D. Robert
//
// The RenderKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DIVtxArray.m,v 1.7 2001/01/19 08:08:06 robert Exp $
//-----------------------------------------------------------------------------

#import "G3DIVtxArray.h"
#import "G3DShape.h"
#import "G3DDefines.h"
#import "G3DArchiver.h"
#import "G3DUnarchiver.h"

NSString *G3DIVtxArrayException = @"G3DIVtxArrayException";
NSString *G3DIVtxArrayFormatException = @"G3DIVtxArrayFormatException";
NSString *G3DIVtxArrayLockingException = @"G3DIVtxArrayLockingException";

@implementation G3DIVtxArray

static void G3DSetOffsetVals(G3DIVtxArray *self, G3DArrayFormat fmt)
{
  self->format = fmt;

  switch (fmt) {
  case G3D_V2:
    self->_vChunkSize = 2;
    self->ver.off= 0;
    self->col.off=-1;
    self->nor.off=-1;
    self->tex.off=-1;
    self->ver.length = 2;
    self->col.length = 0;
    self->nor.length = 0;
    self->tex.length = 0;
    break;
  case G3D_V3:
    self->_vChunkSize = 3;
    self->ver.off= 0;
    self->col.off=-1;
    self->nor.off=-1;
    self->tex.off=-1;
    self->ver.length = 3;
    self->col.length = 0;
    self->nor.length = 0;
    self->tex.length = 0;
    break;
  case G3D_C3V3:
    self->_vChunkSize = 6;
    self->ver.off= 0;
    self->col.off= 3;
    self->nor.off=-1;
    self->tex.off=-1;
    self->ver.length = 3;
    self->col.length = 3;
    self->nor.length = 0;
    self->tex.length = 0;
    break;
  case G3D_N3V3:
    self->_vChunkSize = 6;
    self->ver.off= 3;
    self->col.off=-1;
    self->nor.off= 0;
    self->tex.off=-1;
    self->ver.length = 3;
    self->col.length = 0;
    self->nor.length = 3;
    self->tex.length = 0;
    break;
  case G3D_C4N3V3:
    self->_vChunkSize = 10;
    self->ver.off= 7;
    self->col.off= 0;
    self->nor.off= 4;
    self->tex.off=-1;
    self->ver.length = 3;
    self->col.length = 4;
    self->nor.length = 3;
    self->tex.length = 0;
    break;
  case G3D_T2V3:
    self->_vChunkSize = 5;
    self->ver.off= 2;
    self->col.off=-1;
    self->nor.off=-1;
    self->tex.off= 0;
    self->ver.length = 3;
    self->col.length = 0;
    self->nor.length = 0;
    self->tex.length = 2;
    break;
  case G3D_T4V4:
    self->_vChunkSize = 8;
    self->ver.off= 4;
    self->col.off=-1;
    self->nor.off=-1;
    self->tex.off= 0;
    self->ver.length = 4;
    self->col.length = 0;
    self->nor.length = 0;
    self->tex.length = 4;
    break;
  case G3D_T2C3V3:
    self->_vChunkSize = 8;
    self->ver.off= 5;
    self->col.off= 2;
    self->nor.off=-1;
    self->tex.off= 0;
    self->ver.length = 3;
    self->col.length = 3;
    self->nor.length = 0;
    self->tex.length = 2;
    break;
  case G3D_T2N3V3:
    self->_vChunkSize = 8;
    self->ver.off= 5;
    self->col.off=-1;
    self->nor.off= 2;
    self->tex.off= 0;
    self->ver.length = 3;
    self->col.length = 0;
    self->nor.length = 3;
    self->tex.length = 2;
    break;
  case G3D_T2C4N3V3:
    self->_vChunkSize = 12;
    self->ver.off= 9;
    self->col.off= 2;
    self->nor.off= 6;
    self->tex.off= 0;
    self->ver.length = 3;
    self->col.length = 4;
    self->nor.length = 3;
    self->tex.length = 2;
    break;
  case G3D_T4C4N3V4:
    self->_vChunkSize = 15;
    self->ver.off= 11;
    self->col.off= 4;
    self->nor.off= 8;
    self->tex.off= 0;
    self->ver.length = 4;
    self->col.length = 4;
    self->nor.length = 3;
    self->tex.length = 4;
    break;
  default:
    [NSException raise:G3DIVtxArrayFormatException format:@"No valid format!"];
    break;
  }
}

//-----------------------------------------------------------------------------
// Class methods 
//-----------------------------------------------------------------------------

+ (G3DIVtxArray *)triangleArrayWithData:(const float *)a count:(unsigned)n
{
  return [[[[self class] alloc] initWithMode:G3D_TRIANGLES format:G3D_T4C4N3V4 count:n array:a] autorelease];
}

+ (G3DIVtxArray *)triangleStripArrayWithData:(const float *)a count:(unsigned)n
{
  return [[[[self class] alloc] initWithMode:G3D_TRIANGLE_STRIP format:G3D_T4C4N3V4 count:n array:a] autorelease];
}

//-----------------------------------------------------------------------------
// Init
//-----------------------------------------------------------------------------

- (id)initWithMode:(G3DGeometryMode)mode format:(G3DArrayFormat)fmt count:(int)cnt array:(const float *)varr
{
  NSAssert(cnt>0,@"No valid geometry provided!");

  if ((self = [super init])) {
    size_t n;

    // Set the stride value based on fmt
    G3DSetOffsetVals(self,fmt);

    geometryMode = mode;
    vertices = cnt;    
    n=vertices*_vChunkSize*sizeof(float);      
    array = malloc(n);

    if (varr!=NULL) {
      array = memcpy(array,varr,n);
      _isInitialised = YES;
    }
    else {
      memset(array,0,n);
      _isInitialised = NO;
    }

    // Not yet compiled!
    isLocked = NO;
  }
  return self;
}

- (void)dealloc
{
  if (array) {
    free(array);
  }

#ifdef GL_EXT_compiled_vertex_array
  if (isLocked) {
    glUnlockArraysEXT();
  }
#endif

  [super dealloc];
}

//-----------------------------------------------------------------------------
// Drawing
//-----------------------------------------------------------------------------

- (void)drawGeometry
{
  // Stride will be used for custom extensions!
  stride=0;
  glInterleavedArrays(format,stride,array);
  glDrawArrays(geometryMode,0,vertices);
}

//-----------------------------------------------------------------------------
// Accessor Methods
//-----------------------------------------------------------------------------

- (void)precompile
{
  if (isLocked) {
    [NSException raise:G3DIVtxArrayLockingException 
		 format:@"Array is already locked!"];
    return;
  }

  isLocked = YES;
  
#ifdef GL_EXT_compiled_vertex_array
  glLockArraysEXT(0,vertices);
#endif
}

- (int)vertexCount
{
  return vertices;
}

- (float *)array
{
  return array;
}

- (G3DArrayFormat)format
{
  return format;
}

- (G3DGeometryMode)geometryMode
{
  return geometryMode;
}

- (G3DRange)colour
{
  return col;
}

- (G3DRange)normal
{
  return nor;
}

- (G3DRange)vertex
{
  return ver;
}

- (G3DRange)texture
{
  return tex;
}

- (void)getMin:(float *)min max:(float *)max
{
  float *c;
  register int i;

  if (vertices == 0) {
    [NSException raise:G3DIVtxArrayException format:@"No vertices found in %@ %x!",[self class],self];
    return;
  }

  for (i=0;i<ver.length;i++) {
    c = &array[_vChunkSize+ver.off];

    min[i] = c[i];
    max[i] = c[i];
  }
  
  switch (ver.length) {
    case 2:
      for (i=0;i<vertices;i++) {
        c = &array[(_vChunkSize*i)+ver.off];

        min[0] = MIN(min[0],c[0]);
        min[1] = MIN(min[1],c[1]);        

        max[0] = MAX(max[0],c[0]);
        max[1] = MAX(max[1],c[1]);
      }
      break;
    case 3:
      for (i=0;i<vertices;i++) {
        c = &array[(_vChunkSize*i)+ver.off];

        min[0] = MIN(min[0],c[0]);
        min[1] = MIN(min[1],c[1]);
        min[2] = MIN(min[2],c[2]);        

        max[0] = MAX(max[0],c[0]);
        max[1] = MAX(max[1],c[1]);
        max[2] = MAX(max[2],c[2]);
      }
      break;
    case 4:
      for (i=0;i<vertices;i++) {
        c = &array[(_vChunkSize*i)+ver.off];

        min[0] = MIN(min[0],c[0]);
        min[1] = MIN(min[1],c[1]);
        min[2] = MIN(min[2],c[2]);
        min[3] = MIN(min[3],c[3]);        

        max[0] = MAX(max[0],c[0]);
        max[1] = MAX(max[1],c[1]);
        max[2] = MAX(max[2],c[2]);
        max[3] = MAX(max[3],c[3]);
      }
      break;
  }
}

- (void)setShape:(G3DShape *)aShape
{
  shape = aShape;  
}

//-----------------------------------------------------------------------------
// Geometry
//-----------------------------------------------------------------------------

- (void)setCoordinate:(float *)coord atIndex:(int)index
{
  int i;
  int _index;

  if (isLocked) {
    [NSException raise:G3DIVtxArrayLockingException format:@"Array is already locked!"];
    return;
  }

  NSAssert(index<vertices,@"Invalid index !");

  _index = (_vChunkSize*index)+ver.off;
  for (i=0;i<ver.length;i++) {
    array[_index+i] = coord[i];
  }
  
  if (shape) {
    [shape setNeedsUpdate:YES];
  }
}

- (float *)coordinateAtIndex:(int)index
{
  NSAssert(index<vertices,@"Invalid index !");
  return &array[(_vChunkSize*index)+ver.off];
}

- (void)setCoordinates:(float *)coords atIndex:(int)index count:(int)cnt
{
  int i;
  int j;
  int _index;

  if (isLocked) {
    [NSException raise:G3DIVtxArrayLockingException format:@"Array is already locked!"];
    return;
  }

  NSAssert(index<vertices,@"Invalid index !");

  for (j=0;j<cnt;j++) {
    int k=j*ver.length;

    _index = (_vChunkSize*(index+j))+ver.off;
    for (i=0;i<ver.length;i++) {
      array[_index+i] = coords[k+i];
    }
  }

  if (shape) {
    [shape setNeedsUpdate:YES];
  }
}

- (void)setCoordinates:(float *)coords range:(NSRange)rng
{
  [self setCoordinates:coords atIndex:rng.location count:rng.length];
}

- (void)setNormal:(float *)coord atIndex:(int)index
{
  int i;
  int _index;

  if (isLocked) {
    [NSException raise:G3DIVtxArrayLockingException format:@"Array is already locked!"];
    return;
  }

  NSAssert(index<vertices,@"Invalid index !");

  _index = (_vChunkSize*index)+nor.off;

  for (i=0;i<nor.length;i++) {
    array[_index+i] = coord[i];
  }
}

- (float *)normalAtIndex:(int)index
{
  NSAssert(index<vertices,@"Invalid index!");
  return &array[(_vChunkSize*index)+nor.off];
}

- (void)setNormals:(float *)coords atIndex:(int)index count:(int)cnt
{
  int i;
  int j;
  int _index;

  if (isLocked) {
    [NSException raise:G3DIVtxArrayLockingException format:@"Array is already locked!"];
    return;
  }

  NSAssert(index<vertices,@"Invalid index !");

  for (j=0;j<cnt;j++) {
    int k=j*nor.length;

    _index = (_vChunkSize*(index+j))+nor.off;
    for (i=0;i<nor.length;i++) {
      array[_index+i] = coords[k+i];
    }
  }
}

- (void)setNormals:(float *)coords range:(NSRange)rng
{
  [self setNormals:coords atIndex:rng.location count:rng.length];
}

//-----------------------------------------------------------------------------
// Colours
//-----------------------------------------------------------------------------

- (void)setColour:(float *)acol atIndex:(int)index
{
  int i;
  int _index;

  if (isLocked) {
    [NSException raise:G3DIVtxArrayLockingException format:@"Array is already locked!"];
    return;
  }

  NSAssert(index<vertices,@"Invalid index !");

  _index = (_vChunkSize*index)+col.off;
  for (i=0;i<col.length;i++) {
    array[_index+i] = acol[i];
  }
}

- (float *)colourAtIndex:(int)index
{
  NSAssert(index<vertices,@"Invalid index !");
  return &array[(_vChunkSize*index)+col.off];
}

- (void)setColours:(float *)cols atIndex:(int)index count:(int)cnt
{
  int i;
  int j;
  int _index;

  if (isLocked) {
    [NSException raise:G3DIVtxArrayLockingException format:@"Array is already locked!"];
    return;
  }

  NSAssert(index<vertices,@"Invalid index !");

  for (j=0;j<cnt;j++) {
    int k=j*col.length;

    _index = (_vChunkSize*(index+j))+col.off;
    for (i=0;i<col.length;i++) {
      array[_index+i] = cols[k+i];
    }
  }
}

- (void)setColours:(float *)cols range:(NSRange)rng
{
  [self setColours:cols atIndex:rng.location count:rng.length];
}

//-----------------------------------------------------------------------------
// Texture coordinates
//-----------------------------------------------------------------------------

- (void)setTexCoord:(float *)coord atIndex:(int)index
{
  int i;
  int _index;

  if (isLocked) {
    [NSException raise:G3DIVtxArrayLockingException format:@"Array is already locked!"];
    return;
  }

  NSAssert(index<vertices,@"Invalid index !");

  _index = (_vChunkSize*index)+tex.off;
  for (i=0;i<tex.length;i++) {
    array[_index+i] = coord[i];
  }
}

- (float *)texCoordAtIndex:(int)index
{
  NSAssert(index<vertices,@"Invalid index !");
  return &array[(_vChunkSize*index)+tex.off];
}

- (void)setTexCoords:(float *)coords atIndex:(int)index count:(int)cnt
{
  int i;
  int j;
  int _index;

  if (isLocked) {
    [NSException raise:G3DIVtxArrayLockingException format:@"Array is already locked!"];
    return;
  }

  NSAssert(index<vertices,@"Invalid index !");

  for (j=0;j<cnt;j++) {
    int k=j*tex.length;

    _index = (_vChunkSize*(index+j))+tex.off;
    for (i=0;i<tex.length;i++) {
      array[_index+i] = coords[k+i];
    }
  }
}

- (void)setTexCoords:(float *)coords range:(NSRange)rng
{
  [self setTexCoords:coords atIndex:rng.location count:rng.length];
}

//-----------------------------------------------------------------------------
// Information
//-----------------------------------------------------------------------------

- (NSString *)description
{
  NSString *_descr = [NSString string];
  int i;
  int j;

  _descr = [_descr stringByAppendingFormat:@"VertexArray: fmt %d, mode %d, stride %d\n",format,geometryMode,stride];

  for (i=0;i<vertices;i++) {
    int c = i*_vChunkSize;

    _descr = [_descr stringByAppendingFormat:@"Vertex %d: ",i];
    for (j=0;j<_vChunkSize;j++) {
      _descr = [_descr stringByAppendingFormat:@"%f ",array[c+j]];
    }
    _descr = [_descr stringByAppendingString:@"\n"];
  }

  return _descr;
}

//-----------------------------------------------------------------------------
// NSCoding
//-----------------------------------------------------------------------------

- (void)encodeWithCoder:(NSCoder *)aCoder
{
  [super encodeWithCoder:aCoder];

  [aCoder encodeObject:shape];
  
  [aCoder encodeValueOfObjCType:@encode(int) at:&col.off];
  [aCoder encodeValueOfObjCType:@encode(int) at:&col.length];

  [aCoder encodeValueOfObjCType:@encode(int) at:&ver.off];
  [aCoder encodeValueOfObjCType:@encode(int) at:&ver.length];

  [aCoder encodeValueOfObjCType:@encode(int) at:&nor.off];
  [aCoder encodeValueOfObjCType:@encode(int) at:&nor.length];

  [aCoder encodeValueOfObjCType:@encode(int) at:&tex.off];
  [aCoder encodeValueOfObjCType:@encode(int) at:&tex.length];

  [aCoder encodeValueOfObjCType:@encode(int) at:&vertices];
  [aCoder encodeValueOfObjCType:@encode(int) at:&stride];
  [aCoder encodeValueOfObjCType:@encode(int) at:&format];
  [aCoder encodeValueOfObjCType:@encode(int) at:&geometryMode];
  [aCoder encodeValueOfObjCType:@encode(int) at:&_vChunkSize];
  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&_isInitialised];
  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&isLocked]; // ?

  [aCoder encodeBytes:&array length:vertices*_vChunkSize*sizeof(float)];
}

- (id)initWithCoder:(NSCoder *)aCoder
{
  unsigned n;

  self = [super initWithCoder:aCoder];

  shape = [aCoder decodeObject];
  
  [aCoder decodeValueOfObjCType:@encode(int) at:&col.off];
  [aCoder decodeValueOfObjCType:@encode(int) at:&col.length];

  [aCoder decodeValueOfObjCType:@encode(int) at:&ver.off];
  [aCoder decodeValueOfObjCType:@encode(int) at:&ver.length];

  [aCoder decodeValueOfObjCType:@encode(int) at:&nor.off];
  [aCoder decodeValueOfObjCType:@encode(int) at:&nor.length];

  [aCoder decodeValueOfObjCType:@encode(int) at:&tex.off];
  [aCoder decodeValueOfObjCType:@encode(int) at:&tex.length];

  [aCoder decodeValueOfObjCType:@encode(int) at:&vertices];
  [aCoder decodeValueOfObjCType:@encode(int) at:&stride];
  [aCoder decodeValueOfObjCType:@encode(int) at:&format];
  [aCoder decodeValueOfObjCType:@encode(int) at:&geometryMode];
  [aCoder decodeValueOfObjCType:@encode(int) at:&_vChunkSize];
  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&_isInitialised];
  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&isLocked]; // ?

  n=vertices*_vChunkSize*sizeof(float);
  array = (float*)[aCoder decodeBytesWithReturnedLength:&n];

  return self;
}

//-----------------------------------------------------------------------------
// NSCopying
//-----------------------------------------------------------------------------

- (id)copyWithZone:(NSZone *)zone
{
  G3DIVtxArray *copy = [super copyWithZone:zone];
  size_t n;
  
  copy->col = col;
  copy->ver = ver;
  copy->nor = nor;
  copy->tex = tex;
  copy->vertices = vertices;
  copy->stride = stride;
  copy->format = format;
  copy->shape = shape;
  copy->geometryMode = geometryMode;
  copy->_vChunkSize = _vChunkSize;
  copy->_isInitialised = _isInitialised;
  copy->isLocked = isLocked;

  n=vertices*_vChunkSize*sizeof(float);
  copy->array = malloc(n);
  copy->array = memcpy(copy->array,array,n);

  return copy;
}

//-----------------------------------------------------------------------------
// G3DSceneCoding
//-----------------------------------------------------------------------------

- (void)encodeWithSceneArchiver:(G3DArchiver *)archiver
{
  [super encodeWithSceneArchiver:archiver];

  [archiver encodeObject:shape named:@"shape"];
  
  [archiver encodeInt:col.off named:@"col.off"];
  [archiver encodeInt:col.length named:@"col.length"];

  [archiver encodeInt:ver.off named:@"ver.off"];
  [archiver encodeInt:ver.length named:@"ver.length"];

  [archiver encodeInt:nor.off named:@"nor.off"];
  [archiver encodeInt:nor.length named:@"nor.length"];

  [archiver encodeInt:tex.off named:@"tex.off"];
  [archiver encodeInt:tex.length named:@"tex.length"];

  [archiver encodeInt:vertices named:@"vertices"];
  [archiver encodeInt:stride named:@"stride"];
  [archiver encodeInt:format named:@"format"];
  [archiver encodeInt:geometryMode named:@"geometryMode"];
  [archiver encodeInt:_vChunkSize named:@"_vChunkSize"];

  [archiver encodeBOOL:_isInitialised named:@"_isInitialised"];
  [archiver encodeBOOL:isLocked named:@"isLocked"]; // ?

#warning encodeBytes:length:named: is not yet implemented!
  //[archiver encodeBytes:&array length:vertices*_vChunkSize*sizeof(float)];
}

- (id)initWithSceneUnarchiver:(G3DUnarchiver *)unarchiver
{
  unsigned n;

  self = [super initWithSceneUnarchiver:unarchiver];
  
  shape = [unarchiver decodeObjectNamed:@"shape"];
  
  [unarchiver decodeInt:&col.off named:@"col.off"];
  [unarchiver decodeInt:&col.length named:@"col.length"];

  [unarchiver decodeInt:&ver.off named:@"ver.off"];
  [unarchiver decodeInt:&ver.length named:@"ver.length"];

  [unarchiver decodeInt:&nor.off named:@"nor.off"];
  [unarchiver decodeInt:&nor.length named:@"nor.length"];

  [unarchiver decodeInt:&tex.off named:@"tex.off"];
  [unarchiver decodeInt:&tex.length named:@"tex.length"];

  [unarchiver decodeInt:&vertices named:@"vertices"];
  [unarchiver decodeInt:&stride named:@"stride"];
  [unarchiver decodeInt:&format named:@"format"];
  [unarchiver decodeInt:&geometryMode named:@"geometryMode"];
  [unarchiver decodeInt:&_vChunkSize named:@"_vChunkSize"];

  [unarchiver decodeBOOL:&_isInitialised named:@"_isInitialised"];
  [unarchiver decodeBOOL:&isLocked named:@"isLocked"]; // ?

#warning decodeBytesWithReturnedLength:named: is not yet implemented!
  n=vertices*_vChunkSize*sizeof(float);
  //array = (float*)[unarchiver decodeBytesWithReturnedLength:&n];

  return self;
}

@end






