//-----------------------------------------------------------------------------
// Project            	RenderKit
// Class		G3DLODGroup
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Thu Sep  9 11:05:08 CEST 1999
//
// Copyright (c) Philippe C.D. Robert
//
// The RenderKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DLODGroup.m,v 1.19 2001/07/15 12:39:24 probert Exp $
//-----------------------------------------------------------------------------

#import "G3DLODGroup.h"
#import "G3DTransform.h"
#import "G3DFrustum.h"
#import "G3DBVol.h"
#import "G3DCamera.h"
#import "G3DArchiver.h"
#import "G3DUnarchiver.h"

#import <GeometryKit/G3DDefs.h>

NSString *G3DLODGroupException = @"G3DLODGroupException";

@implementation G3DLODGroup

//-----------------------------------------------------------------------------
// Init and free
//-----------------------------------------------------------------------------

- (id)initWithCapacity:(unsigned int)cap maxDistance:(float)val
{
  NSAssert(val>0.0f,@"Invalid max. distance value!");

  if ((self = [super initWithCapacity:cap])) {
    maxDistance = val;
    mode = G3D_Linear;
  }
  return self;
}

//-----------------------------------------------------------------------------
// Rendering and traversing the scenegraph
//-----------------------------------------------------------------------------

- (void)renderWithMask:(unsigned)rmask frustum:(G3DFrustum *)fr
{
  G3D_CameraPosition *cam_pos;

  float _curr_distance = 0;
  float _curr_pos_cam[3];
  float _curr_pos_obj[3];
  float _tmp_dist[3];
  float _dist_step;
  float *matrix;

  if (!isVisible) {
    return;
  }

  if ((rmask & (1 << VIEW_FRUSTUM_CULLING))) {
    if (_shouldVFCull && [fr containsVolume:bVolume] == G3DOutside) {
      DLOG(@"View-Frustum culling: Node and its children won't be rendered");
      return;
    }
  }

  /*
   * Evaluate the correct index based on the distance of the object
   * to the camera - and then render the correct child. 
   */

  switch (mode) {
    case G3D_Linear:

      cam_pos = [[G3DCamera activeCamera] position];
      _curr_pos_cam[0] = (float)cam_pos->pos[0];
      _curr_pos_cam[1] = (float)cam_pos->pos[1];
      _curr_pos_cam[2] = (float)cam_pos->pos[2];

      matrix = [_fwdT elements];
      _curr_pos_obj[0] = matrix[12];
      _curr_pos_obj[1] = matrix[13];
      _curr_pos_obj[2] = matrix[14];

      G3DSubVectors3fv(_tmp_dist,_curr_pos_cam,_curr_pos_obj);

      _curr_distance = G3DLength3fv(_tmp_dist);	// current distance to 0
      _dist_step = maxDistance/(_children-1); 	// delta distance
      index = FLOOR(_curr_distance/_dist_step);	// next index
      index = (index >= _children) ? _children - 1 : index;
      break;

    case G3D_Exp:
      [NSException raise:G3DLODGroupException format:@"Exponential LoD is not yet implemented!"];
      break;

    default:
      [NSException raise:G3DLODGroupException format:@"Unsupported LoD mode!"];
      break;
  }  

  DLOG(@"LOD node %x renders child %i (distance %f, max %f)",self,index,_curr_distance,maxDistance);

  G3D_PUSH_MATRIX;

  // Render! (if _isCompiled and *current* nodeList -> optimise!)

  if (!_isCompiled) {
    [transformation transform];
  }

  // Do not cull the children here!
  if (rmask & (1<<VIEW_FRUSTUM_CULLING)) {
    rmask &= ~(1<<VIEW_FRUSTUM_CULLING);
  }

  [[children objectAtIndex:index] renderWithMask:rmask frustum:fr];      
  
  G3D_POP_MATRIX;
}

//-----------------------------------------------------------------------------
// Level of detail
//-----------------------------------------------------------------------------

- (void)setMaxDistance:(float)val
{
  NSAssert(val>0.0f,@"Invalid max. distance value!");

  maxDistance = val;
}

- (float)maxDistance
{
  return maxDistance;
}

- (void)setLodMode:(G3D_LODMode)md
{
  mode = md;
}

- (G3D_LODMode)lodMode
{
  return mode;
}

//-----------------------------------------------------------------------------
// NSCoding
//-----------------------------------------------------------------------------

- (void)encodeWithCoder:(NSCoder *)aCoder
{
  [super encodeWithCoder:aCoder];

  [aCoder encodeValueOfObjCType:@encode(float) at:&maxDistance];
  [aCoder encodeValueOfObjCType:@encode(G3D_LODMode) at:&mode];
}

- (id)initWithCoder:(NSCoder *)aCoder
{
  self = [super initWithCoder:aCoder];

  [aCoder decodeValueOfObjCType:@encode(float) at:&maxDistance];
  [aCoder decodeValueOfObjCType:@encode(G3D_LODMode) at:&mode];

  return self;
}

//-----------------------------------------------------------------------------
// NSCopying
//-----------------------------------------------------------------------------

- (id)copyWithZone:(NSZone *)zone
{
  id copy = [super copyWithZone:zone];

  [copy setMaxDistance:maxDistance];
  
  return copy;
}

//-----------------------------------------------------------------------------
// G3DSceneCoding
//-----------------------------------------------------------------------------

- (void)encodeWithSceneArchiver:(G3DArchiver *)archiver
{
  [super encodeWithSceneArchiver:archiver];

  [archiver encodeFloat:maxDistance named:@"maxDistance"];
  [archiver encodeInt:mode named:@"mode"];
}

- (id)initWithSceneUnarchiver:(G3DUnarchiver *)unarchiver
{
  self = [super initWithSceneUnarchiver:unarchiver];
  
  [unarchiver decodeFloat:&maxDistance named:@"maxDistance"];
  [unarchiver decodeInt:&mode named:@"mode"];

  return self;
}

@end






