//-----------------------------------------------------------------------------
// Project            	RenderKit
// Class		G3DLight
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Thu Sep  9 11:05:08 CEST 1999
//
// Copyright (c) Philippe C.D. Robert
//
// The RenderKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DLight.m,v 1.18 2001/01/05 20:50:52 robert Exp $
//-----------------------------------------------------------------------------

#import <GeometryKit/G3DVector3f.h>

#import "G3DLight.h"
#import "G3DDefines.h"
#import "G3DColour.h"
#import "G3DTransform.h"
#import "G3DArchiver.h"
#import "G3DUnarchiver.h"

NSString *G3DLightException = @"G3DLightException";

@implementation G3DLight

//-----------------------------------------------------------------------------
// Class methods
//-----------------------------------------------------------------------------

+ (G3DLight *)distantLightWithDirection:(G3DVector3f *)dir
{
  G3DLight *light = [[[self alloc] init] autorelease];

  light->lightType = G3D_DistantLight;
  [light setPosition:dir];

  return light;
}

+ (G3DLight *)pointLightWithPosition:(G3DVector3f *)p
{
  G3DLight *light = [[[self alloc] init] autorelease];

  light->lightType = G3D_PointLight;
  [light setPosition:p];

  light->attenuation = 1.5f;
  light->linearAttenuation = 0.5f;
  light->quadraticAttenuation = 0.2f;

  return light;
}

+ (G3DLight *)spotLightWithPosition:(G3DVector3f *)pos direction:(G3DVector3f *)d
{
  G3DLight *light = [[[self alloc] init] autorelease];

  light->lightType = G3D_SpotLight;
  [light setDirection:d];
  [light setPosition:pos];

  light->attenuation = 1.5f;
  light->linearAttenuation = 0.5f;
  light->quadraticAttenuation = 0.2f;
  light->spotExponent = 2.0f;
  light->spotCutOff = 30.0f;
  
  return light;
}

//-----------------------------------------------------------------------------
// Init and free
//-----------------------------------------------------------------------------

- (id)init
{
  if ((self = [super init])) {
    /*
     * Defaults - should be retrieved from OpenGL
     *
     */

    specular = [[G3DColour alloc] initWithR:1.0f g:1.0f b:1.0f a:1.0f];
    ambient = [[G3DColour alloc] initWithR:0.0f g:0.0f b:0.0f a:1.0f];
    diffuse = [[G3DColour alloc] initWithR:1.0f g:1.0f b:1.0f a:1.0f];
    
    isSwitchedOn = YES;
    isGlobal = YES;
    
    attenuation = 1.0f;
    linearAttenuation = 0.0f;
    quadraticAttenuation = 0.0f;
    spotExponent = 0.0f;
    spotCutOff = 180.0f;

    _lhandle = -1;
    _lightHasChanged = YES;

    lightType = G3D_DistantLight;

    _pos[0] = 0.0f;
    _pos[0] = 0.0f;
    _pos[0] = 0.0f;
    _pos[0] = 0.0f;

    _dir[0] = 0.0f;
    _dir[1] = 0.0f;
    _dir[2] = 1.0f;
    _dir[3] = 0.0f;

    _firstTime = YES;
    _hasBVolume = NO;
    _shouldVFCull = NO;
  }
  return self;
}

- (void)dealloc
{
  [specular release];
  [ambient release];
  [diffuse release];
  
  [super dealloc];
}

//-----------------------------------------------------------------------------
// Render and Position
//-----------------------------------------------------------------------------

- (void)renderWithMask:(unsigned)mask frustum:(G3DFrustum *)fr
{
  if (isGlobal == NO) {
    DLOG(@"Rendering %@ %x (is local!) of type %d at %f, %f, %f",[self class],self,lightType,_pos[0],_pos[1],_pos[2]);

  }
}

- (void)setDirection:(G3DVector3f *)dir
{
  float *d = [dir elements];

  NSAssert(lightType != G3D_PointLight,@"Pointlights don't have a direction!");

  G3DNormalise3fv(_dir,d);
  _dir[3] = 0.0f;

  if (lightType == G3D_DistantLight) {
    isGlobal = YES;

    // No attenuation is to be used!
    attenuation = 1.0f;
    linearAttenuation = 0.0f;
    quadraticAttenuation = 0.0f;
  }

  G3DMatrix4fXVector4f(_trnDir,(_fwdT)->matrix,_dir);

  _lightHasChanged = YES;
}

- (void)getDirection:(G3DVector3f **)dir
{
  [*dir setElements:_dir];
}

- (void)setPosition:(G3DVector3f *)pos
{
  float *p = [pos elements];

  G3DCopyVector3fv(_pos,p);

  if (lightType != G3D_DistantLight) {
    // Positional light, set w to 1.0 (homogenous coordinates)
    _pos[3] = 1.0f;
  }
  else {
    // Directional light, set w to 0.0 (homogenous coordinates)
    _pos[3] = 0.0f;
  }

  // Correct?
  G3DMatrix4fXVector4f(_trnPos,(_fwdT)->matrix,_pos);
  G3DCopyVector3fv(_trnPos,_pos);

  [transformation setTranslationValues:_pos];

  _lightHasChanged = YES;
}

- (void)getPosition:(G3DVector3f **)pos
{
  [*pos setElements:_pos];
}

- (void)updateIfNeeded
{
  [super updateIfNeeded];

  if (_lightHasChanged && isSwitchedOn) {
    if (_lhandle != -1) {
      G3DUpdateLight(self);
    }
    else {
      G3DEnableLight(self);
    }
    _lightHasChanged = NO;    
  }
  else if (_firstTime) {
    _firstTime = NO;

    G3DEnableLight(self);
  }
}

-(void)updateWithTransformation:(G3DMatrix4f *)t
{
  [super updateWithTransformation:t];

  DLOG(@"Updating light %@",[self description]);

  // Update the direction/position vector!
  G3DMatrix4fXVector4f(_trnDir,(_fwdT)->matrix,_dir);
  G3DMatrix4fXVector4f(_trnPos,(_fwdT)->matrix,_pos);

  // It has changes, since we updated the direction vector...
  if (isSwitchedOn) {
    G3DUpdateLight(self);
    _lightHasChanged = NO;
  }
}

- (void)updateBVolume
{
  DLOG(@"<%@ %x>: no bounding volume needed!",[self class],self);
}

//-----------------------------------------------------------------------------
// Accessor methods
//-----------------------------------------------------------------------------

- (int)lightHandle
{
  return _lhandle;
}

- (G3D_LightType)lightType
{
  return lightType;
}

- (BOOL)isSwitchedOn
{
  return isSwitchedOn;
}

- (void)setIsSwitchedOn:(BOOL)flag
{
  isSwitchedOn = flag;

  if (!isSwitchedOn) {
    G3DDisableLight(self);
  }
  else {
    G3DEnableLight(self);
  }  
}

- (BOOL)isGlobal
{
  return isGlobal;
}

- (void)setIsGlobal:(BOOL)flag
{
  if (parent) {
    [NSException raise:G3DLightException format:@"Light is already bound!"];
    return;
  }

  isGlobal = flag;
  _lightHasChanged = YES;
}

- (G3DColour *)specular
{
  return specular;
}

- (void)setSpecular:(G3DColour *)aColour
{
  [specular autorelease];
  specular = [aColour retain];

  _lightHasChanged = YES;
}

- (G3DColour *)diffuse
{
  return diffuse;
}

- (void)setDiffuse:(G3DColour *)aColour
{
  [diffuse autorelease];
  diffuse = [aColour retain];

  _lightHasChanged = YES;
}

- (G3DColour *)ambient
{
  return ambient;
}

- (void)setAmbient:(G3DColour *)aColour
{
  [ambient autorelease];
  ambient = [aColour retain];

  _lightHasChanged = YES;
}

- (float)spotExponent
{
  return spotExponent;
}

- (void)setSpotExponent:(float)exp
{
  spotExponent = exp;

  _lightHasChanged = YES;
}

- (float)spotCutOff
{
  return spotCutOff;
}

- (void)setSpotCutOff:(float)cutOff
{
  spotCutOff = cutOff;

  _lightHasChanged = YES;
}

- (float)attenuation
{
  return attenuation;
}

- (void)setAttenuation:(float)val
{
  attenuation = val;

  _lightHasChanged = YES;
}

- (float)linearAttenuation
{
  return linearAttenuation;
}

- (void)setLinearAttenuation:(float)val
{
  linearAttenuation = val;

  _lightHasChanged = YES;
}

- (float)quadraticAttenuation
{
  return quadraticAttenuation;
}

- (void)setQuadraticAttenuation:(float)val
{
  quadraticAttenuation = val;

  _lightHasChanged = YES;
}

//-----------------------------------------------------------------------------
// NSCoding
//-----------------------------------------------------------------------------

- (void)encodeWithCoder:(NSCoder *)aCoder
{
  [super encodeWithCoder:aCoder];

  [aCoder encodeValueOfObjCType:@encode(float) at:&attenuation];
  [aCoder encodeValueOfObjCType:@encode(float) at:&linearAttenuation];
  [aCoder encodeValueOfObjCType:@encode(float) at:&quadraticAttenuation];
  [aCoder encodeValueOfObjCType:@encode(float) at:&spotExponent];
  [aCoder encodeValueOfObjCType:@encode(float) at:&spotCutOff];
  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&isSwitchedOn];
  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&isGlobal];
  [aCoder encodeValueOfObjCType:@encode(G3D_LightType) at:&lightType];
  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&_lightHasChanged];
 
  [aCoder encodeObject:specular];
  [aCoder encodeObject:diffuse];
  [aCoder encodeObject:ambient];
}

- (id)initWithCoder:(NSCoder *)aCoder
{
  self = [super initWithCoder:aCoder];

  [aCoder decodeValueOfObjCType:@encode(float) at:&attenuation];
  [aCoder decodeValueOfObjCType:@encode(float) at:&linearAttenuation];
  [aCoder decodeValueOfObjCType:@encode(float) at:&quadraticAttenuation];
  [aCoder decodeValueOfObjCType:@encode(float) at:&spotExponent];
  [aCoder decodeValueOfObjCType:@encode(float) at:&spotCutOff];
  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&isSwitchedOn];
  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&isGlobal];
  [aCoder decodeValueOfObjCType:@encode(G3D_LightType) at:&lightType];
  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&_lightHasChanged];

  specular = [[aCoder decodeObject] retain];
  diffuse = [[aCoder decodeObject] retain];
  ambient = [[aCoder decodeObject] retain];
  
  // It cannot be in use...
  self->_lhandle = -1;

  return self;
}

//-----------------------------------------------------------------------------
// NSCopying
//-----------------------------------------------------------------------------

- (id)copyWithZone:(NSZone *)zone
{
  G3DLight *copy = [super copyWithZone:zone];

  [copy setIsGlobal:isGlobal];
  [copy setSpecular:specular];
  [copy setDiffuse:diffuse];
  [copy setAmbient:ambient];
  [copy setSpotExponent:spotExponent];
  [copy setSpotCutOff:spotCutOff];
  [copy setAttenuation:attenuation];
  [copy setLinearAttenuation:linearAttenuation];
  [copy setQuadraticAttenuation:quadraticAttenuation];

  copy->lightType = lightType;
  
  // Turn it off...
  copy->_lightHasChanged = YES;
  copy->_lhandle = -1;
  copy->isSwitchedOn = NO;
  
  return copy;
}

//-----------------------------------------------------------------------------
// G3DSceneCoding
//-----------------------------------------------------------------------------

- (void)encodeWithSceneArchiver:(G3DArchiver *)archiver
{
  [super encodeWithSceneArchiver:archiver];

  [archiver encodeFloat:attenuation named:@"attenuation"];
  [archiver encodeFloat:linearAttenuation named:@"linearAttenuation"];
  [archiver encodeFloat:quadraticAttenuation named:@"quadraticAttenuation"];
  [archiver encodeFloat:spotExponent named:@"spotExponent"];
  [archiver encodeFloat:spotCutOff named:@"spotCutOff"];

  [archiver encodeBOOL:isSwitchedOn named:@"isSwitchedOn"];
  [archiver encodeBOOL:isGlobal named:@"isGlobal"];
  [archiver encodeInt:lightType named:@"lightType"];
  [archiver encodeBOOL:_lightHasChanged named:@"_lightHasChanged"];
 
  [archiver encodeObject:specular named:@"specular"];
  [archiver encodeObject:diffuse named:@"diffuse"];
  [archiver encodeObject:ambient named:@"ambient"];
}

- (id)initWithSceneUnarchiver:(G3DUnarchiver *)unarchiver
{
  self = [super initWithSceneUnarchiver:unarchiver];

  [unarchiver decodeFloat:&attenuation named:@"attenuation"];
  [unarchiver decodeFloat:&linearAttenuation named:@"linearAttenuation"];
  [unarchiver decodeFloat:&quadraticAttenuation named:@"quadraticAttenuation"];
  [unarchiver decodeFloat:&spotExponent named:@"spotExponent"];
  [unarchiver decodeFloat:&spotCutOff named:@"spotCutOff"];
  [unarchiver decodeBOOL:&isSwitchedOn named:@"isSwitchedOn"];
  [unarchiver decodeBOOL:&isGlobal named:@"isGlobal"];
  [unarchiver decodeInt:&lightType named:@"lightType"];
  [unarchiver decodeBOOL:&_lightHasChanged named:@"_lightHasChanged"];

  specular = [[unarchiver decodeObjectNamed:@"specular"] retain];
  diffuse = [[unarchiver decodeObjectNamed:@"diffuse"] retain];
  ambient = [[unarchiver decodeObjectNamed:@"ambient"] retain];
  
  // It cannot be in use...
  self->_lhandle = -1;

  return self;
}

@end







