//-----------------------------------------------------------------------------
// Project            	RenderKit
// Class		G3DMaterial
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Thu Sep  9 11:05:08 CEST 1999
//
// Copyright (c) Philippe C.D. Robert
//
// The RenderKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DMaterial.m,v 1.13 2001/01/05 20:50:52 robert Exp $
//-----------------------------------------------------------------------------

#import "G3DMaterial.h"

#import "G3DColour.h"
#import "G3DState.h"
#import "G3DFunctions.h"
#import "G3DArchiver.h"
#import "G3DUnarchiver.h"

#define ENABLE(x) if (state) enableState(state,x)

@implementation G3DMaterial

//-----------------------------------------------------------------------------
// Class methods
//-----------------------------------------------------------------------------

+ (G3DMaterial *)materialWithState:(G3DState *)st
{
  return [[[G3DMaterial alloc] initWithState:st] autorelease];
}

//-----------------------------------------------------------------------------
// Init and free
//-----------------------------------------------------------------------------

- (id)initWithState:(G3DState *)aState
{
  if ((self = [super init])) {
    [self setName:@"Generic material"];

    if (aState) {
      state = aState;
    }

    /*
     * I use OpenGL default values, so I don't have to set them!
     * LAter on I will retrieve those values from OpenGL
     *
     */
    
    ambient = [[G3DColour alloc] initWithR:0.2 g:0.2 b:0.2 a:1.0];
    //ENABLE(G3D_AMBIENT);

    emissive = [[G3DColour alloc] initWithR:0.0 g:0.0 b:0.0 a:1.0];
    //ENABLE(G3D_EMISSION);

    diffuse = [[G3DColour alloc] initWithR:0.8 g:0.8 b:0.8 a:1.0];
    //ENABLE(G3D_DIFFUSE);

    specular = [[G3DColour alloc] initWithR:0.0 g:0.0 b:0.0 a:1.0];
    //ENABLE(G3D_SPECULAR);
    
    shininess = 0.0f;
    //ENABLE(G3D_SHININESS);
     
    face = G3D_Front;

    materialMode = G3D_AmbientDiffuse;
    //ENABLE(G3D_COLOR_MATERIAL);
  }
  return self;
}

- (void)dealloc
{
  [ambient release];
  [emissive release];
  [specular release];
  [diffuse release];

  [super dealloc];
}

//-----------------------------------------------------------------------------
// Accessor methods
//-----------------------------------------------------------------------------

- (void)setState:(G3DState *)aState
{
  state = aState;
}

- (G3DState *)state
{
  return state;
}

- (void)setAmbientColour:(G3DColour *)aCol
{
  [ambient autorelease];
  ambient = [aCol copy];

  ENABLE(G3D_AMBIENT);
}

- (G3DColour *)ambientColour
{
  return ambient;
}

- (void)setEmissiveColour:(G3DColour *)eCol
{
  [emissive autorelease];
  emissive = [eCol copy];

  ENABLE(G3D_EMISSION);
}

- (G3DColour *)emissiveColour
{
  return emissive;
}

- (void)setDiffuseColour:(G3DColour *)dCol
{
  [diffuse autorelease];
  diffuse = [dCol copy];

  ENABLE(G3D_DIFFUSE);
}

- (G3DColour *)diffuseColour
{
  return diffuse;
}

- (void)setSpecularColour:(G3DColour *)sCol
{
  [specular autorelease];
  specular = [sCol copy];

  ENABLE(G3D_SPECULAR);
}

- (G3DColour *)specularColour
{
  return specular;
}

- (void)setShininess:(double)sh
{
  shininess = sh;

  ENABLE(G3D_SHININESS);
}

- (double)shininess
{
  return shininess;
}

- (void)setMaterialMode:(G3D_MaterialMode)mode
{
  materialMode = mode;

  ENABLE(G3D_COLOR_MATERIAL);
}

- (G3D_MaterialMode)materialMode
{
  return materialMode;
}

- (void)setFace:(G3D_MaterialFace)fc
{
  face = fc;
}

- (G3D_MaterialFace)face
{
  return face;
}

- (NSString *)description
{
  return [NSString stringWithFormat:@"<%@ %x, thread %x>:\n ambient %@\n, emissive %@\n, diffuse %@\n, specular %@\n shininess %f, material mode %d, face %d", [self class], self, [NSThread currentThread], [ambient description],[emissive description],[diffuse description],[specular description],shininess,materialMode,face];
}

//-----------------------------------------------------------------------------
// NSCoding
//-----------------------------------------------------------------------------

- (void)encodeWithCoder:(NSCoder *)aCoder
{
  [super encodeWithCoder:aCoder];

  [aCoder encodeObject:ambient];
  [aCoder encodeObject:emissive];
  [aCoder encodeObject:diffuse];
  [aCoder encodeObject:specular];
  [aCoder encodeValueOfObjCType:@encode(double) at:&shininess];
  [aCoder encodeValueOfObjCType:@encode(int) at:&materialMode];
  [aCoder encodeValueOfObjCType:@encode(int) at:&face];
}

- (id)initWithCoder:(NSCoder *)aCoder
{
  self = [super initWithCoder:aCoder];
  
  ambient = [[aCoder decodeObject] retain];
  emissive = [[aCoder decodeObject] retain];
  diffuse = [[aCoder decodeObject] retain];
  specular = [[aCoder decodeObject] retain];
  [aCoder decodeValueOfObjCType:@encode(double) at:&shininess];
  [aCoder decodeValueOfObjCType:@encode(int) at:&materialMode];
  [aCoder decodeValueOfObjCType:@encode(int) at:&face];

  return self;
}

//-----------------------------------------------------------------------------
// NSCopying
//-----------------------------------------------------------------------------

- (id)copyWithZone:(NSZone *)zone
{
  G3DMaterial *copy = [super copyWithZone:zone];

  copy->state = state;

  [copy setFace:[self face]];
  [copy setMaterialMode:[self materialMode]];
  [copy setShininess:[self shininess]];
  [copy setAmbientColour:ambient];
  [copy setEmissiveColour:emissive];
  [copy setDiffuseColour:diffuse];
  [copy setSpecularColour:specular];
  
  return copy;
}

//-----------------------------------------------------------------------------
// G3DSceneCoding
//-----------------------------------------------------------------------------

- (void)encodeWithSceneArchiver:(G3DArchiver *)archiver
{
  [super encodeWithSceneArchiver:archiver];

  [archiver encodeObject:ambient named:@"ambient"];
  [archiver encodeObject:emissive named:@"emissive"];
  [archiver encodeObject:diffuse named:@"diffuse"];
  [archiver encodeObject:specular named:@"specular"];

  [archiver encodeDouble:shininess named:@"shininess"];
  [archiver encodeInt:materialMode named:@"materialMode"];
  [archiver encodeInt:face named:@"face"];
}

- (id)initWithSceneUnarchiver:(G3DUnarchiver *)unarchiver
{
  self = [super initWithSceneUnarchiver:unarchiver];

  ambient = [[unarchiver decodeObjectNamed:@"ambient"] retain];
  emissive = [[unarchiver decodeObjectNamed:@"emissive"] retain];
  diffuse = [[unarchiver decodeObjectNamed:@"diffuse"] retain];
  specular = [[unarchiver decodeObjectNamed:@"specular"] retain];

  [unarchiver decodeDouble:&shininess named:@"shininess"];
  [unarchiver decodeInt:&materialMode named:@"materialMode"];
  [unarchiver decodeInt:&face named:@"face"];
  
  return self;
}

@end






