//-----------------------------------------------------------------------------
// Project            	RenderKit
// Class		G3DNodeList
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Tue May  2 15:34:06 CEST 2000
//
// Copyright (c) Philippe C.D. Robert
//
// The RenderKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DNodeList.m,v 1.5 2001/01/05 20:50:53 robert Exp $
//-----------------------------------------------------------------------------

#import <RenderKit/G3DNodeList.h>

#import <RenderKit/G3DNodeEnumerator.h>
#import <RenderKit/G3DGraphNode.h>

NSString *G3DNodeListAccessException = @"G3DNodeListAccessException";

@implementation G3DNodeList

//-----------------------------------------------------------------------------
// Init and free
//-----------------------------------------------------------------------------

- (id)initWithNodes:(NSArray *)nds
{
  int i;

  self = [self initWithCapacity:[nds count]];

  for (i=0;i<[nds count];i++) {
    nodes[i] = [[nds objectAtIndex:i] retain];
  }
  _cnt = [nds count];
  _curr_ind = 0;

  return self;
}

- (id)initWithCapacity:(unsigned int)cap
{
  NSAssert(cap>0,@"Capacity is not valid!");

  if ((self = [super init])) {
    int i;

    _cnt = 0;
    _cap = cap;
    _curr_ind = -1;

    nodes = NSZoneMalloc([self zone], sizeof(id)*cap);
    for(i=0; i < cap; i++) {
      nodes[i] = nil;
    }
  }
  return self;
}

- (void)dealloc
{
  int i;

  for (i = 0; i < _cnt; i++) {
    [nodes[i] release];
  }
  NSZoneFree([self zone], nodes);
  
  [super dealloc];
}

//-----------------------------------------------------------------------------
// List accessing
//-----------------------------------------------------------------------------

- (void)renderWithMask:(unsigned)mask frustum:(G3DFrustum *)fr
{
  int i;

  for (i=0;i<_cnt;i++) {
    [nodes[i] renderWithMask:mask frustum:fr];
  }
}

- (id*)nodes
{
  return nodes;
}

- (int)count
{
  return _cnt;
}

- (G3DGraphNode *)nextNode
{
  if (_curr_ind == _cnt) {
    _curr_ind = 0;
  }
  return nodes[_curr_ind++];
}

- (unsigned int)indexOfNextNode
{
  return _curr_ind;
}

- (G3DNodeEnumerator *)nodeEnumerator
{
  return [G3DNodeEnumerator nodeEnumeratorWithList:self startIndex:0 stopIndex:(_cnt-1)];
}

//-----------------------------------------------------------------------------
// List handling
//-----------------------------------------------------------------------------

- (void)addNode:(G3DGraphNode *)aNode
{
  if (_cnt == _cap) {
    [NSException raise:G3DNodeListAccessException format:@"List already complete!"];
  }

  nodes[_cnt++] = [aNode retain];
}

- (void)removeNode:(G3DGraphNode *)aNode
{
  int i;
  int _index = -1;

  NSAssert(aNode,@"No valid node!");

  for (i=0;i<_cnt;i++) {
    if (nodes[i] == aNode) {
      _index = i;
      break;
    }
  }

  if (_index != -1) {
    [self removeNodeAtIndex:_index];
  }
}

- (void)removeNodeAtIndex:(unsigned int)index
{
  int i;
  id obj = nodes[index];

  NSAssert(index>=0,@"No valid index!");

  if (index >= _cnt) {
    [NSException raise:G3DNodeListAccessException format:@"Index too high!"];
  }

  for (i=index;i<(_cnt-1);i++) {  
    nodes[i] = nodes[i+1];
  }
  nodes[_cnt-1] = nil;
  [obj autorelease];
  
  _cnt--;
}

- (G3DGraphNode *)nodeAtIndex:(unsigned int)index
{
  if (index >= _cnt) {
    [NSException raise:G3DNodeListAccessException format:@"Index too high!"];
  }

  return (G3DGraphNode *)nodes[index];
}

@end






