//-----------------------------------------------------------------------------
// Project            	RenderKit
// Class		G3DOverlay
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Wed Jun 28 12:38:33 CEST 2000
//
// Copyright (c) Philippe C.D. Robert
//
// The RenderKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DOverlay.m,v 1.17 2001/07/15 12:39:24 probert Exp $
//-----------------------------------------------------------------------------

#import "G3DOverlay.h"

#import "G3DTexture.h"
#import "G3DState.h"
#import "G3DTransform.h"
#import "G3DColour.h"
#import "G3DArchiver.h"
#import "G3DUnarchiver.h"

@implementation G3DOverlay

//-----------------------------------------------------------------------------
// Init and Free
//-----------------------------------------------------------------------------

+ (G3DOverlay *)overlayWithTexture:(G3DTexture *)tx position:(G3DTuple2f*)p
{
  return [[[self alloc] initWithTexture:tx position:p] autorelease];
}

//-----------------------------------------------------------------------------
// Init and Free
//-----------------------------------------------------------------------------

- (id)initWithTexture:(G3DTexture *)tx position:(G3DTuple2f*)p;
{
  if ((self = [super init])) {
    float pv[3];
    
    state = [[G3DState alloc] init];
    [state setShadeMode:G3D_Flat];

    _hasBVolume = NO;
    _renderAsBVol = NO;
    _bVolIsDirty = NO;

    enableBlending = YES;
    _blendCol[0] = 0.0f;
    _blendCol[1] = 0.0f;
    _blendCol[2] = 0.0f;
    _blendCol[3] = 0.5f;

    [state enableFeature:G3D_Culling];
    [state enableFeature:G3D_Texturing];
    [state enableFeature:G3D_Blending];

    [state disableFeature:G3D_ColourMaterial];
    [state disableFeature:G3D_Lighting];
    [state disableFeature:G3D_AlphaTest];

    [p getElements:pv];
    pv[2] = 0.0f;
    
    [self setTexture:tx];
    [self setPositionWithValues:pv];
    
  }
  return self;
}

- (void)dealloc
{
  [state release];
  [texture release];

  [super release];
}

//-----------------------------------------------------------------------------
// Rendering
//-----------------------------------------------------------------------------

- (void)renderWithMask:(unsigned)mask frustum:(G3DFrustum *)fr
{
  if (!texture) {
    DLOG(@"No texture in %@ %x...",[self class],self);
    return;
  }
  else {
    G3D_PUSH_MATRIX;
        
    [transformation transform];
    [state activateIfNeeded];

    if (enableBlending) {    
      glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
      glColor4f(_blendCol[0],_blendCol[1],_blendCol[2],_blendCol[3]);
    }
    
    glBegin(GL_QUADS);
    glTexCoord2f(0,0);
    glVertex2i(0,0);
    glTexCoord2f(1,0);
    glVertex2i(_w,0);
    glTexCoord2f(1,1);
    glVertex2i(_w,_h);
    glTexCoord2f(0,1);
    glVertex2i(0,_h);
    glEnd();
        
    G3D_POP_MATRIX;
  }
}

//-----------------------------------------------------------------------------
// Accessor methods
//-----------------------------------------------------------------------------

- (void)setTexture:(G3DTexture *)tex
{
  [texture autorelease];
  texture = [tex retain];

  if (texture) {
    G3DTexGeometry *tg;

    [texture bindTextureIfNeeded];
    
    [state setTextureHandle:[texture handle]];

    tg = [texture texGeometry];

    if ([texture blendColour]) {
      G3DCopyVector4fv(_blendCol,[[texture blendColour] values]);
    }

    _w = (*tg).width;
    _h = (*tg).height;
  }
  else {
    [state setTextureHandle:0];

    _blendCol[0] = 0.0f;
    _blendCol[1] = 0.0f;
    _blendCol[2] = 0.0f;
    _blendCol[3] = 0.5f;

    _w = 0;
    _h = 0;
  }
}

- (G3DTexture *)texture
{
  return texture;
}

- (void)setEnableBlending:(BOOL)yn
{
  enableBlending = yn;

  if (yn) {
    [state enableFeature:G3D_Blending];
  }
  else {
    [state disableFeature:G3D_Blending];
  }
}

- (BOOL)enableBlending
{
  return [state isFeatureEnabled:G3D_Blending];
}

- (NSString *)description
{
  return [NSString stringWithFormat:@"%@ %x, thread %x, named %@",[self class], self, [NSThread currentThread],name];
}

//-----------------------------------------------------------------------------
// NSCoding
//-----------------------------------------------------------------------------

- (void)encodeWithCoder:(NSCoder *)aCoder
{
  [super encodeWithCoder:aCoder];  

  [aCoder encodeObject:state];
  [aCoder encodeObject:texture];

  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&enableBlending];
  [aCoder encodeValueOfObjCType:@encode(int) at:&_w];
  [aCoder encodeValueOfObjCType:@encode(int) at:&_h];
  [aCoder encodeArrayOfObjCType:@encode(float) count:4 at:_blendCol];
}

- (id)initWithCoder:(NSCoder *)aCoder
{
  self = [super initWithCoder:aCoder];

  state = [[aCoder decodeObject] retain];
  texture = [[aCoder decodeObject] retain];

  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&enableBlending];
  [aCoder decodeValueOfObjCType:@encode(int) at:&_w];
  [aCoder decodeValueOfObjCType:@encode(int) at:&_h];
  [aCoder decodeArrayOfObjCType:@encode(float) count:4 at:_blendCol];

  return self;
}

//-----------------------------------------------------------------------------
// NSCopying
//-----------------------------------------------------------------------------

- (id)copyWithZone:(NSZone *)zone
{
  G3DOverlay *copy = [super copyWithZone:zone];
  
  [copy setTexture:texture];
  [copy setEnableBlending:enableBlending];

  copy->state = [state copy];

  copy->_w = _w;
  copy->_h = _h;

  G3DCopyVector4fv(copy->_blendCol,_blendCol);

  return copy;
}

//-----------------------------------------------------------------------------
// G3DSceneCoding
//-----------------------------------------------------------------------------

- (void)encodeWithSceneArchiver:(G3DArchiver *)archiver
{
  [super encodeWithSceneArchiver:archiver];

  [archiver encodeObject:state named:@"state"];
  [archiver encodeObject:texture named:@"texture"];

  [archiver encodeBOOL:enableBlending named:@"enableBlending"];

  [archiver encodeInt:_w named:@"_w"];
  [archiver encodeInt:_h named:@"_h"];

  [archiver encodeArrayOfTypeFloatAt:_blendCol count:4 named:@"_blendCol"];
}

- (id)initWithSceneUnarchiver:(G3DUnarchiver *)unarchiver
{
  self = [super initWithSceneUnarchiver:unarchiver];

  state = [[unarchiver decodeObjectNamed:@"state"] retain];
  texture = [[unarchiver decodeObjectNamed:@"texture"] retain];

  [unarchiver decodeBOOL:&enableBlending named:@"enableBlending"];

  [unarchiver decodeInt:&_w named:@"_w"];
  [unarchiver decodeInt:&_h named:@"_h"];

  [unarchiver decodeArrayOfTypeFloatAt:_blendCol count:4 named:@"_blendCol"];

  return self;
}

@end






