//-----------------------------------------------------------------------------
// Project            	RenderKit
// Class		G3DRenderEngine
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Tue May  2 12:22:38 CEST 2000
//
// Copyright (c) Philippe C.D. Robert
//
// The RenderKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DRenderEngine.m,v 1.30 2001/07/07 15:35:47 probert Exp $
//-----------------------------------------------------------------------------

#import "G3DRenderEngine.h"

#import "G3DFog.h"
#import "G3DLight.h"
#import "G3DColour.h"

#if defined(__G3DUSEOPENAL__)
#include <AL/al.h>
#include <AL/alc.h>
#include <AL/alut.h>
#endif

#if defined(GNUSTEP)
#include <GL/gl.h>
#else
#include <OpenGL/gl.h>
#endif

/*
 *
 *
 */

#define G3DHIGHFREQUENCY 44100
#define G3DLOWFREQUENCY  22050

NSString *G3DRenderEngineInitialisationException = @"G3DRenderEngineInitialisationException";

static NSString *cvsID = @"Version 0.3.1r2";

static short lightingEnabled = 0;
static int lights[MAX_DYN_LIGHTS];
 
/*
 * Miscellaneous
 *
 */

void G3DBindFog(G3DFog *fog);

/*
 * Lighting
 *
 */

void G3DEnableLighting(void);
void G3DDisableLighting(void);
int G3DEnableLight(G3DLight *light);
int G3DUpdateLight(G3DLight *light);
void G3DDisableLight(G3DLight *light);
int G3DCreateLightHandle(void);

/*
 * Miscellaneous
 *
 */

__G3DI__ void G3DBindFog(G3DFog *fog)
{
  int mode = [fog fogMode];
  
  glEnable(GL_FOG);

  glFogi(GL_FOG_MODE,mode);
  glFogf(GL_FOG_DENSITY,[fog density]);
  glFogf(GL_FOG_START,[fog start]);
  glFogf(GL_FOG_END,[fog end]);
  glFogfv(GL_FOG_COLOR,[[fog fogColour] values]);

  if (mode == G3D_EXP_FOG) {
    glHint(GL_FOG_HINT,GL_FASTEST);
  }
  else {
    glHint(GL_FOG_HINT,GL_NICEST);
  }  
}

/*
 * Lighting
 *
 */

__G3DI__ void G3DEnableLighting(void)
{
  if (!lightingEnabled) {
    glEnable(GL_LIGHTING);
    lightingEnabled = 1;
  }  
}

__G3DI__ void G3DDisableLighting(void)
{
  if (lightingEnabled) {
    glDisable(GL_LIGHTING);
    lightingEnabled = 0;
  }
}

__G3DI__ int G3DEnableLight(G3DLight *light)
{
  if (light->isSwitchedOn) {
    int h;

    DLOG(@"Enabling light %@",[light description]);
    
    if (!lightingEnabled) {
      glEnable(GL_LIGHTING);
      lightingEnabled = 1;
    }
    
    if ((h = G3DCreateLightHandle()) == -1) {
      return -1;
    }

    light->_lhandle = h;
    glEnable(GL_LIGHT0+h);

    glLightfv((GL_LIGHT0+h), GL_AMBIENT, [[light ambient] values]);
    glLightfv((GL_LIGHT0+h), GL_DIFFUSE, [[light diffuse] values]);
    glLightfv((GL_LIGHT0+h), GL_SPECULAR,[[light specular] values]);
    
    glLightf((GL_LIGHT0+h), GL_CONSTANT_ATTENUATION,[light attenuation]);
    glLightf((GL_LIGHT0+h), GL_LINEAR_ATTENUATION,[light linearAttenuation]);
    glLightf((GL_LIGHT0+h), GL_QUADRATIC_ATTENUATION,[light quadraticAttenuation]);
    
    glLightfv((GL_LIGHT0+h),GL_POSITION,light->_trnPos);

    if (light->lightType == G3D_SpotLight) {
      glLightfv((GL_LIGHT0+h),GL_SPOT_DIRECTION,light->_trnDir);

      glLightf((GL_LIGHT0+h), GL_SPOT_CUTOFF,[light spotCutOff]);
      glLightf((GL_LIGHT0+h), GL_SPOT_EXPONENT,[light spotExponent]);
    }

#ifdef GL_VERSION_1_2
    glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL, GL_SEPARATE_SPECULAR_COLOR);
#endif
  }
  return 0;
}

__G3DI__ int G3DUpdateLight(G3DLight *light)
{
  int h = light->_lhandle;

  if (lights[h]) {
    DLOG(@"Updating light %@",[light description]);

    glEnable(GL_LIGHT0+h);

    glLightfv((GL_LIGHT0+h), GL_AMBIENT, [[light ambient] values]);
    glLightfv((GL_LIGHT0+h), GL_DIFFUSE, [[light diffuse] values]);
    glLightfv((GL_LIGHT0+h), GL_SPECULAR,[[light specular] values]);
    
    glLightf((GL_LIGHT0+h), GL_CONSTANT_ATTENUATION,[light attenuation]);
    glLightf((GL_LIGHT0+h), GL_LINEAR_ATTENUATION,[light linearAttenuation]);
    glLightf((GL_LIGHT0+h), GL_QUADRATIC_ATTENUATION,[light quadraticAttenuation]);
    
    glLightfv((GL_LIGHT0+h),GL_POSITION,light->_trnPos);

    if (light->lightType == G3D_SpotLight) {
      glLightfv((GL_LIGHT0+h),GL_SPOT_DIRECTION,light->_trnDir);

      glLightf((GL_LIGHT0+h), GL_SPOT_CUTOFF,[light spotCutOff]);
      glLightf((GL_LIGHT0+h), GL_SPOT_EXPONENT,[light spotExponent]);
    }

#ifdef GL_VERSION_1_2
    glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL, GL_SEPARATE_SPECULAR_COLOR);
#endif
  }
  return 0;
}

__G3DI__ void G3DDisableLight(G3DLight *light)
{
  int handle = light->_lhandle;

  if (handle != -1) {
    lights[handle] = 0;
    light->isSwitchedOn = 0;
    light->_lhandle = -1;
  }
}

__G3DI__ int G3DCreateLightHandle(void)
{
  register int i;

  for (i=0;i<MAX_DYN_LIGHTS;i++) {
    if (lights[i] == 0) { lights[i] = 1; return i;}
  }

  return -1;
}

@implementation G3DRenderEngine

static BOOL renderBoundingVolumes;
static G3DRenderEngine *_engine = nil;

//-----------------------------------------------------------------------------
// Class methods
//-----------------------------------------------------------------------------

+ (G3DRenderEngine *)renderEngine
{
  if (!_engine) {
    _engine = [[G3DRenderEngine alloc] init];
  }
  return _engine;
}

//-----------------------------------------------------------------------------
// Information
//-----------------------------------------------------------------------------

- (void)initEngine
{
#if defined(__G3DUSEOPENAL__)
  int attrlist[3];

  attrlist[0] = ALC_FREQUENCY;
  attrlist[1] = G3DHIGHFREQUENCY;
  attrlist[2] = 0;
  
  context_id = alcCreateContext(attrlist);
  if(context_id == NULL) {
    [NSException raise:G3DRenderEngineInitialisationException 
		 format:@"Could not initialise the audio backend!"];
  }

  DLOG(@"<%@ %x>: OpenAL context %x successfully created!",[self class],self,context_id);
#endif

  glDisable(GL_NORMALIZE);

  DLOG([self description]);
}

- (void)cleanup
{
#if defined(__G3DUSEOPENAL__)
  alcDestroyContext(context_id);
#endif

  glDisableClientState(GL_NORMAL_ARRAY);
  glDisableClientState(GL_COLOR_ARRAY);
  glDisableClientState(GL_VERTEX_ARRAY);
  glDisableClientState(GL_TEXTURE_COORD_ARRAY);
}

- (void)setRenderMode:(G3D_RenderMode)mode
{
  if (mode == G3D_RealGeometry) {
    renderBoundingVolumes = NO;
  }
  else {
    renderBoundingVolumes = YES;
  }
}

- (G3D_RenderMode)renderMode
{
  return (renderBoundingVolumes) ? G3D_BoundingVolume : G3D_RealGeometry;
}

- (void)toggleRenderMode
{
  renderBoundingVolumes = (renderBoundingVolumes) ? NO : YES;
}

//-----------------------------------------------------------------------------
// Lighting
//-----------------------------------------------------------------------------

- (void)enableVectorNormalisation
{
  glEnable(GL_NORMALIZE);
}

- (void)disableVectorNormalisation
{
  glDisable(GL_NORMALIZE);
}

- (void)setGlobalAmbientColour:(G3DColour *)aCol
{
  float *values = [aCol values];

  glLightModelfv(GL_LIGHT_MODEL_AMBIENT,values);
}

- (void)setUseInfiniteViewpoint:(BOOL)yn
{
  if (yn) {
    glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,GL_FALSE);
  }
  else {
    glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER,GL_TRUE);
  }
}

- (void)enableTwoSidedLighting
{
  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_TRUE);
}

- (void)disableTwoSidedLighting
{
  glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_FALSE);
}

//-----------------------------------------------------------------------------
// Description
//-----------------------------------------------------------------------------

- (NSString *)description
{
  NSString *str;

  const char *vendor;
  const char *version;
  const char *renderer;
  const char *ext;

  GLint bits;

  str = [NSString stringWithFormat:@"\n\n****************************************\n\n"];

  str = [str stringByAppendingFormat:@"GNU 3DKit (%@)\n\n",cvsID];

  vendor = glGetString(GL_VENDOR);
  str = [str stringByAppendingFormat:@"GL Vendor: %s\n",vendor];

  version = glGetString(GL_VERSION);
  str = [str stringByAppendingFormat:@"GL Version: %s\n",version];

  glGetIntegerv(GL_INDEX_BITS, &bits);
  str = [str stringByAppendingFormat:@"GL Index Bits: %d\n",bits];

  renderer = glGetString(GL_RENDERER);
  str = [str stringByAppendingFormat:@"GL Renderer: %s\n\n",renderer];

  ext = glGetString(GL_EXTENSIONS);
  str = [str stringByAppendingFormat:@"Supported GL Extensions: %s\n\n",ext];

  str = [str stringByAppendingFormat:@"\n****************************************\n\n"];

  return str;
}

@end






