//-----------------------------------------------------------------------------
// Project            	RenderKit
// Class		G3DShape
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Thu Sep  9 11:05:08 CEST 1999
//
// Copyright (c) Philippe C.D. Robert
//
// The RenderKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DShape.m,v 1.39 2001/01/12 19:22:50 robert Exp $
//-----------------------------------------------------------------------------

#import "G3DShape.h"

#import "G3DState.h"
#import "G3DMaterial.h"
#import "G3DTexture.h"
#import "G3DDefines.h"
#import "G3DGroup.h"
#import "G3DTransform.h"
#import "G3DBVol.h"
#import "G3DFrustum.h"
#import "G3DArchiver.h"
#import "G3DUnarchiver.h"

@interface G3DShape (DListGeneration)

- (void)_createDList;
- (void)_createDListIfNeeded;

@end

@implementation G3DShape (DListGeneration)

- (void)_createDList
{
  _dlistIsDirty = NO;
}

- (void)_createDListIfNeeded
{
  if (_dlistIsDirty) {
    [self _createDList];
  }
}

@end

@implementation G3DShape

//-----------------------------------------------------------------------------
// Class methods
//-----------------------------------------------------------------------------

+ (G3DShape *)shape
{
  return [[[G3DShape alloc] initWithTexture:nil geometry:nil] autorelease];
}

+ (G3DShape *)shapeWithTexture:(G3DTexture *)aTex geometry:(id<NSObject,G3DGeometryBackend>)geom
{
  return [[[G3DShape alloc] initWithTexture:aTex geometry:geom] autorelease];
}

//-----------------------------------------------------------------------------
// Init and free
//-----------------------------------------------------------------------------

- (id)init
{
  return [self initWithTexture:nil geometry:nil];
}

- (id)initWithTexture:(G3DTexture *)aTexture geometry:(id<NSObject,G3DGeometryBackend>)geom
{
  if ((self = [super init])) {
    state = [[G3DState alloc] init];

    if (aTexture) {
      [self setTexture:aTexture];
    }

    if (geom) {
      [self setGeometry:geom];
    }

    // Some defaults...
    bVolType = G3D_Sphere;
    
    _loadsLazy = NO;
    _isLoaded = YES;

    _dlistIsDirty = YES;

    [name release];
    name = [[NSString alloc] initWithString:@"Generic Shape"];
  }

  return self;
}

- (void)dealloc
{
  [state release];
  [geometry release];
  
  if (texture) {
    [texture release];
  }

  [super dealloc];
}

//-----------------------------------------------------------------------------
// Rendering
//-----------------------------------------------------------------------------

- (void)renderWithMask:(unsigned)mask frustum:(G3DFrustum *)fr
{
  if (!isVisible || geometry == nil) {
    return;
  }

  if ((mask & (1 << VIEW_FRUSTUM_CULLING))) {
    if (_shouldVFCull && [fr containsVolume:bVolume] == G3DOutside) {
      DLOG(@"View-Frustum culling: Node and its children won't be rendered");
      return;
    }
  }

  if (_shouldNotifyBeforeRendering) {
    [delegate graphNodeWillRender:self];
  }

  /*
   * Either use a DL or do everything manually
   *
   */

  G3D_PUSH_MATRIX;

  if (_useDisplayList) {
    if (_dlistIsDirty) {
      [self _createDList];
    }
    glCallList(dlist);
  }
  else {
    if (_isCompiled) {
      // Check if we have an identity here?!
      glMultMatrixf(_fwdT->matrix);
    }
    else {
      if (![transformation isIdentity]) {
	[transformation transform];
      }
    }

    if (_renderAsBVol || [[G3DRenderEngine renderEngine] renderMode] == G3D_BoundingVolume) {
      [bVolume renderBoundingVolume];
    }
    else {
      [state activateIfNeeded];
      [geometry drawGeometry];
    }    
  }

  if (_shouldNotifyAfterRendering) {
    [delegate graphNodeDidRender:self];
  }

  G3D_POP_MATRIX;
}

//-----------------------------------------------------------------------------
// Optimisation
//-----------------------------------------------------------------------------

- (void)compile
{
  if (!_isCompiled) {
    DLOG(@"<%@ %x> node becomes compiled...",[self class],self);
 
    [geometry precompile];
    _isCompiled = YES;
  }
}

- (void)updateBVolume
{
  /*
  if (_isCompiled) {
    [NSException raise:G3DGraphNodeCompilationException format:@"Cannot modify a compiled node."];
    return;
  }
  */

  if (!_hasBVolume) {
    DLOG(@"%@ %x has no bounding volume - ignoring...",[self class],self);
    _bVolIsDirty = NO;
    return;    
  }
  
  if (!bVolume) {
    bVolume = [[G3DBVol alloc] initWithType:bVolType];

    DLOG(@"%@ %x created new bounding volume %x",[self class],self,bVolume);
  }
  else if (bVolType != [bVolume type]) {
    [bVolume autorelease];
    bVolume = [[G3DBVol alloc] initWithType:bVolType];

    DLOG(@"%@ %x created new bounding volume %x",[self class],self,bVolume);
  }

  [bVolume setNode:self];
  [bVolume updateWithGeometry:geometry];
  _bVolIsDirty = NO;

  DLOG(@"%@ %x updated bounding volume %x",[self class],self,bVolume);

  if (parent != nil) {
    [parent setNeedsUpdate:YES];
  }
}

- (void)updatePriority
{
  renderPriority.value = 0;

  /*
   * First set the appropriate bits
   */

  if (texture && [state isFeatureEnabled:G3D_Texturing]) {
    G3D_ENABLE_PRIORITY(G3D_PRIORITY_TEXTURING);
    renderPriority.value += G3D_PRIORITY_TEXTURING;
    if (texture) {
      renderPriority.value += [texture handle];
    }
  }
  else {
    G3D_DISABLE_PRIORITY(G3D_PRIORITY_TEXTURING);
  }

  if ([state isFeatureEnabled:G3D_ColourMaterial]) {
    G3D_ENABLE_PRIORITY(G3D_PRIORITY_COLMATERIAL);
    renderPriority.value += G3D_PRIORITY_COLMATERIAL;
  }
  else {
    G3D_DISABLE_PRIORITY(G3D_PRIORITY_COLMATERIAL);
  }

  if ([state shadeMode] == G3D_Smooth) {
    G3D_ENABLE_PRIORITY(G3D_PRIORITY_SMOOTH);
    renderPriority.value += G3D_PRIORITY_SMOOTH;
  }
  else {
    G3D_DISABLE_PRIORITY(G3D_PRIORITY_SMOOTH);
  }

  if ([state isFeatureEnabled:G3D_Lighting]) {
    G3D_ENABLE_PRIORITY(G3D_PRIORITY_LIGHTING);
    renderPriority.value += G3D_PRIORITY_LIGHTING;
  }
  else {
    G3D_ENABLE_PRIORITY(G3D_PRIORITY_LIGHTING);
  }

  if ([state isFeatureEnabled:G3D_Blending]) {
    G3D_ENABLE_PRIORITY(G3D_PRIORITY_BLENDING);
  }
  else {
    G3D_ENABLE_PRIORITY(G3D_PRIORITY_BLENDING);
  }

  DLOG(@"<%@ %x> updated render priority (%d)",[self class],self,renderPriority.value);
}

//-----------------------------------------------------------------------------
// Accessor methods
//-----------------------------------------------------------------------------

- (void)setBVolType:(G3D_BVolType)tp
{
  bVolType = tp;

  // Next time we update it!
  if (_hasBVolume) {
    _bVolIsDirty = YES;
  }
}

- (G3D_BVolType)bVolType
{
  return bVolType;
}

- (void)setBoundingVolume:(G3DBVol *)bVol
{
  [bVolume autorelease];
  bVolume = [bVol copy];

  [bVolume setNode:self];

  _hasBVolume = bVolume ? YES : NO;
  _bVolIsDirty = NO;
    
  if (parent) {
    [parent setNeedsUpdate:YES];
  }
}

- (void)setGeometry:(id<NSObject,G3DGeometryBackend>)aGeom
{
  NSAssert(aGeom,@"No valid geometry provided!");

  //~ Cache the drawGeometry selector, here!?
  [geometry autorelease];
  geometry = [aGeom retain];

  _bVolIsDirty = _hasBVolume ? YES : NO;
  _dlistIsDirty = YES;
}

- (id<NSObject,G3DGeometryBackend>)geometry
{
  return geometry;
}

- (void)setTexture:(G3DTexture *)tex
{
  [texture autorelease];
  texture = [tex retain];

  if (texture) {
    [texture bindTextureIfNeeded];
    
    [state setTextureHandle:[texture handle]];
    [self setEnableTexturing:YES];
  }
  else {
    [state setTextureHandle:0];
    [self setEnableTexturing:NO];
  }
}

- (G3DTexture *)texture
{
  return texture;
}

- (void)setState:(G3DState *)aState
{
  NSAssert(aState,@"No valid state provided!");

  [state autorelease];
  state = [aState copy];

  // Cache the active selector!

  _dlistIsDirty = YES;
}

- (G3DState *)state
{
  return state;
}

//-----------------------------------------------------------------------------
// State facade methods
//-----------------------------------------------------------------------------

- (G3DMaterial *)material
{
  return [state material];
}

- (void)setEnableTexturing:(BOOL)flag
{
  if (flag) {
    [state enableFeature:G3D_Texturing];
  }
  else {
    [state disableFeature:G3D_Texturing];
  }
  _dlistIsDirty = YES;
}

- (BOOL)enableTexturing
{
  return [state isFeatureEnabled:G3D_Texturing];
}

- (void)setEnableBackfaceCulling:(BOOL)flag
{
  if (flag) {
    [state enableFeature:G3D_Culling];
  }
  else {
    [state disableFeature:G3D_Culling];
  }
  _dlistIsDirty = YES;
}

- (BOOL)enableBackfaceCulling
{
  return [state isFeatureEnabled:G3D_Culling];
}

- (void)setEnableBlending:(BOOL)flag
{
  if (flag) {
    [state enableFeature:G3D_Blending];
  }
  else {
    [state disableFeature:G3D_Blending];
  }
  _dlistIsDirty = YES;
}

- (BOOL)enableBlending
{
  return [state isFeatureEnabled:G3D_Blending];
}

- (void)setShadeMode:(G3D_ShadeMode)mode
{
  [state setShadeMode:mode];
}

- (G3D_ShadeMode)shadeMode
{
  return [state shadeMode];
}

- (void)setEnableLighting:(BOOL)flag
{
  if (flag) {
    [state enableFeature:G3D_Lighting];
  }
  else {
    [state disableFeature:G3D_Lighting];
  }
  _dlistIsDirty = YES;
}

- (BOOL)enableLighting
{
  return [state isFeatureEnabled:G3D_Lighting];
}

//-----------------------------------------------------------------------------
// NSCopying
//-----------------------------------------------------------------------------

- (id)copyWithZone:(NSZone *)zone
{
  G3DShape *copy = [super copyWithZone:zone];

  [copy setState:state];
  [copy setTexture:texture];

  [copy setGeometry:geometry];
  copy->_dlistIsDirty = _dlistIsDirty;
  copy->_loadsLazy = _loadsLazy;
  copy->_isLoaded = _isLoaded;
  copy->surface = surface;
  
  return copy;
}

//-----------------------------------------------------------------------------
// Encoding/Decoding
//-----------------------------------------------------------------------------

- (void)encodeWithCoder:(NSCoder *)aCoder
{
  [super encodeWithCoder:aCoder];

  [aCoder encodeObject:texture];
  [aCoder encodeObject:geometry];
  [aCoder encodeObject:state];

  [aCoder encodeValueOfObjCType:@encode(int) at:&surface];
  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&_dlistIsDirty];
  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&_loadsLazy];
  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&_isLoaded];
}

- (id)initWithCoder:(NSCoder *)aCoder
{
  self = [super initWithCoder:aCoder];

  texture = [[aCoder decodeObject] retain];
  geometry = [[aCoder decodeObject] retain];
  state = [[aCoder decodeObject] retain];

  [aCoder decodeValueOfObjCType:@encode(int) at:&surface];
  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&_dlistIsDirty];
  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&_loadsLazy];
  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&_isLoaded];

  return self;
}

//-----------------------------------------------------------------------------
// G3DSceneCoding
//-----------------------------------------------------------------------------

- (void)encodeWithSceneArchiver:(G3DArchiver *)archiver
{
  [super encodeWithSceneArchiver:archiver];

  [archiver encodeObject:texture named:@"texture"];
  [archiver encodeObject:geometry named:@"geometry"];
  [archiver encodeObject:state named:@"state"];

  [archiver encodeInt:surface named:@"surface"];
  [archiver encodeBOOL:_dlistIsDirty named:@"_dlistIsDirty"];
  [archiver encodeBOOL:_loadsLazy named:@"_loadsLazy"];
  [archiver encodeBOOL:_isLoaded named:@"_isLoaded"];
}

- (id)initWithSceneUnarchiver:(G3DUnarchiver *)unarchiver
{
  self = [super initWithSceneUnarchiver:unarchiver];

  texture = [[unarchiver decodeObjectNamed:@"texture"] retain];
  geometry = [[unarchiver decodeObjectNamed:@"geometry"] retain];
  state = [[unarchiver decodeObjectNamed:@"state"] retain];

  [unarchiver decodeInt:&surface named:@"surface"];
  [unarchiver decodeBOOL:&_dlistIsDirty named:@"_dlistIsDirty"];
  [unarchiver decodeBOOL:&_loadsLazy named:@"_loadsLazy"];
  [unarchiver decodeBOOL:&_isLoaded named:@"_isLoaded"];

  return self;
}

@end






