//-----------------------------------------------------------------------------
// Project            	RenderKit
// Class		G3DSurface
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Fri Jun 16 13:50:36 CEST 2000
//
// Copyright (c) Philippe C.D. Robert
//
// The RenderKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DSurface.m,v 1.7 2001/07/22 11:04:21 probert Exp $
//-----------------------------------------------------------------------------

#import "G3DSurface.h"
#import <RenderKit/RenderKit.h>

NSString *G3DSurfaceException = @"G3DSurfaceException";

@implementation G3DSurface

//-----------------------------------------------------------------------------
// Class methods
//-----------------------------------------------------------------------------

+ (G3DSurface *)surfaceWithScale:(float)sc
{
  return [[[G3DSurface alloc] initWithScale:sc] autorelease];
}

//-----------------------------------------------------------------------------
// Init and free
//-----------------------------------------------------------------------------

- (id)initWithScale:(float)sc
{
  if ((self = [super init])) {
    _sc = sc;
    _hasDL = NO;
 
    _tResX = 1.0f;
    _tResY = 1.0f;

    G3DCopyVector4f(rgba,1.0f,1.0f,1.0f,1.0f);
  }
  return self;
}

- (void)dealloc
{
  if (_hasDL) {
    glDeleteLists(dl,1);
  }
  
  [super dealloc];
}

//-----------------------------------------------------------------------------
// Geometry backend
//-----------------------------------------------------------------------------

#define RES 2.0f
- (void)setTextureCoordinateX:(float)xval y:(float)yval
{
  _tResX = (xval <= 0.0f) ? 1.0f : xval;
  _tResY = (yval <= 0.0f) ? 1.0f : yval;
}

- (void)drawGeometry
{
  if (!_hasDL) {  
    G3D_PUSH_MATRIX;
    
    glScalef(_sc,1.0f,_sc);

    glColor4fv(rgba);

    glBegin(GL_QUADS);

    glTexCoord2f(0.0f,0.0f);
    glNormal3i(0,1,0);
    glVertex3f(0.0f,0.0f,0.0f);

    glTexCoord2f(0.0f, _tResY);
    glNormal3i(0,1,0);
    glVertex3f(0.0f,0.0f,1.0f);

    glTexCoord2f(_tResX,_tResY);
    glNormal3i(0,1,0);
    glVertex3f(1.0f,0.0f,1.0f);

    glTexCoord2f(_tResX,0.0f);
    glNormal3i(0,1,0);
    glVertex3f(1.0f,0.0f,0.0f);
    
    glEnd();
        
    G3D_POP_MATRIX;
  }
  else {
    glCallList(dl);
  }
}

- (void)precompile
{
  if (!_hasDL) {
#ifdef DEBUG
    GLenum err;
    const GLubyte *str;
#endif DEBUG
    
    dl = glGenLists(1);
    glNewList(dl,GL_COMPILE);

    G3D_PUSH_MATRIX;

    glScalef(_sc,1.0f,_sc);

    glColor4fv(rgba);

    glBegin(GL_QUADS);
    
    glTexCoord2f(0.0f,0.0f);
    glNormal3i(0,1,0);
    glVertex3f(0.0f,0.0f,0.0f);

    glTexCoord2f(0.0f,_tResY);
    glNormal3i(0,1,0);
    glVertex3f(0.0f,0.0f,1.0f);

    glTexCoord2f(_tResX,_tResY);
    glNormal3i(0,1,0);
    glVertex3f(1.0f,0.0f,1.0f);

    glTexCoord2f(_tResX,0.0f);
    glNormal3i(0,1,0);
    glVertex3f(1.0f,0.0f,0.0f);
    
    glEnd();
    
    G3D_POP_MATRIX;    

    glEndList();

    if (!glIsList(dl)){
      _hasDL = NO;
    }
    else {
      _hasDL = YES;
    }
  }
  else {
    DLOG(@"%@ %x already uses a displaylist!",[self class],self);
  }
}

- (int)vertexCount
{
  return -1;
}

- (void)getMin:(float *)min max:(float *)max
{
  min[0] = (float) 0.0f;
  min[1] = (float)-G3DEPSILON;
  min[2] = (float) 0.0f;

  max[0] = (float)1.0f*_sc;
  max[1] = (float)G3DEPSILON;
  max[2] = (float)1.0f*_sc;
}

- (void)setShape:(G3DShape *)aShape
{
  shape = aShape;
}

- (void)setColour:(G3DColour *)col
{
  G3DCopyVector4fv(rgba,[col values]);
}

- (G3DColour *)colour
{
  return [G3DColour colourWithR:rgba[0] g:rgba[1] b:rgba[2] a:rgba[3]];
}

//-----------------------------------------------------------------------------
// NSCoding
//-----------------------------------------------------------------------------

- (void)encodeWithCoder:(NSCoder *)aCoder
{
  [super encodeWithCoder:aCoder];

  [aCoder encodeObject:shape];
  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&_hasDL];
  [aCoder encodeValueOfObjCType:@encode(float) at:&_sc];
  [aCoder encodeArrayOfObjCType:@encode(float) count:4 at:rgba];
  [aCoder encodeValueOfObjCType:@encode(float) at:&_tResX];
  [aCoder encodeValueOfObjCType:@encode(float) at:&_tResY];
}

- (id)initWithCoder:(NSCoder *)aCoder
{
  self = [super initWithCoder:aCoder];

  shape = [[aCoder decodeObject] retain];
  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&_hasDL];
  [aCoder decodeValueOfObjCType:@encode(float) at:&_sc];
  [aCoder decodeArrayOfObjCType:@encode(float) count:4 at:rgba];
  [aCoder decodeValueOfObjCType:@encode(float) at:&_tResX];
  [aCoder decodeValueOfObjCType:@encode(float) at:&_tResY];

  if (_hasDL) {
    [self precompile];
  }

  return self;
}

//-----------------------------------------------------------------------------
// NSCopying
//-----------------------------------------------------------------------------

- (id)copyWithZone:(NSZone *)zone
{
  G3DSurface *copy = [[G3DSurface allocWithZone:zone] init];

  copy->shape  = shape;
  copy->_hasDL = _hasDL;
  copy->_tResX = _tResX;
  copy->_tResY = _tResY;

  G3DCopyVector4fv(copy->rgba,rgba);

  if (_hasDL) {
    copy->dl = dl;
  }

  copy->_sc = _sc;

  return copy;
}

//-----------------------------------------------------------------------------
// G3DSceneCoding
//-----------------------------------------------------------------------------

- (void)encodeWithSceneArchiver:(G3DArchiver *)archiver
{
  [super encodeWithSceneArchiver:archiver];

  [archiver encodeObject:shape named:@"shape"];  

  [archiver encodeFloat:_sc named:@"_sc"];
  [archiver encodeBOOL:_hasDL named:@"_hasDL"];
  [archiver encodeArrayOfTypeFloatAt:rgba count:4 named:@"rgba"];
  [archiver encodeFloat:_tResX named:@"_tResX"];
  [archiver encodeFloat:_tResY named:@"_tResY"];
}

- (id)initWithSceneUnarchiver:(G3DUnarchiver *)unarchiver
{
  self = [super initWithSceneUnarchiver:unarchiver];

  shape = [unarchiver decodeObjectNamed:@"shape"];

  [unarchiver decodeFloat:&_sc named:@"_sc"];
  [unarchiver decodeBOOL:&_hasDL named:@"_hasDL"];
  [unarchiver decodeArrayOfTypeFloatAt:rgba count:4 named:@"rgba"];
  [unarchiver decodeFloat:&_tResX named:@"_tResX"];
  [unarchiver decodeFloat:&_tResY named:@"_tResY"];

  if (_hasDL) {
    [self precompile];
  }

  return self;
}

@end









