//-----------------------------------------------------------------------------
// Project            	RenderKit
// Class		G3DSwitchGroup
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Thu Sep  9 11:05:08 CEST 1999
//
// Copyright (c) Philippe C.D. Robert
//
// The RenderKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DSwitchGroup.m,v 1.21 2001/07/15 12:39:24 probert Exp $
//-----------------------------------------------------------------------------

#import "G3DSwitchGroup.h"
#import "G3DTransform.h"
#import "G3DFunctions.h"
#import "G3DRenderEngine.h"
#import "G3DDefines.h"
#import "G3DBVol.h"
#import "G3DFrustum.h"
#import "G3DArchiver.h"
#import "G3DUnarchiver.h"

@implementation G3DSwitchGroup

//-----------------------------------------------------------------------------
// Init and free
//-----------------------------------------------------------------------------

- (id)initWithCapacity:(unsigned int)cap
{
  if ((self = [super initWithCapacity:cap allowReordering:NO])) {
  }
  return self;
}

//-----------------------------------------------------------------------------
// Rendering and traversing the scenegraph
//-----------------------------------------------------------------------------

- (void)renderWithMask:(unsigned)rmask frustum:(G3DFrustum *)fr
{
  if (!isVisible) {
    return;
  }

  if (_shouldNotifyBeforeRendering) {
    [delegate graphNodeWillRender:self];
  }

  if ((rmask & (1 << VIEW_FRUSTUM_CULLING))) {
    if (_shouldVFCull && [fr containsVolume:bVolume] == G3DOutside) {
      DLOG(@"View-Frustum culling: Node and its children won't be rendered");
      return;
    }
  }

  G3D_PUSH_MATRIX;

  if (!_isCompiled) {
    [transformation transform];
  }

  // Render! (if _isCompiled and *current* nodeList -> optimise!)
  [[children objectAtIndex:index] renderWithMask:rmask frustum:fr];
  
  index++;
  
  if (index == _children) {
    index = 0;
  }
  
  G3D_POP_MATRIX;

  if (_shouldNotifyAfterRendering) {
    [delegate graphNodeDidRender:self];
  }
}

- (BOOL)canBeFlattened
{
  return NO;
}

//-----------------------------------------------------------------------------
// Switching methods
//-----------------------------------------------------------------------------

- (G3DGraphNode *)currentChild
{
  return [children objectAtIndex:index];
}

- (G3DGraphNode *)nextChild
{
  int _i = (index+1);

  if (_i == _children) {
    _i = 0;
  }

  return [children objectAtIndex:_i];
}

- (void)setNextChild:(G3DGraphNode *)aChild
{
  NSAssert([children containsObject:aChild],@"aChild is not a valid child!");

  index = [children indexOfObject:aChild];
}

- (void)setSelectionMask:(int)aMask
{
  mask = aMask;
}

- (int)selectionMask
{
  return mask;
}

//-----------------------------------------------------------------------------
// NSCoding
//-----------------------------------------------------------------------------

- (void)encodeWithCoder:(NSCoder *)aCoder
{
  [super encodeWithCoder:aCoder];

  [aCoder encodeValueOfObjCType:@encode(int) at:&index];
  [aCoder encodeValueOfObjCType:@encode(int) at:&mask];
}

- (id)initWithCoder:(NSCoder *)aCoder
{
  self = [super initWithCoder:aCoder];

  [aCoder decodeValueOfObjCType:@encode(int) at:&index];
  [aCoder decodeValueOfObjCType:@encode(int) at:&mask];

  return self;
}

//-----------------------------------------------------------------------------
// NSCopying
//-----------------------------------------------------------------------------

- (id)copyWithZone:(NSZone *)zone
{
  G3DSwitchGroup *copy = [super copyWithZone:zone];

  [copy setSelectionMask:mask];
  copy->index = index;
  
  return copy;
}

//-----------------------------------------------------------------------------
// G3DSceneCoding
//-----------------------------------------------------------------------------

- (void)encodeWithSceneArchiver:(G3DArchiver *)archiver
{
  [super encodeWithSceneArchiver:archiver];

  [archiver encodeInt:index named:@"index"];
  [archiver encodeInt:mask named:@"mask"];
}

- (id)initWithSceneUnarchiver:(G3DUnarchiver *)unarchiver
{
  self = [super initWithSceneUnarchiver:unarchiver];

  [unarchiver decodeInt:&index named:@"index"];
  [unarchiver decodeInt:&mask named:@"mask"];
  
  return self;
}

@end






