//-----------------------------------------------------------------------------
// Project            	RenderKit
// Class		G3DTexture
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Thu Sep  9 11:05:08 CEST 1999
//
// Copyright (c) Philippe C.D. Robert
//
// The RenderKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DTexture.m,v 1.20 2001/07/22 11:04:21 probert Exp $
//-----------------------------------------------------------------------------

#import "G3DTexture.h"

#import "G3DColour.h"
#import "G3DDefines.h"
#import "G3DArchiver.h"
#import "G3DUnarchiver.h"

#import <GeometryKit/GeometryKit.h>

NSString *G3DTextureScalingException = @"G3DTextureScalingException";
NSString *G3DTextureMipmapException = @"G3DTextureMipmapException";
NSString *G3DTextureImageSizeException = @"G3DTextureImageSizeException";
NSString *G3DTextureFormatException = @"G3DTextureFormatException";
NSString *G3DTextureReadingException = @"G3DTextureReadingException";

#define TGA_OFFSET 18

@interface G3DTexture (TextureLoading)

- (void)readTextureFromFile:(NSString *)aFile;
- (void)readBMPTextureFromFile:(NSString *)aPath;
- (void)readTGATextureFromFile:(NSString *)aPath;
  // Never invoke those methods directly, use the initialiser instead.

@end

@implementation G3DTexture

//-----------------------------------------------------------------------------
// Class methods
//-----------------------------------------------------------------------------

+ (G3DTexture *)textureWithContentsOfFile:(NSString *)aPath
{
  return [[[G3DTexture alloc] initWithContentsOfFile:aPath] autorelease];
}

//-----------------------------------------------------------------------------
// Init and free
//-----------------------------------------------------------------------------

- (id)initWithContentsOfFile:(NSString *)aPath
{
  NSAssert(aPath,@"No valid path passed!");

  if (![[NSFileManager defaultManager] isReadableFileAtPath:aPath]) {
    [self autorelease];
    [NSException raise:G3DTextureReadingException 
		 format:@"%@ not a valid path!",aPath];
    return nil;
  }

  if ((self = [super init])) {

    /*
     * Set up the defaults...
     */

    wrap.s = G3D_REPEAT;
    wrap.t = G3D_REPEAT;

    mag = G3D_MAG_NEAREST;
    min = G3D_MIN_NEAREST;

    mode = G3D_MODULATE;

    ifmt = G3D_RGBA_INT;
    fmt = G3D_RGBA;
    type = G3D_UBYTE;

    _isBound = NO;

    blendcol = [[G3DColour alloc] initWithR:0.0f g:0.0f b:0.0f a:0.0f];
    bordercol = [[G3DColour alloc] initWithR:0.0f g:0.0f b:0.0f a:0.0f];

    _createMipmaps = NO;
    _components = 3;

    _texFileName = [aPath copy];

    /*
     * Load the texture into memory and initialises the flags
     */

    [self readTextureFromFile:_texFileName];
  }
  return self;
}

- (void)dealloc
{
  // Delete the texture obj - check the ref count?
  glDeleteTextures(1,texObj);

  [blendcol release];
  [bordercol release];
  [imageData release];

  [_texFileName release];

  [super dealloc];
}

//-----------------------------------------------------------------------------
// Texture creation
//-----------------------------------------------------------------------------

- (void)bindTextureIfNeeded
{
  if (_isBound) {
    DLOG(@"Texture object %d <%x> is already bound!",texObj[0],self);
    return;
  }

  glPixelStorei(GL_UNPACK_ALIGNMENT,4); // 4 is OpenGL default

  // Generate the texture object
  glGenTextures(1,texObj);
  glBindTexture(GL_TEXTURE_2D,texObj[0]);

  DLOG(@"Setting up the texture object %d <%x>",texObj[0],self);

  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, wrap.s);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, wrap.t);

  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, mag);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, min);

  // Modulation or blending?
  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, mode);
  glTexEnvfv(GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, [blendcol values]);

  /*
   * Using mipmapping means more memory consumption!
   *
   */

  if (_createMipmaps) {
    [NSException raise:G3DTextureMipmapException 
                format:@"Mipmapping not supported in this release!"];
    /*
    if(gluBuild2DMipmaps(GL_TEXTURE_2D, _components, texGeometry.width, texGeometry.height,fmt, type, [imageData bytes]) != 0) {
      [NSException raise:G3DTextureMipmapException format:@"Could not create mipmaps of the specified texture!"];
    }
    */
  }
  else {
    glTexImage2D(GL_TEXTURE_2D, 0, ifmt, texGeometry.width, texGeometry.height, texGeometry.border, fmt, type, [imageData bytes]);
  }

  _isBound = YES;
}

- (unsigned int)handle
{
  return texObj[0];
}

//-----------------------------------------------------------------------------
// Texture handling
//-----------------------------------------------------------------------------

- (void)scaleWithFormat:(G3D_TexFmt)fmt type:(G3D_TexType)tp geometry:(G3DTexGeometry)gm
{
}

- (void)addSubImage:(void*)img level:(int)l format:(G3D_TexFmt)fmt type:(G3D_TexType)tp geometry:(G3DTexGeometry)gm
{
}

- (void)copySubImageAtLevel:(int)l geometry:(G3DTexGeometry)gm to:(G3DOffset)tg
{
}

//-----------------------------------------------------------------------------
// Texture filtering and environment handling
//-----------------------------------------------------------------------------

- (void)setBuildMipmaps:(BOOL)yn
{
  _createMipmaps = yn;
}

- (BOOL)buildMipmaps
{
  return _createMipmaps;
}

- (void)setMagFilter:(G3D_TexMagFilter)magf
{
  mag = magf;
}

- (G3D_TexMagFilter)magFilter;
{
  return mag;
}

- (void)setMinFilter:(G3D_TexMinFilter)minf
{
  min = minf;
}

- (G3D_TexMinFilter)minFilter
{
  return min;
}

- (G3D_TexEnvMode)textureEnvironmentMode
{
  return mode;
}

- (void)setTextureEnvironmentMode:(G3D_TexEnvMode)md
{
  mode = md;
}

- (void)setBlendColour:(G3DColour *)aCol
{
  [blendcol autorelease];
  blendcol = [aCol retain];
}

- (G3DColour *)blendColour
{
  return blendcol;
}

- (void)setTextureWrapping:(G3DWrapping)wr
{
  wrap = wr;
}

- (G3DWrapping)textureWrapping
{
  return wrap;
}

- (void)setBorderColour:(G3DColour *)bcol
{
  [bordercol autorelease];
  bordercol = [bcol retain];
}

- (G3DColour *)borderColour
{
  return bordercol;
}

- (void)setPriority:(double)p
{
  priority = CLAMP(p,0.0,1.0);
}

- (double)priority
{
  return priority;
}

- (G3DTexGeometry *)texGeometry;
{
  return &texGeometry;
}

//-----------------------------------------------------------------------------
// NSCopying
//-----------------------------------------------------------------------------

- (id)copyWithZone:(NSZone *)zone
{
  G3DTexture *copy = [super copyWithZone:zone];

  [copy setBlendColour:blendcol];
  [copy setBorderColour:bordercol];

  copy->imageData = [imageData retain];
  copy->_texFileName = _texFileName;

  copy->ifmt = ifmt;
  copy->fmt = fmt;
  copy->type = type;
  copy->mode = mode;
  copy->mag = mag;
  copy->min = min;

  copy->texGeometry = texGeometry;
  copy->wrap = wrap;

  copy->priority = priority;
  copy->_createMipmaps = _createMipmaps;
  copy->_components = _components;

  copy->_isBound = _isBound;

  return copy;
}

//-----------------------------------------------------------------------------
// NSCoding
//-----------------------------------------------------------------------------

- (void)encodeWithCoder:(NSCoder *)aCoder
{
  [super encodeWithCoder:aCoder];

  [aCoder encodeObject:imageData];
  [aCoder encodeObject:blendcol];
  [aCoder encodeObject:bordercol];
  [aCoder encodeObject:_texFileName];

  [aCoder encodeValueOfObjCType:@encode(int) at:&ifmt];
  [aCoder encodeValueOfObjCType:@encode(int) at:&fmt];
  [aCoder encodeValueOfObjCType:@encode(int) at:&type];
  [aCoder encodeValueOfObjCType:@encode(int) at:&mode];
  [aCoder encodeValueOfObjCType:@encode(int) at:&mag];
  [aCoder encodeValueOfObjCType:@encode(int) at:&min];

  [aCoder encodeValueOfObjCType:@encode(int) at:&wrap.s];
  [aCoder encodeValueOfObjCType:@encode(int) at:&wrap.t];

  [aCoder encodeValueOfObjCType:@encode(int) at:&texGeometry.width];
  [aCoder encodeValueOfObjCType:@encode(int) at:&texGeometry.height];
  [aCoder encodeValueOfObjCType:@encode(int) at:&texGeometry.border];
  [aCoder encodeValueOfObjCType:@encode(int) at:&texGeometry.offset.x];
  [aCoder encodeValueOfObjCType:@encode(int) at:&texGeometry.offset.y];

  [aCoder encodeValueOfObjCType:@encode(double) at:&priority];

  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&_createMipmaps];
  [aCoder encodeValueOfObjCType:@encode(int) at:&_components];
}

- (id)initWithCoder:(NSCoder *)aCoder
{
  self = [super initWithCoder:aCoder];

  imageData = [[aCoder decodeObject] retain];
  blendcol = [[aCoder decodeObject] retain];
  bordercol = [[aCoder decodeObject] retain];
  _texFileName = [[aCoder decodeObject] retain];

  [aCoder decodeValueOfObjCType:@encode(int) at:&ifmt];
  [aCoder decodeValueOfObjCType:@encode(int) at:&fmt];
  [aCoder decodeValueOfObjCType:@encode(int) at:&type];
  [aCoder decodeValueOfObjCType:@encode(int) at:&mode];
  [aCoder decodeValueOfObjCType:@encode(int) at:&mag];
  [aCoder decodeValueOfObjCType:@encode(int) at:&min];

  [aCoder decodeValueOfObjCType:@encode(int) at:&wrap.s];
  [aCoder decodeValueOfObjCType:@encode(int) at:&wrap.t];

  [aCoder decodeValueOfObjCType:@encode(int) at:&texGeometry.width];
  [aCoder decodeValueOfObjCType:@encode(int) at:&texGeometry.height];
  [aCoder decodeValueOfObjCType:@encode(int) at:&texGeometry.border];
  [aCoder decodeValueOfObjCType:@encode(int) at:&texGeometry.offset.x];
  [aCoder decodeValueOfObjCType:@encode(int) at:&texGeometry.offset.y];

  [aCoder decodeValueOfObjCType:@encode(double) at:&priority];

  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&_createMipmaps];
  [aCoder decodeValueOfObjCType:@encode(int) at:&_components];

  /*
   * Lazy bind the texture if needed!
   */

  _isBound = NO;

  return self;
}

//-----------------------------------------------------------------------------
// G3DSceneCoding
//-----------------------------------------------------------------------------

- (void)encodeWithSceneArchiver:(G3DArchiver *)archiver
{
  [super encodeWithSceneArchiver:archiver];

  [archiver encodeObject:imageData named:@"imageData"];
  [archiver encodeObject:blendcol named:@"blendcol"];
  [archiver encodeObject:bordercol named:@"bordercol"];
  [archiver encodeObject:_texFileName named:@"_texFileName"];

  [archiver encodeInt:ifmt named:@"ifmt"];
  [archiver encodeInt:fmt named:@"fmt"];
  [archiver encodeInt:type named:@"type"];
  [archiver encodeInt:mode named:@"mode"];
  [archiver encodeInt:mag named:@"mag"];
  [archiver encodeInt:min named:@"min"];

  [archiver encodeDouble:priority named:@"priority"];

  [archiver encodeInt:wrap.s named:@"wrap.s"];
  [archiver encodeInt:wrap.t named:@"wrap.t"];

  [archiver encodeInt:texGeometry.width named:@"texGeometry.width"];
  [archiver encodeInt:texGeometry.height named:@"texGeometry.height"];
  [archiver encodeInt:texGeometry.border named:@"texGeometry.border"];
  [archiver encodeInt:texGeometry.offset.x named:@"texGeometry.offset.x"];
  [archiver encodeInt:texGeometry.offset.y named:@"texGeometry.offset.y"];
}

- (id)initWithSceneUnarchiver:(G3DUnarchiver *)unarchiver
{
  self = [super initWithSceneUnarchiver:unarchiver];

  imageData = [[unarchiver decodeObjectNamed:@"imageData"] retain];
  blendcol = [[unarchiver decodeObjectNamed:@"blendcol"] retain];
  bordercol = [[unarchiver decodeObjectNamed:@"bordercol"] retain];
  _texFileName = [[unarchiver decodeObjectNamed:@"_texFileName"] retain];

  [unarchiver decodeInt:&ifmt named:@"ifmt"];
  [unarchiver decodeInt:&fmt named:@"fmt"];
  [unarchiver decodeInt:&type named:@"type"];
  [unarchiver decodeInt:&mode named:@"mode"];
  [unarchiver decodeInt:&mag named:@"mag"];
  [unarchiver decodeInt:&min named:@"min"];

  [unarchiver decodeDouble:&priority named:@"priority"];

  [unarchiver decodeInt:&wrap.s named:@"wrap.s"];
  [unarchiver decodeInt:&wrap.t named:@"wrap.t"];
 
  [unarchiver decodeInt:&texGeometry.width named:@"texGeometry.width"];
  [unarchiver decodeInt:&texGeometry.height named:@"texGeometry.height"];
  [unarchiver decodeInt:&texGeometry.border named:@"texGeometry.border"];
  [unarchiver decodeInt:&texGeometry.offset.x named:@"texGeometry.offset.x"];
  [unarchiver decodeInt:&texGeometry.offset.y named:@"texGeometry.offset.y"];

  /*
   * Lazy binding!
   */
    
  [self readTextureFromFile:_texFileName];
  _isBound = NO;

  return self;
}

@end

@implementation G3DTexture (TextureLoading)

- (void)readTextureFromFile:(NSString *)aFile
{
  if ([[aFile pathExtension] caseInsensitiveCompare:@"BMP"] == NSOrderedSame) {
    [self readBMPTextureFromFile:aFile];
  }
  else if ([[aFile pathExtension] caseInsensitiveCompare:@"TGA"] == NSOrderedSame) {
    [self readTGATextureFromFile:aFile];
  }
  else {
    [NSException raise:G3DTextureReadingException format:@"'%@' is not a supported image format!",[aFile pathExtension]];
  }
}

- (void)readBMPTextureFromFile:(NSString *)aPath
{
  NSData *tex;
  unsigned cur = 18;
  unsigned long buf;
  unsigned long size;
  unsigned short int planes;
  unsigned short int bpp;
  unsigned long i;
  char *c;
  char tmp;

  if (!(tex = [NSData dataWithContentsOfFile:aPath])) {
    [NSException raise:G3DTextureReadingException format:@"Couldn't load %@!",aPath];
    return;  
  }

  [tex deserializeBytes:&buf length:4 atCursor:&cur];
  texGeometry.width = (int)buf;
  if (texGeometry.width < 64) {
    [NSException raise:G3DTextureImageSizeException format:@"Texture width too small (%d)!",texGeometry.width];
    return;
  }

  [tex deserializeBytes:&buf length:4 atCursor:&cur];
  texGeometry.height = (int)buf;
  if (texGeometry.height < 64) {
    [NSException raise:G3DTextureImageSizeException format:@"Texture height too small(%d)!",texGeometry.height];
    return;
  }

  size = texGeometry.width * texGeometry.height * 3;

  [tex deserializeBytes:&planes length:2 atCursor:&cur];
  if (planes != 1) {
    [NSException raise:G3DTextureReadingException format:@"Can't handle multiple planes!"];
    return;
  }

  [tex deserializeBytes:&bpp length:2 atCursor:&cur];
  if (bpp != 24) {
    [NSException raise:G3DTextureReadingException format:@"bpp = %d, but only 24bpp are supported right now!",bpp];
    return;
  }

  cur += 24;
  imageData = [[tex subdataWithRange:NSMakeRange(cur,size)] mutableCopy];

  c = (char *)[imageData mutableBytes];
  for (i=0;i<size;i+=3) {
    tmp = c[i];
    c[i] = c[i+2];
    c[i+2] = tmp;
  }

  // Some more defaults
  texGeometry.offset.x = 0;
  texGeometry.offset.y = 0;
  texGeometry.border = 0;

  fmt = G3D_RGB;
  ifmt = G3D_RGB_INT;
  _components = 3;
}

- (void)readTGATextureFromFile:(NSString *)aPath
{
  NSData *tex;
  unsigned cur = 0;
  int imageBits;
  unsigned size;
  unsigned ln;
  unsigned char _tga_type[4];
  unsigned char info[7];
  char *c;
  char tmp;
  unsigned i;

  // Get the file and some basic information
  if (!(tex = [NSData dataWithContentsOfFile:aPath])) {
    [NSException raise:G3DTextureReadingException format:@"Couldn't load %@!",aPath];
    return;
  }

  [tex deserializeBytes:&_tga_type length:(3 * sizeof (char)) atCursor:&cur];

  cur= 12;
  [tex deserializeBytes:&info length:(6 * sizeof (char)) atCursor:&cur];
  if (_tga_type[1] != 0 || (_tga_type[2] != 2 && _tga_type[2] != 3)) {
    [NSException raise:G3DTextureReadingException format:@"Unsupported TGA type!"];
    return;
  }

  texGeometry.width  = info[0] + info[1] * 256;
  texGeometry.height = info[2] + info[3] * 256;
  imageBits = info[4];

  if (texGeometry.width < 64) {
    [NSException raise:G3DTextureImageSizeException format:@"Texture size must be at least 64x64pixels!"];
    return;
  }

  size = texGeometry.width * texGeometry.height;

  if (imageBits != 32 && imageBits != 24 && imageBits != 8) {
    [NSException raise:G3DTextureReadingException format:@"bpp = %d, but only 32, 24 and 8bpp are supported right now!",imageBits];
    return;
  }

  // Now load the image into memory
  switch (imageBits) {
    case 8:
      ln = size;
      fmt = G3D_ALPHA;
      ifmt = G3D_LUMINANCE_INT; // Correct?
      _components = 1;
      break;
    case 24:
      ln = size * 3;
      fmt = G3D_RGB;  
      ifmt = G3D_RGB_INT;
      _components = 3;
      break;
    case 32:
      ln = size * 4;
      fmt = G3D_RGBA;  
      ifmt = G3D_RGBA_INT;
      _components = 4;
      break;
  }
  
  // Get the image data
  cur += _tga_type[0];
  imageData = [[tex subdataWithRange:NSMakeRange(cur,ln)] mutableCopy];

  if (imageBits == 24) {
    c = (char *)[imageData mutableBytes];
    for (i = 0; i < ln; i += 3) {
      tmp      = c[i];
      c[i]     = c[i + 2];
      c[i+2]   = tmp;
    }    
  }
  else if (imageBits == 32) {
    int h;
    float *tc;
    size_t size = texGeometry.width * 4 * sizeof(char);

    c = (char *)[imageData mutableBytes];
    for (i = 0; i < ln; i += 4) {
      tmp      = c[i];
      c[i]     = c[i + 2];
      c[i+2]   = tmp;
    }    

    tc = malloc(size);

    for (i = 0,h = texGeometry.height-1; i < texGeometry.height/2; i++,h--) {
      char *dhs = c+h*size;
      NSRange lowRange = {i*size, size}; 
      NSRange highRange = {h*size, size}; 

      [imageData getBytes:tc range:lowRange];
      [imageData replaceBytesInRange:lowRange withBytes:dhs];
      [imageData replaceBytesInRange:highRange withBytes:tc];
    }    

    free(tc);
  }
  
  // Some more defaults
  texGeometry.offset.x = 0;
  texGeometry.offset.y = 0;
  texGeometry.border = 0;
}

@end


