//-----------------------------------------------------------------------------
// Project            	RenderKit
// Class		G3DTimedSwitchGroup
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Thu Sep  9 11:05:08 CEST 1999
//
// Copyright (c) Philippe C.D. Robert
//
// The RenderKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DTimedSwitchGroup.m,v 1.13 2001/07/15 12:39:24 probert Exp $
//-----------------------------------------------------------------------------

#import "G3DTimedSwitchGroup.h"
#import "G3DTransform.h"
#import "G3DFunctions.h"
#import "G3DRenderEngine.h"
#import "G3DDefines.h"
#import "G3DBVol.h"
#import "G3DFrustum.h"
#import "G3DArchiver.h"
#import "G3DUnarchiver.h"

@interface G3DTimedSwitchGroup (IndexSelection)

- (void)_selectNextIndex;

@end

@implementation G3DTimedSwitchGroup (IndexSelection)

- (void)_selectNextIndex
{
  index++;

  if (index == _children) {
    index = 0;
  }
}

@end

@implementation G3DTimedSwitchGroup 

//-----------------------------------------------------------------------------
// Init and free
//-----------------------------------------------------------------------------

- (id)initWithCapacity:(unsigned int)cap timeInterval:(NSTimeInterval)ti
{
  if ((self = [super initWithCapacity:cap])) {
    timer = [NSTimer scheduledTimerWithTimeInterval:ti target:self selector:@selector(_selectNextIndex) userInfo:nil repeats:YES];
  }
  return self;
}

- (void)dealloc
{
  [timer invalidate];
  timer = nil;
  
  [super dealloc];  
}

//-----------------------------------------------------------------------------
// Rendering and traversing the scenegraph
//-----------------------------------------------------------------------------

- (void)renderWithMask:(unsigned)rmask frustum:(G3DFrustum *)fr
{
  if (!isVisible) {
    return;
  }

  if (_shouldNotifyBeforeRendering) {
    [delegate graphNodeWillRender:self];
  }

  if ((rmask & (1 << VIEW_FRUSTUM_CULLING))) {
    if (_shouldVFCull && [fr containsVolume:bVolume] == G3DOutside) {
      DLOG(@"View-Frustum culling: Node and its children won't be rendered");
      return;
    }
  }

  G3D_PUSH_MATRIX;

  if (!_isCompiled) {
    [transformation transform];
  }

  // Render! (if _isCompiled and *current* nodeList -> optimise!)
  [[children objectAtIndex:index] renderWithMask:rmask frustum:fr];
  
  G3D_POP_MATRIX;

  if (_shouldNotifyAfterRendering) {
    [delegate graphNodeDidRender:self];
  }
}

//-----------------------------------------------------------------------------
// Accessor methods
//-----------------------------------------------------------------------------

- (void)setInterval:(NSTimeInterval)ti
{
  interval = ti;
}

- (NSTimeInterval)interval
{
  return interval;  
}

//-----------------------------------------------------------------------------
// NSCoding
//-----------------------------------------------------------------------------

- (void)encodeWithCoder:(NSCoder *)aCoder
{
  [super encodeWithCoder:aCoder];

  [aCoder encodeValueOfObjCType:@encode(NSTimeInterval) at:&interval];
}

- (id)initWithCoder:(NSCoder *)aCoder
{
  self = [super initWithCoder:aCoder];

  [aCoder decodeValueOfObjCType:@encode(NSTimeInterval) at:&interval];

  self->timer = [NSTimer scheduledTimerWithTimeInterval:interval target:self selector:@selector(_selectNextIndex) userInfo:nil repeats:YES];

  return self;
}

//-----------------------------------------------------------------------------
// NSCopying
//-----------------------------------------------------------------------------

- (id)copyWithZone:(NSZone *)zone
{
  G3DTimedSwitchGroup *copy = [super copyWithZone:zone];

  [copy setInterval:interval];
  copy->timer = [NSTimer scheduledTimerWithTimeInterval:interval target:copy selector:@selector(_selectNextIndex) userInfo:nil repeats:YES];

  return copy;
}

//-----------------------------------------------------------------------------
// G3DSceneCoding
//-----------------------------------------------------------------------------

- (void)encodeWithSceneArchiver:(G3DArchiver *)archiver
{
  [super encodeWithSceneArchiver:archiver];

  [archiver encodeDouble:interval named:@"interval"];
}

- (id)initWithSceneUnarchiver:(G3DUnarchiver *)unarchiver
{
  self = [super initWithSceneUnarchiver:unarchiver];
  
  [unarchiver decodeDouble:&interval named:@"interval"];

  self->timer = [NSTimer scheduledTimerWithTimeInterval:interval target:self selector:@selector(_selectNextIndex) userInfo:nil repeats:YES];

  return self;
}

@end






