//-----------------------------------------------------------------------------
// Project            	RenderKit
// Class		G3DTransform
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Thu Sep  9 11:05:08 CEST 1999
//
// Copyright (c) Philippe C.D. Robert
//
// The RenderKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DTransform.m,v 1.18 2001/04/05 16:51:58 phr Exp $
//-----------------------------------------------------------------------------

#import "G3DTransform.h"
#import "G3DGroup.h"
#import "G3DGraphNode.h"
#import "G3DArchiver.h"
#import "G3DUnarchiver.h"
#import "G3DFunctions.h"

#import <GeometryKit/G3DMatrix4f.h>
#import <GeometryKit/G3DVector3f.h>
#import <GeometryKit/G3DFunctions.h>

NSString *G3DTransformInversionException = @"G3DTransformInversionException";
NSString *G3DTransformException = @"G3DTransformException";

@interface G3DTransform (Private)

- (int)type;

@end

@implementation G3DTransform (Private)

- (int)type
{
  return _type;
}

@end

@implementation G3DTransform

static G3DTransform *_identity = nil;

//-----------------------------------------------------------------------------
// Class methods
//-----------------------------------------------------------------------------

+ (G3DTransform *)identity
{
  if (!_identity) {
    _identity = [[G3DTransform alloc] init];
  }
  return _identity;
}

+ (G3DTransform *)transform
{
  return [[[G3DTransform alloc] init] autorelease];
}

+ (G3DTransform *)transformWithScale:(G3DVector3f *)scale
{
  return [[[G3DTransform alloc] initWithScale:scale hpr:nil translation:nil] autorelease];
}

+ (G3DTransform *)transformWithTranslation:(G3DVector3f *)transl
{
  return [[[G3DTransform alloc] initWithScale:nil hpr:nil translation:transl] autorelease];
}

+ (G3DTransform *)transformWithHPR:(G3DVector3f *)hpr
{
  return [[[G3DTransform alloc] initWithScale:nil hpr:hpr translation:nil] autorelease];
}

+ (G3DTransform *)transformWithScale:(G3DVector3f *)scale hpr:(G3DVector3f *)hpr translation:(G3DVector3f *)translation
{
  return [[[G3DTransform alloc] initWithScale:scale hpr:hpr translation:translation] autorelease];
}

//-----------------------------------------------------------------------------
// Init
//-----------------------------------------------------------------------------

- (id)init
{
  if ((self = [super init])) {
    G3DMakeIdentity4fv(m);
    
    G3DMakeIdentity4fv(transformation.scale);
    G3DMakeIdentity4fv(transformation.hpr);
    G3DMakeIdentity4fv(transformation.translation);

    _type = IDENTITY;
    _tmpMat = [[G3DMatrix4f alloc] init];
  }
  return self;
}

- (id)initWithScale:(G3DVector3f *)scaleVec hpr:(G3DVector3f *)hprVec translation:(G3DVector3f *)transVec;
{
  if ((self = [self init])) {
    if (scaleVec) {
      G3DCopyVector3fv(_scale,[scaleVec elements]);
      G3DMakeScale4fv(transformation.scale,_scale);
      shouldScale = YES;
    }
    else {
      G3DMakeIdentity4fv(transformation.scale);
      shouldScale = NO;
    }

    if (hprVec) {
      G3DCopyVector3fv(_hpr,[hprVec elements]);
      G3DMakeEulerTransform4fv(transformation.hpr,_hpr);
      shouldHPR = YES;
    }
    else {
      G3DMakeIdentity4fv(transformation.hpr);
      shouldHPR = NO;
    }

    if (transVec) {
      G3DCopyVector3fv(_translation,[transVec elements]);
      G3DMakeTranslation4f(transformation.translation,_translation);
      shouldTranslate = YES;
    }
    else {
      G3DMakeIdentity4fv(transformation.translation);
      shouldTranslate = NO;
    }

    _type = GENERIC;
  }
  return self;
}

- (void)dealloc
{
  [_tmpMat release];

  [super dealloc];
}

//-----------------------------------------------------------------------------
// Transformation
//-----------------------------------------------------------------------------

- (void)transform
{
  if (shouldTranslate) {
    glTranslatef(_translation[0],_translation[1],_translation[2]);
  }
  if (shouldHPR) {
    glMultMatrixf(transformation.hpr);
  }
  if (shouldScale) {
    glScalef(_scale[0],_scale[1],_scale[2]);
  }
}

//-----------------------------------------------------------------------------
// Accessor methods
//-----------------------------------------------------------------------------

- (BOOL)isIdentity
{
  return (_type == IDENTITY) ? YES : NO;
}

- (void)makeIdentity
{
  G3DMakeIdentity4fv(m);
  
  G3DMakeIdentity4fv(transformation.scale);
  G3DMakeIdentity4fv(transformation.hpr);
  G3DMakeIdentity4fv(transformation.translation);

  shouldScale = NO;
  shouldHPR = NO;
  shouldTranslate = NO;
  
  _type = IDENTITY;
}

- (void)setScale:(G3DVector3f *)scaleVec
{
  [self setScaleValues:[scaleVec elements]];
}

- (void)setScaleValues:(float *)scaleValues
{
  G3DCopyVector3fv(_scale,scaleValues);
  G3DMakeScale4fv(transformation.scale,_scale);
  shouldScale = YES;

  _type = GENERIC;

  if (node) {
    node->_transfIsDirty=YES;
  }
}

- (void)addScale:(G3DVector3f *)scaleVec
{
  [self addScaleValues:[scaleVec elements]];
}

- (void)addScaleValues:(float *)scaleValues
{
  float newScale[3];

  G3DAddVectors3fv(newScale,scaleValues,_scale);
  [self setScaleValues:newScale];
}

- (void)getScaleValues:(float *)scaleValues
{
  G3DCopyVector3fv(scaleValues,_scale);
}

- (void)getScaleMatrix:(G3DMatrix4f **)mat
{
  [*mat setElements:transformation.scale];
}

- (void)setTranslation:(G3DVector3f *)translationVec
{
  [self setTranslationValues:[translationVec elements]];
}

- (void)setTranslationValues:(float *)translationValues
{
  G3DCopyVector3fv(_translation,translationValues);
  G3DMakeTranslation4f(transformation.translation,_translation);
  shouldTranslate = YES;

  _type = GENERIC;

  if (node) {
    node->_transfIsDirty=YES;
  }
}

- (void)addTranslation:(G3DVector3f *)translationVec
{
  [self addTranslationValues:[translationVec elements]];
}

- (void)addTranslationValues:(float *)translationValues
{
  float newTr[3];

  G3DAddVectors3fv(newTr,translationValues,_translation);
  [self setTranslationValues:newTr];
}

- (void)getTranslationValues:(float *)translationValues
{
  G3DCopyVector3fv(translationValues,_translation);
}

- (void)getTranslationMatrix:(G3DMatrix4f **)mat
{
  [*mat setElements:transformation.translation];
}

- (void)setHPR:(G3DVector3f *)hprVec
{
  [self setHPRValues:[hprVec elements]];
}

- (void)setHPRValues:(float *)hprValues
{
  G3DCopyVector3fv(_hpr,hprValues);
  G3DMakeEulerTransform4fv(transformation.hpr,_hpr);
  shouldHPR = YES;

  _type = GENERIC;
  _rotMatIsUsed = NO;

  if (node) {
    node->_transfIsDirty=YES;
  }
}

- (void)addHPR:(G3DVector3f *)hprVec
{
  [self addHPRValues:[hprVec elements]];
}

- (void)addHPRValues:(float *)hprValues
{
  float newHPR[3];

  if (_rotMatIsUsed) {
    [NSException raise:G3DTransformException format:@"Using a custom rotation matrix!"];
    return;
  }

  G3DAddVectors3fv(newHPR,hprValues,_hpr);
  [self setHPRValues:newHPR];
}

- (void)getHPRValues:(float *)hprValues
{
  G3DCopyVector3fv(hprValues,_hpr);
}

- (void)getHPRMatrix:(G3DMatrix4f **)mat
{
  [*mat setElements:transformation.hpr];
}

- (void)setRotationWithElements:(float *)elements
{
  G3DCopyMatrix4fv(transformation.hpr,elements);

  _rotMatIsUsed = YES;
  _hpr[0] = -1.0;
  _hpr[1] = -1.0;
  _hpr[2] = -1.0;

  shouldHPR = YES;
}

- (void)getTransformationValues:(float *)mat
{
  float tmp[16];

  G3DCopyMatrix4fv(tmp,transformation.scale);
  G3DMultiplyMatrix4fv(m,transformation.hpr,tmp);
  G3DMultiplyMatrix4fv(mat,transformation.translation,m);
}

- (void)getTransformation:(G3DMatrix4f **)mat
{
  float tmp[16];

  G3DCopyMatrix4fv(tmp,transformation.scale);
  G3DMultiplyMatrix4fv(m,transformation.hpr,tmp);
  G3DMultiplyMatrix4fv(tmp,transformation.translation,m);

  [*mat setElements:tmp];
}

- (void)getInverseTransformation:(G3DMatrix4f **)mat
{
  float tmp[16];

  // We should cache the inverse...
  G3DCopyMatrix4fv(tmp,transformation.scale);
  G3DMultiplyMatrix4fv(m,transformation.hpr,tmp);
  G3DMultiplyMatrix4fv(tmp,transformation.translation,m);

  if (!G3DInvertMatrix4fv(im,tmp)) {
    [NSException raise:G3DTransformInversionException format:@"Could not invert the transformation %x!",self];
  }
  [*mat setElements:im];
}

- (void)setNode:(G3DGraphNode *)nd;
{
  node = nd;
}

- (G3DGraphNode *)node
{
  return node;  
}

- (NSString *)description
{
  return [NSString stringWithFormat:@"%@ %x in thread %x:\nRotation %@\n Translation %@\n Scale %@",[self class],self,[NSThread currentThread],G3DStringFromMatrix4f(transformation.hpr),G3DStringFromMatrix4f(transformation.translation),G3DStringFromMatrix4f(transformation.scale)];
}

//-----------------------------------------------------------------------------
// NSCoding
//-----------------------------------------------------------------------------

- (void)encodeWithCoder:(NSCoder *)aCoder
{
  [super encodeWithCoder:aCoder];

  [aCoder encodeObject:_tmpMat];
  [aCoder encodeArrayOfObjCType:@encode(float) count:16 at:m];
  [aCoder encodeValueOfObjCType:@encode(int) at:&_type];

  [aCoder encodeArrayOfObjCType:@encode(float) count:16 at:transformation.scale];
  [aCoder encodeArrayOfObjCType:@encode(float) count:16 at:transformation.hpr];
  [aCoder encodeArrayOfObjCType:@encode(float) count:16 at:transformation.translation];

  [aCoder encodeArrayOfObjCType:@encode(float) count:3 at:_scale];
  [aCoder encodeArrayOfObjCType:@encode(float) count:3 at:_hpr];
  [aCoder encodeArrayOfObjCType:@encode(float) count:3 at:_translation];

  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&shouldScale];
  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&shouldHPR];
  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&shouldTranslate];
  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&_rotMatIsUsed];

  [aCoder encodeObject:node];
}

- (id)initWithCoder:(NSCoder *)aCoder
{
  self = [super initWithCoder:aCoder];

  _tmpMat = [[aCoder decodeObject] retain];
  [aCoder decodeArrayOfObjCType:@encode(float) count:16 at:m];
  [aCoder decodeValueOfObjCType:@encode(int) at:&_type];

  [aCoder decodeArrayOfObjCType:@encode(float) count:16 at:transformation.scale];
  [aCoder decodeArrayOfObjCType:@encode(float) count:16 at:transformation.hpr];
  [aCoder decodeArrayOfObjCType:@encode(float) count:16 at:transformation.translation];

  [aCoder decodeArrayOfObjCType:@encode(float) count:3 at:_scale];
  [aCoder decodeArrayOfObjCType:@encode(float) count:3 at:_hpr];
  [aCoder decodeArrayOfObjCType:@encode(float) count:3 at:_translation];

  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&shouldScale];
  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&shouldHPR];
  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&shouldTranslate];
  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&_rotMatIsUsed];

  node = [aCoder decodeObject]; // Correct?

  return self;
}

//-----------------------------------------------------------------------------
// NSCopying
//-----------------------------------------------------------------------------

- (id)copyWithZone:(NSZone *)zone
{
  G3DTransform *copy = [super copyWithZone:zone];

  copy->_type = _type;
  copy->_tmpMat = [_tmpMat retain];
  [copy setNode:node];

  copy->transformation = transformation;

  G3DCopyMatrix4fv(copy->m,m);
  G3DCopyMatrix4fv(copy->im,im);

  G3DCopyVector3fv(copy->_translation,_translation);
  G3DCopyVector3fv(copy->_scale,_scale);
  G3DCopyVector3fv(copy->_hpr,_hpr);

  copy->shouldHPR = shouldHPR;
  copy->shouldTranslate = shouldTranslate;
  copy->shouldScale = shouldScale;

  copy->_rotMatIsUsed = _rotMatIsUsed;

  return copy;
}

//-----------------------------------------------------------------------------
// G3DSceneCoding
//-----------------------------------------------------------------------------

- (void)encodeWithSceneArchiver:(G3DArchiver *)archiver
{
  [super encodeWithSceneArchiver:archiver];

  [archiver encodeObject:_tmpMat named:@"_tmpMat"];
  [archiver encodeArrayOfTypeFloatAt:&m[0] count:16 named:@"m"];
  [archiver encodeInt:_type named:@"_type"];
  [archiver encodeObject:node named:@"node"];

  [archiver encodeArrayOfTypeFloatAt:&transformation.scale[0] count:16 named:@"transformation.scale"];
  [archiver encodeArrayOfTypeFloatAt:&transformation.hpr[0] count:16 named:@"transformation.hpr"];
  [archiver encodeArrayOfTypeFloatAt:&transformation.translation[0] count:16 named:@"transformation.translation"];

  [archiver encodeArrayOfTypeFloatAt:&_scale[0] count:3 named:@"_scale"];
  [archiver encodeArrayOfTypeFloatAt:&_hpr[0] count:3 named:@"_hpr"];
  [archiver encodeArrayOfTypeFloatAt:&_translation[0] count:3 named:@"_translation"];

  [archiver encodeBOOL:shouldHPR named:@"shouldHPR"];
  [archiver encodeBOOL:shouldTranslate named:@"shouldTranslate"];
  [archiver encodeBOOL:shouldScale named:@"shouldScale"];
  [archiver encodeBOOL:_rotMatIsUsed named:@"_rotMatIsUsed"];
}

- (id)initWithSceneUnarchiver:(G3DUnarchiver *)unarchiver
{
  self = [super initWithSceneUnarchiver:unarchiver];
  
  _tmpMat = [[unarchiver decodeObjectNamed:@"_tmpMat"] retain];
  [unarchiver decodeArrayOfTypeFloatAt:&m[0] count:16 named:@"m"];
  [unarchiver decodeInt:&_type named:@"_type"];
  node = [unarchiver decodeObjectNamed:@"node"]; // Correct?

  [unarchiver decodeArrayOfTypeFloatAt:&transformation.scale[0] count:16 named:@"transformation.scale"];
  [unarchiver decodeArrayOfTypeFloatAt:&transformation.hpr[0] count:16 named:@"transformation.hpr"];
  [unarchiver decodeArrayOfTypeFloatAt:&transformation.translation[0] count:16 named:@"transformation.translation"];

  [unarchiver decodeArrayOfTypeFloatAt:&_scale[0] count:3 named:@"_scale"];
  [unarchiver decodeArrayOfTypeFloatAt:&_hpr[0] count:3 named:@"_hpr"];
  [unarchiver decodeArrayOfTypeFloatAt:&_translation[0] count:3 named:@"_translation"];

  [unarchiver decodeBOOL:&shouldHPR named:@"shouldHPR"];
  [unarchiver decodeBOOL:&shouldTranslate named:@"shouldTranslate"];
  [unarchiver decodeBOOL:&shouldScale named:@"shouldScale"];

  [unarchiver decodeBOOL:&_rotMatIsUsed named:@"_rotMatIsUsed"];

  return self;
}

@end






