//-----------------------------------------------------------------------------
// Project            	RenderKit
// Class		G3DUnarchiver
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Wed Nov 24 18:49:25 CET 1999
//
// Copyright (c) Philippe C.D. Robert
//
// The RenderKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DUnarchiver.m,v 1.13 2001/07/07 15:35:47 probert Exp $
//-----------------------------------------------------------------------------

#import "G3DUnarchiver.h"
#import "G3DArchiver.h"
#import "G3DFunctions.h"

@implementation G3DUnarchiver

//-----------------------------------------------------------------------------
// Class methods
//-----------------------------------------------------------------------------

static NSMutableDictionary* classToAliasMappings = nil;

+ (void) initialize
{
  static BOOL tooLate = NO;

  if (!tooLate) {
    classToAliasMappings = [[NSMutableDictionary alloc] init];
    tooLate = YES;
  }
}

+ (id)unarchiveObjectTreeWithContentsOfFile:(NSString*)filename
{
  id plist = [[NSString stringWithContentsOfFile:filename] propertyList];

  if (!plist) {
    return nil;
  }

  return [G3DUnarchiver unarchiveObjectNamed:@"RootObject" fromPropertyList:plist];
}

+ (id)unarchiveObjectNamed:(NSString*)name fromPropertyList:(id)prList
{
  G3DUnarchiver *ua;

  ua = [[[self alloc] initForReadingWithPropertyList:prList] autorelease];

  return [ua decodeObjectNamed:name];
}

+ (id)unarchiveObjectNamed:(NSString*)name fromFile:(NSString*)path
{
  G3DUnarchiver *ua = [self unarchiveObjectTreeWithContentsOfFile:path];

  return [ua decodeObjectNamed:name];
}

+ (NSString*)decodedClassNameForArchiveClassName:(NSString*)archName
{
  return archName;
}

+ (void)decodeClassName:(NSString*)archName asClassName:(NSString*)trueName
{
  [classToAliasMappings setObject:trueName forKey:archName];
}

//-----------------------------------------------------------------------------
// Init and free
//-----------------------------------------------------------------------------

- (id)init
{
  return [self initForReadingWithPropertyList:nil];
}

- (id)initForReadingWithPropertyList:(id)prList
{
  if ((self = [super init])) {
    NSString* versionString;
    
    pList = [prList copy];
    currentDecodedObjectRep = pList;

    // 119?
    namesToObjects = [NSMutableDictionary dictionaryWithCapacity:119];
    
    // Version information
    versionString = [pList objectForKey:@"Version"];
    [[NSScanner scannerWithString: versionString] scanInt:&version];
    
    objectZone = NSDefaultMallocZone ();
  }
  return self;
}

- (void)dealloc
{
  [pList release];
  [super dealloc];
}

//-----------------------------------------------------------------------------
// Unarchiving
//-----------------------------------------------------------------------------

- (NSZone*)objectZone
{
  return objectZone;
}

- (void)setObjectZone:(NSZone*)zone
{
  objectZone = zone;
}

- (unsigned int)systemVersion
{
  return version;
}

- (NSString*)decodedClassNameForArchiveClassName:(NSString*)archName
{
  NSString* className = [classToAliasMappings objectForKey:archName];

  return className ? className : archName;
}

- (void)decodeClassName:(NSString*)archName asClassName:(NSString*)trueName
{
  [classToAliasMappings setObject:trueName forKey:archName];
}

- (unsigned int)versionForClassName:(NSString*)className
{
  return 1;
}

//-----------------------------------------------------------------------------
// Decoding objects
//-----------------------------------------------------------------------------

- (id)decodeObjectNamed:(NSString*)name
{
  id object;
  id label;
  id  representation;
  id className;
  id upperObjectRepresentation;
  BOOL objectOnTopLevel = NO;
  id newObject;
  Class class;
  NSString* decodeAsName;

  if (!name) {
    return nil;
  }

  if (level) {
    // First try to see if the object has been already decoded
    if ((object = [namesToObjects objectForKey: name])) {
      return object;
    }
  }

  // The object has not been decoded yet. Read its label from the current object dictionary representation.
  label = [currentDecodedObjectRep objectForKey:name];

  if (label) {
    // Try to see if the object has been decoded using `label' as name
    if ((object = [namesToObjects objectForKey: label])) {
      return object;
    }
  }
  else {
    // Try to find the object on the top level
    label = [pList objectForKey: name];

    if (label) {
      objectOnTopLevel = YES;
    }
    else {
      // There is no object with this name within the current object or on the top level.

#ifdef DEBUG
      NSLog (@"No object named '%@' in object representation '%@'", name, currentDecodedObjectRep);
#endif DEBUG
      return nil;
    }
  }
  
  // If we are on the top level the description is really the representation of the object. Otherwise the value is the name of an object on the top level.

  if (currentDecodedObjectRep != pList && !objectOnTopLevel){
    NSAssert1 ([label isKindOfClass: [NSString class]], @"Label is not a string: '%@'!", label);

    // label is either a name of an object on the top level dictionary or the string "nil" which means the object has the nil value.
    if ([label isEqual: @"nil"]) {
      return nil;
    }

    representation = [pList objectForKey:label];
  }
  else {
    representation = label;
    label = name;
  }

  if (!representation) {
    // There is no object with such a label on the top level dictionary
    NSLog (@"No object object named '%@' on the top level dictionary! (error " @"within object representation '%@')", label, currentDecodedObjectRep);
    return nil;
  }

  // Temporary save the current object representation
  upperObjectRepresentation = currentDecodedObjectRep;
  currentDecodedObjectRep = representation;

  // Create the object
  className = [representation objectForKey: @"isa"];
  decodeAsName = [classToAliasMappings objectForKey: className];

  if ( decodeAsName ) {
    NSLog(@"G3DArchiver: %@ to be decoded as %@", className, decodeAsName);
    className = decodeAsName;
  }

  class = NSClassFromString(className);
  object = [class createObjectForSceneUnarchiver:self];

  if (!class) {
    NSLog (@"Class %@ not linked into application!", className);
    return nil;
  }

  // Push it into the dictionary of known objects
  [namesToObjects setObject:object forKey:label];

  // Read it from dictionary
  level++;
  newObject = [object initWithSceneUnarchiver:self];
  level--;

  if (newObject != object) {
    object = newObject;
    [namesToObjects setObject:object forKey:label];
  }

  // Restore the current object representation
  currentDecodedObjectRep = upperObjectRepresentation;

  return object;
}

- (NSString*)decodeStringNamed:(NSString*)name
{
  id string;

  if (!name) {
    return nil;
  }

  string = [currentDecodedObjectRep objectForKey: name];
  if (!string) {
    NSLog (@"Couldn't find the string value for key '%@' (object '%@')", name, currentDecodedObjectRep);
    return nil;
  }

  if (![string isKindOfClass:[NSString class]]) {
    NSLog (@"Decoded object is not a string: '%@'! (key '%@', object '%@')", string, name, currentDecodedObjectRep);
    return nil;
  }

  if ([string isEqualToString: @"nil"]) {
    return nil;
  }
  
  return string;
}

- (NSArray*)decodeArrayNamed:(NSString*)name
{
  id array;
  id decodedArray;
  int i;
  int count;

  if (!name) {
    return nil;
  }

  array = [currentDecodedObjectRep objectForKey: name];
  if (!array) {
    NSLog (@"Couldn't find the array value for key '%@' (object '%@')", name, currentDecodedObjectRep);
    return nil;
  }

  if (![array isKindOfClass:[NSArray class]]) {
    NSLog (@"Decoded object is not an array: '%@'! (key '%@', object '%@')", array, name, currentDecodedObjectRep);
    return nil;
  }

  count = [array count];
  decodedArray = [NSMutableArray arrayWithCapacity: count];

  for (i = 0; i < count; i++) {
    id label = [array objectAtIndex: i];
    id objectDescription = [pList objectForKey: label];

    if (!objectDescription) {
      NSLog (@"warning: couldn't find the description for object labeled '%@' "@"in the array description '%@ = %@'!", label, name, array);
      continue;
    }

    [decodedArray addObject: [self decodeObjectNamed:label]];
  }

  return decodedArray;
}

- (NSDictionary*)decodeDictionaryNamed:(NSString*)name
{
  id dictionary;
  id decodedDictionary;
  id enumerator;
  id keyLabel;
  id valueLabel;

  if (!name) {
    return nil;
  }

  dictionary = [currentDecodedObjectRep objectForKey:name];
  if (!dictionary) {
    NSLog (@"Couldn't find the dictionary value for key '%@' (object '%@')", name, currentDecodedObjectRep);
    return nil;
  }

  if (![dictionary isKindOfClass: [NSDictionary class]]) {
    NSLog (@"Decoded object is not a dictionary: '%@'! (key '%@', object '%@')", dictionary, name, currentDecodedObjectRep);
    return nil;
  }

  decodedDictionary = [NSMutableDictionary dictionaryWithCapacity:[dictionary count]];
  enumerator = [dictionary keyEnumerator];
  while ((keyLabel = [enumerator nextObject])) {
    id key;
    id value;
    id objectDescription;

    objectDescription = [pList objectForKey: keyLabel];
    if (!objectDescription) {
      NSLog (@"warning: couldn't find the description for object labeled '%@' "@"in the dictionary description '%@ = %@'!",keyLabel, name, dictionary);
      continue;
    }
    key = [self decodeObjectNamed: keyLabel];

    valueLabel = [dictionary objectForKey: keyLabel];
    objectDescription = [pList objectForKey: valueLabel];
    if (!objectDescription) {
      NSLog (@"warning: couldn't find the description for object labeled '%@' "@"in the dictionary description '%@ = %@'!",valueLabel, name, dictionary);
      continue;
    }

    value = [self decodeObjectNamed:valueLabel];
    [decodedDictionary setObject:value forKey:key];
  }

  return decodedDictionary;
}

- (NSData*)decodeDataNamed:(NSString*)name
{
  id data;

  if (!name) {
    return nil;
  }

  data = [currentDecodedObjectRep objectForKey: name];
  if (!data) {
    NSLog (@"Couldn't find the data value for key '%@' (object '%@')", name, currentDecodedObjectRep);
    return nil;
  }

  if (![data isKindOfClass: [NSData class]]) {
    NSLog (@"Decoded object is not a data: '%@'! (key '%@', object '%@')", data, name, currentDecodedObjectRep);
    return nil;
  }
  
  return data;
}

- (Class)decodeClassNamed:(NSString*)name
{
  NSString* className = [self decodeStringNamed:name];
  
  return className ? NSClassFromString (className) : Nil;
}

- (SEL)decodeSelectorNamed:(NSString*)name
{
  NSString* selectorName = [self decodeStringNamed:name];

  return selectorName ? NSSelectorFromString (selectorName) : NULL;
}

//-----------------------------------------------------------------------------
// Decoding C Types
//-----------------------------------------------------------------------------

- (void)decodeBytesWithReturnedLength:(unsigned *)l named:(NSString *)name
{
#warning decodeBytesWithReturnedLength:named: not yet done!
}

- (void)decodeArrayOfTypeDoubleAt:(double *)address count:(unsigned)cnt named:(NSString*)name
{
  NSArray *array;
  int i;

  if (!name) {
    return;
  }

  array = [currentDecodedObjectRep objectForKey:name];
  if (!array) {
    return;
  }

  for (i=0;i<cnt;i++) {
    address[i] = [[array objectAtIndex:i] doubleValue];;
  }
}

- (void)decodeArrayOfTypeFloatAt:(float *)address count:(unsigned)cnt named:(NSString*)name
{
  NSArray *array;
  int i;

  if (!name) {
    return;
  }

  array = [currentDecodedObjectRep objectForKey:name];
  if (!array) {
    return;
  }

  for (i=0;i<cnt;i++) {
    address[i] = [[array objectAtIndex:i] floatValue];;
  }
}

- (void)decodeArrayOfTypeIntAt:(int *)address count:(unsigned)cnt named:(NSString*)name
{
  NSArray *array;
  int i;

  if (!name) {
    return;
  }

  array = [currentDecodedObjectRep objectForKey:name];
  if (!array) {
    return;
  }

  for (i=0;i<cnt;i++) {
    address[i] = [[array objectAtIndex:i] intValue];;
  }
}

- (void)decodeChar:(char *)value named:(NSString*)name;
{
  NSString* valueString;

  if (!name) {
    return;
  }

  valueString = [currentDecodedObjectRep objectForKey:name];
  if (!valueString) {
    return;
  }
  else {
    strcpy(value,[valueString cString]);
  }
}

- (void)decodeUnsignedChar:(unsigned char *)value named:(NSString*)name;
{
  NSString* valueString;

  if (!name) {
    return;
  }

  valueString = [currentDecodedObjectRep objectForKey:name];
  if (!valueString) {
    return;
  }
  else {
    strcpy(value,[valueString cString]);
  }
}

- (void)decodeBOOL:(BOOL *)value named:(NSString*)name;
{
  NSString* valueString;

  if (!name) {
    return;
  }

  valueString = [currentDecodedObjectRep objectForKey: name];
  if (!valueString) {
    return;
  }

  *value = [valueString compare:@"YES" options:NSCaseInsensitiveSearch] == NSOrderedSame;
}

- (void)decodeShort:(short *)value named:(NSString*)name;
{
  [self decodeInt:(int *)value named:name];
}

- (void)decodeUnsignedShort:(unsigned short *)value named:(NSString*)name;
{
  [self decodeInt:(int *)value named:name];
}

- (void)decodeInt:(int *)value named:(NSString*)name
{
  NSString* valueString;
  int val;

  if (!name) {
    return;
  }

  valueString = [currentDecodedObjectRep objectForKey: name];
  if (!valueString) {
    return;
  }

  if (![[NSScanner scannerWithString:valueString] scanInt:&val]) {
    return;
  }
  else {
    *value = val;
  }
}

- (void)decodeUnsignedInt:(unsigned *)value named:(NSString*)name;
{
  [self decodeInt:(int *)value named:name];
}

- (void)decodeLong:(long *)value named:(NSString*)name;
{
  [self decodeInt:(int *)value named:name];
}

- (void)decodeUnsignedLong:(unsigned long *)value named:(NSString*)name;
{
  [self decodeInt:(int *)value named:name];
}

- (void)decodeFloat:(float *)value named:(NSString*)name;
{
  NSString* valueString;
  float val;

  if (!name) {
    return;
  }

  valueString = [currentDecodedObjectRep objectForKey: name];
  if (!valueString) {
    return;
  }

  if (![[NSScanner scannerWithString:valueString] scanFloat:&val]) {
    return;
  }
  else {
    *value = val;
  }
}

- (void)decodeDouble:(double *)value named:(NSString*)name;
{
  NSString* valueString;
  double val;

  if ( !name ) {
    return;
  }

  valueString = [currentDecodedObjectRep objectForKey: name];
  if (!valueString) {
    return;
  }

  if ( ![[NSScanner scannerWithString:valueString] scanDouble: &val]) {
    return;
  }
  else {
    *value = val;
  }
}

//-----------------------------------------------------------------------------
// Decoding other DTs
//-----------------------------------------------------------------------------

- (NSPoint)decodePointNamed:(NSString*)name
{
  NSString *valueString;

  if (!name) {
    return NSZeroPoint;
  }

  valueString = [currentDecodedObjectRep objectForKey: name];
  if (!valueString) {
    return NSZeroPoint;
  }

  return NSPointFromString(valueString);
}

- (NSSize)decodeSizeNamed:(NSString*)name
{
  NSString* valueString;

  if (!name) {
    return NSZeroSize;
  }

  valueString = [currentDecodedObjectRep objectForKey: name];
  if (!valueString) {
    return NSZeroSize;
  }

  return NSSizeFromString(valueString);
}

- (NSRect)decodeRectNamed:(NSString*)name
{
  NSString* valueString;

  if (!name) {
    return NSZeroRect;
  }

  valueString = [currentDecodedObjectRep objectForKey: name];
  if (!valueString) {
    return NSZeroRect;
  }

  return NSRectFromString(valueString);
}

- (G3DFrustumValues)decodeFrustumNamed:(NSString*)name
{
}

- (G3DCoordinate)decodeCoordNamed:(NSString*)name
{
}

- (G3DRange)decodeRangeNamed:(NSString*)name
{
}

- (G3DBoundingSphere)decodeBoundingSphereNamed:(NSString*)name
{
}

- (G3DBoundingBox)decodeBoundingBoxNamed:(NSString*)name
{
}

/*
- (G3D_CameraFlags)decodeCameraFlagsNamed:(NSString*)name
{
}
 */

@end






