//-----------------------------------------------------------------------------
// Project            	RenderKit
// Class		G3DVtxArray
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Thu Sep  9 11:05:08 CEST 1999
//
// Copyright (c) Philippe C.D. Robert
//
// The RenderKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DVtxArray.m,v 1.23 2001/01/28 17:01:11 robert Exp $
//-----------------------------------------------------------------------------

#import "G3DVtxArray.h"
#import "G3DShape.h"
#import "G3DDefines.h"
#import "G3DArchiver.h"
#import "G3DUnarchiver.h"

NSString *G3DVtxArrayException = @"G3DVtxArrayException";
NSString *G3DVtxArrayFormatException = @"G3DVtxArrayFormatException";
NSString *G3DVtxArrayLockingException = @"G3DVtxArrayLockingException";

@implementation G3DVtxArray

//-----------------------------------------------------------------------------
// Init
//-----------------------------------------------------------------------------

- (id)initWithMode:(G3DGeometryMode)mode count:(int)cnt
{
  NSAssert(cnt>0,@"No valid geometry provided!");

  if ((self = [super init])) {
    size_t n;

    geometryMode = mode;

    vertices = cnt;    

    n=vertices*SIZE_VERTEX*sizeof(float);      
    if ((vertexArray = malloc(n)) != NULL) {
      memset(vertexArray,0,n);
    }
    else {
      [self autorelease];
      [NSException raise:G3DVtxArrayException 
		   format:@"<%@ %x>: Could not allocate space for a vertex array",[self class],self];
      return nil;
    }

    /*
     * These arrays should be allocated lazily!
     *
     */

    n=vertices*SIZE_COLOUR*sizeof(float);      
    if ((colourArray = malloc(n)) != NULL) {
      memset(colourArray,0,n);
    }
    else {
      [self autorelease];
      [NSException raise:G3DVtxArrayException 
		   format:@"<%@ %x>: Could not allocate space for a vertex array",[self class],self];
      return nil;
    }

    n=vertices*SIZE_NORMAL*sizeof(float);      
    if ((normalArray = malloc(n)) != NULL) {
      memset(normalArray,0,n);
    }
    else {
      [self autorelease];
      [NSException raise:G3DVtxArrayException 
		   format:@"<%@ %x>: Could not allocate space for a vertex array",[self class],self];
      return nil;
    }

    n=vertices*SIZE_TCOORD*sizeof(float);      
    if ((texCoordArray = malloc(n)) != NULL) {
      memset(texCoordArray,0,n);
    }
    else {
      [self autorelease];
      [NSException raise:G3DVtxArrayException 
		   format:@"<%@ %x>: Could not allocate space for a vertex array",[self class],self];
      return nil;
    }

    isLocked = NO;

    _hasNormals = NO;
    _hasColours = NO;
    _hasTexCoords = NO;
  }
  return self;
}

- (void)dealloc
{
  free(vertexArray);
  free(colourArray);
  free(normalArray);
  free(texCoordArray);

#ifdef GL_EXT_compiled_vertex_array
  if (isLocked) {
    glUnlockArraysEXT();
  }
#endif

  [super dealloc];
}

//-----------------------------------------------------------------------------
// Drawing
//-----------------------------------------------------------------------------

- (void)drawGeometry
{
  G3D_PUSH_MATRIX;

  /*
   * The client state handling should be optimised!
   */

  if (_hasColours) {
    glEnableClientState(GL_COLOR_ARRAY);
    glColorPointer(SIZE_COLOUR,GL_FLOAT,0,colourArray);
  }
  else {
    glDisableClientState(GL_COLOR_ARRAY);
    glColor4f(0.8,0.2,0.3,1.0f);
  }

  if (_hasNormals) {
    glEnableClientState(GL_NORMAL_ARRAY);
    glNormalPointer(GL_FLOAT,0,normalArray);
  }
  else {
    glDisableClientState(GL_NORMAL_ARRAY);
  }

  if (_hasTexCoords) {
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glTexCoordPointer(SIZE_TCOORD,GL_FLOAT,0,texCoordArray);
  }
  else {
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
  }

  glEnableClientState(GL_VERTEX_ARRAY);
  glVertexPointer(SIZE_VERTEX,GL_FLOAT,0,vertexArray);

  glDrawArrays(geometryMode,0,vertices);

  G3D_POP_MATRIX;
}

//-----------------------------------------------------------------------------
// Accessor Methods
//-----------------------------------------------------------------------------

- (void)precompile
{
  if (isLocked) {
    [NSException raise:G3DVtxArrayLockingException 
		 format:@"Array is already locked!"];
    return;
  }

  isLocked = YES;

#ifdef GL_EXT_compiled_vertex_array
  glLockArraysEXT(0,vertices);
#endif
}

- (int)vertexCount
{
  return vertices;
}

- (float *)colourArray
{
  return colourArray;
}

- (float *)vertexArray
{
  return vertexArray;
}

- (float *)normalArray
{
  return normalArray;
}

- (float *)texCoordArray
{
  return texCoordArray;
}

- (G3DGeometryMode)geometryMode
{
  return geometryMode;
}

- (void)getMin:(float *)min max:(float *)max
{
  float *c;
  register int i;

  if (vertices == 0) {
    [NSException raise:G3DVtxArrayException 
		 format:@"No vertices found in %@ %x!",[self class],self];
    return;
  }

  DLOG(@"<%@ %x>: getMin:max: with %d vertices",[self class],self,vertices);

  for (i=0;i<SIZE_VERTEX;i++) {
    min[i] = vertexArray[i];
    max[i] = vertexArray[i];
  }

  for (i=0;i<vertices;i++) {
    c = &vertexArray[SIZE_VERTEX*i];
    
    min[0] = MIN(min[0],c[0]);
    min[1] = MIN(min[1],c[1]);
    min[2] = MIN(min[2],c[2]);
    min[3] = MIN(min[3],c[3]);        
    
    max[0] = MAX(max[0],c[0]);
    max[1] = MAX(max[1],c[1]);
    max[2] = MAX(max[2],c[2]);
    max[3] = MAX(max[3],c[3]);
  }
}

- (void)setShape:(G3DShape *)aShape
{
  shape = aShape;  
}

- (void)setUseVertexNormals:(BOOL)yn
{
  _hasNormals = yn;
}

//-----------------------------------------------------------------------------
// Geometry
//-----------------------------------------------------------------------------

- (void)setCoordinate:(float *)coord atIndex:(int)index
{
  int i;
  int _index;

  if (isLocked) {
    [NSException raise:G3DVtxArrayLockingException format:@"Array is already locked!"];
    return;
  }

  NSAssert(index<vertices,@"Invalid index !");

  _index = (SIZE_VERTEX*index);
  for (i=0;i<SIZE_VERTEX;i++) {
    vertexArray[_index+i] = coord[i];
  }
  
  if (shape) {
    [shape setNeedsUpdate:YES];
  }
}

- (float *)coordinateAtIndex:(int)index
{
  NSAssert(index<vertices,@"Invalid index !");
  return &vertexArray[SIZE_VERTEX*index];
}

- (void)setCoordinates:(float *)coords atIndex:(int)index count:(int)cnt
{
  int i;
  int j;
  int _index;

  if (isLocked) {
    [NSException raise:G3DVtxArrayLockingException format:@"Array is already locked!"];
    return;
  }

  NSAssert(index<vertices,@"Invalid index !");

  for (j=0;j<cnt;j++) {
    int k=j*SIZE_VERTEX;

    _index = SIZE_VERTEX*(index+j);
    for (i=0;i<SIZE_VERTEX;i++) {
      vertexArray[_index+i] = coords[k+i];
    }
  }

  if (shape) {
    [shape setNeedsUpdate:YES];
  }
}

- (void)setCoordinates:(float *)coords range:(NSRange)rng
{
  [self setCoordinates:coords atIndex:rng.location count:rng.length];
}

- (void)setNormal:(float *)coord atIndex:(int)index
{
  int i;
  int _index;

  if (isLocked) {
    [NSException raise:G3DVtxArrayLockingException format:@"Array is already locked!"];
    return;
  }

  NSAssert(index<vertices,@"Invalid index !");

  _index = (SIZE_NORMAL*index);
  for (i=0;i<SIZE_NORMAL;i++) {
    normalArray[_index+i] = coord[i];
  }
  
  if (shape) {
    [shape setNeedsUpdate:YES];
  }
}

- (float *)normalAtIndex:(int)index
{
  NSAssert(index<vertices,@"Invalid index !");
  return &normalArray[SIZE_NORMAL*index];
}

- (void)setNormals:(float *)coords atIndex:(int)index count:(int)cnt
{
  int i;
  int j;
  int _index;

  if (isLocked) {
    [NSException raise:G3DVtxArrayLockingException format:@"Array is already locked!"];
    return;
  }

  NSAssert(index<vertices,@"Invalid index !");

  for (j=0;j<cnt;j++) {
    int k=j*SIZE_NORMAL;

    _index = SIZE_NORMAL*(index+j);
    for (i=0;i<SIZE_NORMAL;i++) {
      normalArray[_index+i] = coords[k+i];
    }
  }

  if (shape) {
    [shape setNeedsUpdate:YES];
  }
}

- (void)setNormals:(float *)coords range:(NSRange)rng
{
  [self setNormals:coords atIndex:rng.location count:rng.length];
}

//-----------------------------------------------------------------------------
// Colours
//-----------------------------------------------------------------------------

- (void)setColour:(float *)acol atIndex:(int)index
{
  int i;
  int _index;

  if (isLocked) {
    [NSException raise:G3DVtxArrayLockingException format:@"Array is already locked!"];
    return;
  }

  NSAssert(index<vertices,@"Invalid index !");

  _index = (SIZE_COLOUR*index);
  for (i=0;i<SIZE_COLOUR;i++) {
    colourArray[_index+i] = acol[i];
  }
  
  if (shape) {
    [shape setNeedsUpdate:YES];
  }
}

- (float *)colourAtIndex:(int)index
{
  NSAssert(index<vertices,@"Invalid index !");
  return &colourArray[SIZE_COLOUR*index];
}

- (void)setColours:(float *)cols atIndex:(int)index count:(int)cnt
{
  int i;
  int j;
  int _index;

  if (isLocked) {
    [NSException raise:G3DVtxArrayLockingException format:@"Array is already locked!"];
    return;
  }

  NSAssert(index<vertices,@"Invalid index !");

  for (j=0;j<cnt;j++) {
    int k=j*SIZE_COLOUR;

    _index = SIZE_COLOUR*(index+j);
    for (i=0;i<SIZE_COLOUR;i++) {
      colourArray[_index+i] = cols[k+i];
    }
  }

  if (shape) {
    [shape setNeedsUpdate:YES];
  }
}

- (void)setColours:(float *)cols range:(NSRange)rng
{
  [self setColours:cols atIndex:rng.location count:rng.length];
}

//-----------------------------------------------------------------------------
// Texture coordinates
//-----------------------------------------------------------------------------

- (void)setTexCoord:(float *)coord atIndex:(int)index
{
  int i;
  int _index;

  if (isLocked) {
    [NSException raise:G3DVtxArrayLockingException 
		 format:@"Array is already locked!"];
    return;
  }

  NSAssert(index<vertices,@"Invalid index !");

  _index = (SIZE_TCOORD*index);
  for (i=0;i<SIZE_TCOORD;i++) {
    texCoordArray[_index+i] = coord[i];
  }
  
  if (shape) {
    [shape setNeedsUpdate:YES];
  }
}

- (float *)texCoordAtIndex:(int)index
{
  NSAssert(index<vertices,@"Invalid index !");
  return &texCoordArray[SIZE_TCOORD*index];
}

- (void)setTexCoords:(float *)coords atIndex:(int)index count:(int)cnt
{
  int i;
  int j;
  int _index;

  if (isLocked) {
    [NSException raise:G3DVtxArrayLockingException 
		 format:@"Array is already locked!"];
    return;
  }

  NSAssert(index<vertices,@"Invalid index !");

  for (j=0;j<cnt;j++) {
    int k=j*SIZE_TCOORD;

    _index = SIZE_TCOORD*(index+j);
    for (i=0;i<SIZE_TCOORD;i++) {
      texCoordArray[_index+i] = coords[k+i];
    }
  }

  if (shape) {
    [shape setNeedsUpdate:YES];
  }
}

- (void)setTexCoords:(float *)coords range:(NSRange)rng
{
  [self setTexCoords:coords atIndex:rng.location count:rng.length];
}

//-----------------------------------------------------------------------------
// Information
//-----------------------------------------------------------------------------

- (NSString *)description
{
  return [NSString stringWithFormat:@"<%@ %x>: mode %d\n",
		   geometryMode,
		   [self class],
		   self];
}

//-----------------------------------------------------------------------------
// NSCoding
//-----------------------------------------------------------------------------

- (void)encodeWithCoder:(NSCoder *)aCoder
{
  [super encodeWithCoder:aCoder];

  [aCoder encodeObject:shape];  
  [aCoder encodeValueOfObjCType:@encode(int) at:&vertices];
  [aCoder encodeValueOfObjCType:@encode(int) at:&geometryMode];
  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&isLocked]; // ?

  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&_hasColours];
  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&_hasNormals];
  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&_hasTexCoords];

  [aCoder encodeBytes:&vertexArray length:vertices*SIZE_VERTEX*sizeof(float)];
  [aCoder encodeBytes:&colourArray length:vertices*SIZE_COLOUR*sizeof(float)];
  [aCoder encodeBytes:&normalArray length:vertices*SIZE_NORMAL*sizeof(float)];
  [aCoder encodeBytes:&texCoordArray length:vertices*SIZE_TCOORD*sizeof(float)];
}

- (id)initWithCoder:(NSCoder *)aCoder
{
  unsigned n;

  self = [super initWithCoder:aCoder];

  shape = [aCoder decodeObject];
  
  [aCoder decodeValueOfObjCType:@encode(int) at:&vertices];
  [aCoder decodeValueOfObjCType:@encode(int) at:&geometryMode];
  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&isLocked]; // ?

  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&_hasColours];
  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&_hasNormals];
  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&_hasTexCoords];

  n=vertices*SIZE_VERTEX*sizeof(float);
  vertexArray = (float*)[aCoder decodeBytesWithReturnedLength:&n];

  n=vertices*SIZE_COLOUR*sizeof(float);
  colourArray = (float*)[aCoder decodeBytesWithReturnedLength:&n];

  n=vertices*SIZE_NORMAL*sizeof(float);
  normalArray = (float*)[aCoder decodeBytesWithReturnedLength:&n];

  n=vertices*SIZE_TCOORD*sizeof(float);
  texCoordArray = (float*)[aCoder decodeBytesWithReturnedLength:&n];

  return self;
}

//-----------------------------------------------------------------------------
// NSCopying
//-----------------------------------------------------------------------------

- (id)copyWithZone:(NSZone *)zone
{
  G3DVtxArray *copy = [super copyWithZone:zone];
  size_t n;
  
  copy->vertices = vertices;
  copy->shape = shape;
  copy->geometryMode = geometryMode;
  copy->isLocked = isLocked;

  copy->_hasColours = _hasColours;
  copy->_hasNormals = _hasNormals;
  copy->_hasTexCoords = _hasTexCoords;
  
  n=vertices*SIZE_VERTEX*sizeof(float);      
  if ((copy->vertexArray = malloc(n)) != NULL) {
    copy->vertexArray = memcpy(copy->vertexArray,vertexArray,n);
  }
  else {
    [copy autorelease];
    [NSException raise:G3DVtxArrayException 
		 format:@"<%@ %x>: Could not allocate space for a vertex array",[self class],self];
    return nil;
  }
  
    n=vertices*SIZE_COLOUR*sizeof(float);      
    if ((copy->colourArray = malloc(n)) != NULL) {
      copy->colourArray = memcpy(copy->colourArray,colourArray,n);
    }
    else {
      [copy autorelease];
      [NSException raise:G3DVtxArrayException 
		   format:@"<%@ %x>: Could not allocate space for a vertex array",[self class],self];
      return nil;
    }

    n=vertices*SIZE_NORMAL*sizeof(float);      
    if ((copy->normalArray = malloc(n)) != NULL) {
      copy->normalArray = memcpy(copy->normalArray,normalArray,n);
    }
    else {
      [copy autorelease];
      [NSException raise:G3DVtxArrayException 
		   format:@"<%@ %x>: Could not allocate space for a vertex array",[self class],self];
      return nil;
    }

    n=vertices*SIZE_TCOORD*sizeof(float);      
    if ((copy->texCoordArray = malloc(n)) != NULL) {
      copy->texCoordArray = memcpy(copy->texCoordArray,texCoordArray,n);
    }
    else {
      [copy autorelease];
      [NSException raise:G3DVtxArrayException 
		   format:@"<%@ %x>: Could not allocate space for a vertex array",[self class],self];
      return nil;
    }

  return copy;
}

//-----------------------------------------------------------------------------
// G3DSceneCoding
//-----------------------------------------------------------------------------

- (void)encodeWithSceneArchiver:(G3DArchiver *)archiver
{
  [super encodeWithSceneArchiver:archiver];

  [archiver encodeObject:shape named:@"shape"];  

  [archiver encodeInt:vertices named:@"vertices"];
  [archiver encodeInt:geometryMode named:@"geometryMode"];
  [archiver encodeBOOL:isLocked named:@"isLocked"]; // ?

  [archiver encodeBOOL:_hasColours named:@"_hasColours"];
  [archiver encodeBOOL:_hasNormals named:@"_hasNormals"];
  [archiver encodeBOOL:_hasTexCoords named:@"_hasTexCoords"];

#warning encodeBytes:length:named: is not yet implemented!

  /*
  [archiver encodeBytes:&vertexArray length:vertices*SIZE_VERTEX*sizeof(float)];
  [archiver encodeBytes:&colourArray length:vertices*SIZE_COLOUR*sizeof(float)];
  [archiver encodeBytes:&normalArray length:vertices*SIZE_NORMAL*sizeof(float)];
  [archiver encodeBytes:&texCoordArray length:vertices*SIZE_TCOORD*sizeof(float)];
  */
}

- (id)initWithSceneUnarchiver:(G3DUnarchiver *)unarchiver
{
  // unsigned n;

  self = [super initWithSceneUnarchiver:unarchiver];
  
  shape = [unarchiver decodeObjectNamed:@"shape"];
  
  [unarchiver decodeInt:&vertices named:@"vertices"];
  [unarchiver decodeInt:&geometryMode named:@"geometryMode"];
  [unarchiver decodeBOOL:&isLocked named:@"isLocked"]; // ?

  [unarchiver decodeBOOL:&_hasColours named:@"_hasColours"];
  [unarchiver decodeBOOL:&_hasNormals named:@"_hasNormals"];
  [unarchiver decodeBOOL:&_hasTexCoords named:@"_hasTexCoords"];

#warning decodeBytesWithReturnedLength:named: is not yet implemented!

  /*
  n=vertices*SIZE_VERTEX*sizeof(float);
  vertexArray = (float*)[unarchiver decodeBytesWithReturnedLength:&n];

  n=vertices*SIZE_COLOUR*sizeof(float);
  colourArray = (float*)[unarchiver decodeBytesWithReturnedLength:&n];

  n=vertices*SIZE_NORMAL*sizeof(float);
  normalArray = (float*)[unarchiver decodeBytesWithReturnedLength:&n];

  n=vertices*SIZE_TCOORD*sizeof(float);
  texCoordArray = (float*)[unarchiver decodeBytesWithReturnedLength:&n];
  */

  return self;
}

@end






