//-----------------------------------------------------------------------------
// Project            	RenderKit
// Class		G3DWorld
// Creator            	Philippe C.D. Robert
// Maintainer         	Philippe C.D. Robert
// Creation Date      	Thu Sep  9 11:05:08 CEST 1999
//
// Copyright (c) Philippe C.D. Robert
//
// The RenderKit is free software; you can redistribute it and/or modify it 
// under the terms of the GNU LGPL Version 2 as published by the Free 
// Software Foundation
//
// $Id: G3DWorld.m,v 1.37 2001/07/15 12:39:24 probert Exp $
//-----------------------------------------------------------------------------

#import "G3DWorld.h"

#import "G3DNodeList.h"
#import "G3DGroup.h"
#import "G3DFunctions.h"
#import "G3DRenderEngine.h"
#import "G3DFog.h"
#import "G3DLight.h"
#import "G3DCamera.h"
#import "G3DArchiver.h"
#import "G3DUnarchiver.h"

NSString *G3DWorldParentException = @"G3DWorldParentException";

@interface G3DWorld (LightRendering)

- (void)lightAdded:(NSNotification *)notif;
- (void)lightRemoved:(NSNotification *)notif;

@end

@implementation G3DWorld

static Class defaultWorldClass = NULL;
static BOOL useMultipleThreads = NO;

//-----------------------------------------------------------------------------
// Class methods
//-----------------------------------------------------------------------------

+ (void)initialize
{
  static BOOL isInitialised = NO;

  if (isInitialised == NO) {
    isInitialised = YES;

    defaultWorldClass = [G3DWorld class];
  }
}

+ (void)setDefaultWorldClass:(Class)rootClass
{
  // Test if rootClass is subclass of G3DWorld or implements G3DRootNode
  defaultWorldClass = rootClass;
}

+ (Class)defaultWorldClass
{
  return defaultWorldClass;
}

+ (void)setUseMultipleThreads:(BOOL)yn
{
  useMultipleThreads = yn;
}

+ (G3DWorld *)world
{
  return [[[defaultWorldClass alloc] init] autorelease];
}

//-----------------------------------------------------------------------------
// Init and Free
//-----------------------------------------------------------------------------

- (id)init
{
  if ((self = [super init])) {
    staticObjects = [[G3DGroup alloc] init];
    [staticObjects setName:@"Static Objects"];

    dynamicObjects = [[G3DGroup alloc] init];
    [dynamicObjects setName:@"Dynamic Objects"];

    overlays = [[G3DGroup alloc] init];
    [overlays setName:@"Overlays"];

    _s_objs = 0;
    _d_objs = 0;
    _d_lights = 0;
    _overlays = 0;
    
    _useFog = NO;

    /*
     * Dynamic GL lighting
     */

    _dLights = [[NSMutableArray alloc] initWithCapacity:8];
    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(lightRemoved:) name:G3DLightRemovedNotification object:nil];

    [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(lightAdded:) name:G3DLightAddedNotification object:nil];

    /*
     * Multithreading
     */

    //updater = [[G3DGraphUpdater alloc] init];
  }
  return self;
}

- (void)dealloc
{
  [[NSNotificationCenter defaultCenter] removeObserver:self];

  [staticObjects release];
  [dynamicObjects release];  
  [overlays release];
  [_dLights release];

  //[updater release];

  [super dealloc];
}

//-----------------------------------------------------------------------------
// Rendering
//-----------------------------------------------------------------------------

- (void)preprocess
{
  G3DFog *f;

  DLOG(@"Preprocessing world node %x (%@)...",self,[self class]);

  /*
   * Update the cached transformations etc.
   */

  DLOG(@"Optimising dynamic subgraphs...");

  if (_d_objs) {
    G3DEnableLighting();
    [dynamicObjects updateIfNeeded];
    [dynamicObjects optimiseIfNeeded];
    G3DDisableLighting();
  }

  DLOG(@"Optimising overlay subgraphs...");

  if (_overlays) {
    [overlays updateIfNeeded];
    [overlays optimiseIfNeeded];
  }

  DLOG(@"Compiling static subgraphs...");

  if (_s_objs) {
    [staticObjects updateIfNeeded];
    [staticObjects optimiseIfNeeded];

    [staticObjects compile];
  }
  
  DLOG(@"Setting up fog if needed...");

  if ((f = [[G3DCamera activeCamera] fog])) {
    G3DBindFog(f);
    _useFog = YES;
  }
  else {
    _useFog = NO;
  }

  DLOG(@"...done!");
}

- (void)renderWithMask:(unsigned)mask frustum:(G3DFrustum *)fr
{
  G3DCamera *cam = [G3DCamera activeCamera];
  NSSize s = [cam size];

  if (_shouldNotifyBeforeRendering) {
    [delegate graphNodeWillRender:self];
  }

  if ((mask & (1 << ENABLE_GL_LIGHTS))) {
    register int i;

    DLOG(@"Rendering with dynamic lighting...");
    G3DEnableLighting();

    // Update the lights' transformations here!!!
    for (i=0;i<_d_lights;i++) {
      G3DLight *l = [_dLights objectAtIndex:i];
      int h = l->_lhandle;
      
      // If the light is directional, this is the direction!
      glLightfv((GL_LIGHT0+h),GL_POSITION,l->_trnPos);
      
      if (l->lightType == G3D_SpotLight) {
	glLightfv((GL_LIGHT0+h),GL_SPOT_DIRECTION,l->_trnDir);
      }
    }
  }
  else {
    DLOG(@"Rendering without dynamic lighting...");
    G3DDisableLighting();
  }

  /*
   * Update and check the sensors, if any...
   *
   */


  /*
   * Render the static and dynamic subgraphs
   *
   */

  [staticObjects renderWithMask:mask frustum:fr];

  if (useMultipleThreads == NO) {
    [dynamicObjects updateIfNeeded];
    [dynamicObjects updateBVolumeIfNeeded];
    [dynamicObjects renderWithMask:mask frustum:fr];

    // At the end render overlays
    if ((mask & (1 << ENABLE_GL_LIGHTS))) {
      G3DDisableLighting();
    }
    
    if (_overlays) {
      glDisable(GL_DEPTH_TEST);
      glMatrixMode(GL_PROJECTION);
      
      G3D_PUSH_MATRIX;
      
      glLoadIdentity();
      glOrtho(0,s.width,0,s.height,-1,1);
      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();
      
      [overlays updateIfNeeded];
      [overlays renderWithMask:mask frustum:fr];

      glMatrixMode(GL_PROJECTION);  // Select The Projection Matrix
      
      G3D_POP_MATRIX;
      
      glMatrixMode(GL_MODELVIEW);
      glEnable(GL_DEPTH_TEST);
    }
  }
  else { // Use multithreaded updating mechanisms
    /*
    [NSThread detachNewThreadSelector:@selector(updateTransformations:) 
	      toTarget:updater
	      withObject:dynamicObjects];
    
    // Once it's finished
    [NSThread detachNewThreadSelector:@selector(updateBVols:) 
	      toTarget:updater
	      withObject:dynamicObjects];

    */

    [NSException raise:@"G3DWorldMultiThreadingFailedException" 
		 format:@"Multithreading is not yet supported"];

    [dynamicObjects renderWithMask:mask frustum:fr];
    
    // At the end render overlays
    if ((mask & (1 << ENABLE_GL_LIGHTS))) {
      G3DDisableLighting();
    }
    
    if (_overlays) {
      /*
      [NSThread detachNewThreadSelector:@selector(updateTransformations:) 
		toTarget:updater
		withObject:overlays];
      */

      glDisable(GL_DEPTH_TEST);
      glMatrixMode(GL_PROJECTION);
      
      G3D_PUSH_MATRIX;
      
      glLoadIdentity();
      glOrtho(0,s.width,0,s.height,-1,1);
      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();

      /*
       * If finished updating...
       */
      
      [overlays renderWithMask:mask frustum:fr];

      glMatrixMode(GL_PROJECTION);  // Select The Projection Matrix
      
      G3D_POP_MATRIX;
      
      glMatrixMode(GL_MODELVIEW);
      glEnable(GL_DEPTH_TEST);
    }
  }
  
  if (_shouldNotifyAfterRendering) {
    [delegate graphNodeDidRender:self];
  }
}

//-----------------------------------------------------------------------------
// The parent
//-----------------------------------------------------------------------------

- (void)setParent:(G3DGraphNode *)aParent
{
  [NSException raise:G3DWorldParentException format:@"A world has no parent!"];
}

- (G3DGroup *)staticObjects
{
  return staticObjects;
}

- (void)addStaticObject:(G3DGraphNode *)aChild
{
  [staticObjects addChild:aChild];
  _s_objs++;
}

- (void)removeStaticObject:(G3DGraphNode *)aChild
{
  [staticObjects removeChild:aChild];
  _s_objs--;
}

- (G3DGroup *)dynamicObjects
{
  return dynamicObjects;
}

- (void)addDynamicObject:(G3DGraphNode *)aChild
{
  [dynamicObjects addChild:aChild];
  _d_objs++;
}

- (void)removeDynamicObject:(G3DGraphNode *)aChild
{
  [dynamicObjects removeChild:aChild];
  _d_objs--;
}

- (G3DGroup *)overlays
{
  return overlays;  
}

- (void)addOverlay:(G3DGraphNode *)aChild
{
  [overlays addChild:aChild];
  _overlays++;  
}

- (void)removeOverlay:(G3DGraphNode *)aChild
{
  [overlays removeChild:aChild];
  _overlays--;    
}

//-----------------------------------------------------------------------------
// NSCoding
//-----------------------------------------------------------------------------

- (void)encodeWithCoder:(NSCoder *)aCoder
{
  [super encodeWithCoder:aCoder];

  [aCoder encodeObject:staticObjects];
  [aCoder encodeObject:dynamicObjects];
  [aCoder encodeObject:_dLights];

  [aCoder encodeValueOfObjCType:@encode(int) at:&_s_objs];
  [aCoder encodeValueOfObjCType:@encode(int) at:&_d_objs];
  [aCoder encodeValueOfObjCType:@encode(int) at:&_d_lights];

  [aCoder encodeValueOfObjCType:@encode(BOOL) at:&_useFog];
}

- (id)initWithCoder:(NSCoder *)aCoder
{
  self = [super initWithCoder:aCoder];

  staticObjects = [[aCoder decodeObject] retain];
  dynamicObjects = [[aCoder decodeObject] retain];
  _dLights = [[aCoder decodeObject] retain];

  [aCoder decodeValueOfObjCType:@encode(int) at:&_s_objs];
  [aCoder decodeValueOfObjCType:@encode(int) at:&_d_objs];
  [aCoder decodeValueOfObjCType:@encode(int) at:&_d_lights];

  [aCoder decodeValueOfObjCType:@encode(BOOL) at:&_useFog];
 
  return self;
}

//-----------------------------------------------------------------------------
// NSCopying
//-----------------------------------------------------------------------------

- (id)copyWithZone:(NSZone *)zone
{
  G3DWorld *copy = [super copyWithZone:zone];

  /*
   * Deep copy!
   */
  
  copy->staticObjects = [staticObjects copy];
  copy->dynamicObjects = [dynamicObjects copy];
  copy->overlays = [overlays copy];
  
  copy->_s_objs = _s_objs;
  copy->_d_objs = _d_objs;
  copy->_d_lights = _d_lights;

  copy->_useFog = _useFog;

  copy->_dLights = [_dLights copy];
  
  return copy;
}

//-----------------------------------------------------------------------------
// G3DSceneCoding
//-----------------------------------------------------------------------------

- (void)encodeWithSceneArchiver:(G3DArchiver *)archiver
{
  [super encodeWithSceneArchiver:archiver];

  [archiver encodeObject:staticObjects named:@"staticObjects"];
  [archiver encodeObject:dynamicObjects named:@"dynamicObjects"];
  [archiver encodeObject:_dLights named:@"_dLights"];

  [archiver encodeInt:_s_objs named:@"_s_objs"];
  [archiver encodeInt:_d_objs named:@"_d_objs"];
  [archiver encodeInt:_d_lights named:@"_d_lights"];

  [archiver encodeBOOL:_useFog named:@"_useFog"];
}

- (id)initWithSceneUnarchiver:(G3DUnarchiver *)unarchiver
{
  self = [super initWithSceneUnarchiver:unarchiver];

  staticObjects = [[unarchiver decodeObjectNamed:@"staticObjects"] retain];
  dynamicObjects = [[unarchiver decodeObjectNamed:@"dynamicObjects"] retain];
  _dLights = [[unarchiver decodeObjectNamed:@"_dLights"] retain];

  [unarchiver decodeInt:&_s_objs named:@"_s_objs"];
  [unarchiver decodeInt:&_d_objs named:@"_d_objs"];
  [unarchiver decodeInt:&_d_lights named:@"_d_lights"];

  [unarchiver decodeBOOL:&_useFog named:@"_useFog"];

  return self;
}

@end

@implementation G3DWorld (LightRendering)

- (void)lightAdded:(NSNotification *)notif
{
  [_dLights addObject:[notif object]];

  _d_lights = [_dLights count];
}

- (void)lightRemoved:(NSNotification *)notif
{
  [_dLights removeObject:[notif object]];

  _d_lights = [_dLights count];
}

@end

